/* global Parameters, StdIcon_Error, StdButton_Ok, VERSION, TITLE, File, ImageOp_Mul, ImageOp_And, ImageOp_Sub, ImageOp_Add, Resample, StdButton_Ignore, StdIcon_Warning, StdButton_Cancel, NSG_EXTRA_CONTROLS, ImageIntegration, ProcessInstance, UndoFlag_ImageId, UndoFlag_NoSwapFile, TEST, ImageOp_Div, ImageWindow, DIVIDE_BY, ColorSpace_Gray, ColorSpace_RGB, ABORT, ASK_USER, StdButton_Abort */

// Version 1.0 (c) John Murphy 4th-Apr-2021
//
// ======== #license ===============================================================
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, version 3 of the License.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.
// =================================================================================
"use strict";
#feature-id NormalizeScaleGradient : Batch Processing > NormalizeScaleGradient 1.6

#feature-icon @script_icons_dir/NormalizeScaleGradient.svg

#feature-info Normalizes the scale and gradient to that of the reference image. \
Photometry is used to determine the scale, and a surface spline to model the relative gradient.<br/>\
Copyright &copy; 2019-2021 John Murphy.<br/> \
StarDetector.jsh: Copyright &copy; 2003-2020 Pleiades Astrophoto S.L. All Rights Reserved.<br/>

#define TITLE "NormalizeScaleGradient"
#define VERSION "1.6"
#define TEST false
#define DIVIDE_BY 1.05
#define ABORT 1
#define ASK_USER 2
#define MAX_SMOOTHNESS 4

#include <pjsr/UndoFlag.jsh>
#include <pjsr/DataType.jsh>
#include <pjsr/ImageOp.jsh>
#include <pjsr/ColorSpace.jsh>
#include "lib/NsgData.js"
#include "lib/FitsHeader.js"
#include "lib/FitsHeaderReader.js"
#include "lib/DialogControls.js"
#include "lib/LeastSquareFit.js"
#include "lib/StarLib.js"
#include "lib/Cache.js"
#include "lib/NsgDialog.js"
#include "lib/STFAutoStretch.js"
#include "lib/PreviewControl.js"
#include "lib/SampleGrid.js"
#include "lib/SampleGridDialog.js"
#include "lib/Graph.js"
#include "lib/DetectedStarsDialog.js"
#include "lib/PhotometryStarsDialog.js"
#include "lib/PhotometryGraphDialog.js"
#include "lib/Gradient.js"
#include "lib/GradientGraph.js"
#include "lib/GradientGraphDialog.js"
#include "extraControls/BinnedSampleGridDialog.js"

/**
 * Normalize tgtFile to the cached reference image
 * @param {NsgData} data
 * @param {String} tgtFile Full target filename, including path
 * @param {Noise} refNoiseRaw Noise estimates read from NOISExx FITS header(s)
 * @param {Number[]} refNoiseMRS First array item contains noise estimate for all channels
 * @param {Number[]} refNoiseKSigma First array item contains noise estimate for all channels
 * @param {SavedFile[]} savedFiles
 * @return {{summary: resultString, file: savedFilename, Number: weight, Number[] scaleFactors, Number: exposure, Boolean: failed} | undefined} 
 */
function normalizeScaleGradient(data, tgtFile, refNoiseRaw, refNoiseMRS, refNoiseKSigma, savedFiles){
    let resultString = File.extractName(tgtFile);
    if (!data.outputDir || data.outputDir.trim().length < 2){
        resultString = "Invalid output directory: '" + data.outputDir + "'";
        console.criticalln(resultString);
        return new Result(resultString, undefined, undefined, undefined, undefined, tgtFile, true);
    }
    let refFile = data.cache.getRefFilename();
    if (!refFile){
        return new Result("Error: No reference image.", undefined, undefined, undefined, undefined, tgtFile, true);
    }
    if (!tgtFile){
        return new Result("Error: No target image.", undefined, undefined, undefined, undefined, tgtFile, true);
    }
    if (!data.cache.setTgtFilename(tgtFile)){
        return new Result("Failed to read '" + File.extractName(tgtFile) + "'", undefined, undefined, undefined, undefined, tgtFile, true);
    }
    let nChannels = data.cache.isColor() ? 3 : 1;
    let savedFilename;
    let imageWindow = data.cache.getTgtImageWindow();
    let keywords = imageWindow.keywords;
    let fitsHeader = new FitsHeader();
    let exposure = fitsHeader.getExposure(keywords);
    resultString += ", " + (exposure ? exposure : 0) + "s" ;
    let noiseValues = [];
    let signalToNoise = [];
    let weights = [];
    let averageWeight;
    let scaleFactors;
    if (tgtFile !== refFile){
        // target file is not the reference file so it needs normalizing
        data.setPhotometryAutoValues(data.useAutoPhotometry, true);
        data.setSampleGenerationAutoValues(data.useAutoSampleGeneration, true);

        // Scale factors
        let colorStarPairs = getColorStarPairs(nChannels, data);
        scaleFactors = getScaleFactors(colorStarPairs, data);
        
        // === Sample pairs ===
        let sampleGrid = data.cache.getSampleGrid(data, tgtFile);
        let stars = data.cache.getRefStars(data.logStarDetection);
        let colorSamplePairs = sampleGrid.createColorScaledSamplePairs(stars, data, scaleFactors);

        for (let c=0; c<colorSamplePairs.length; c++){
            if (colorSamplePairs[c].length < 3) {
                let errMsg = "ERROR [" + File.extractName(tgtFile) + "] Too few samples to create a Surface Spline.";
                return new Result(errMsg, undefined, undefined, undefined, undefined, tgtFile, true);
            }
        }

        let surfaceSplines = [];
        console.writeln("\n<b><u>Calculating surface spline</u></b>");
        let binnedColorSamplePairs = createColorBinnedSamplePairs(data, colorSamplePairs);
        for (let c = 0; c < nChannels; c++) {
            let samplePairs = binnedColorSamplePairs[c];
            surfaceSplines[c] = calcSurfaceSpline(samplePairs, data.gradientSmoothness);
            console.writeln("Created surface spline [", c, "] smoothness ", 
                    data.gradientSmoothness, " from ", binnedColorSamplePairs[c].length, " samples");
            if (console.abortRequested){
                return undefined;
            }
        }

        console.writeln("\n<b><u>Normalizing</u></b>");
        let tgtImage = data.cache.getTgtView().image;
        imageWindow.mainView.beginProcess(UndoFlag_ImageId);
        let rect = new Rect(tgtImage.width, tgtImage.height);
        let tgtSamples = new Float32Array(rect.area);
        let gradSamples = new Float32Array(rect.area);
        let gradientImages = [];
        for (let c=0; c<nChannels; c++){
            gradientImages[c] = createGradientImage(data, rect, surfaceSplines[c]);
            if (!gradientImages[c].image){
                let errMsg = "ERROR [" + File.extractName(tgtFile) + "] Failed to create gradient image.";
                return new Result(errMsg, undefined, undefined, undefined, undefined, tgtFile, true);
            }
            if (console.abortRequested || !gradientImages[c].image){
                // TODO potential memory leak for images 0 to c-1 ?
                return undefined;
            }

            // Apply scale factor and then subtract gradient provided the tgtImage is non zero
            tgtImage.getSamples(tgtSamples, rect, c);
            gradientImages[c].image.getSamples(gradSamples, rect, 0);
            for (let i=0; i<tgtSamples.length; i++){
                if (tgtSamples[i]){
                    tgtSamples[i] = tgtSamples[i] * scaleFactors[c].m - gradSamples[i];
                }
            }
            // Finished with gradSamples typed array contents. gradSamples will be reused.
            gradientImages[c].image.free();
            tgtImage.setSamples(tgtSamples, rect, c);
        }
        console.writeln();
        
        let weightType;
        let tgtNoiseRaw = fitsHeader.getNoise(keywords);
        let useRawNoise = refNoiseRaw && tgtNoiseRaw && refNoiseRaw.index === tgtNoiseRaw.index;
        
        if (useRawNoise){
            let nWeights = 0;
            averageWeight = 0;
            for (let c=0; c<nChannels; c++){
                let index = refNoiseRaw.index !== -1 ? refNoiseRaw.index : c;
                signalToNoise[c] = refNoiseRaw.getSignalToNoise(refNoiseRaw, tgtNoiseRaw, index, scaleFactors[c].m);
                weights[c] = refNoiseRaw.getWeight(signalToNoise[c]);
                if (weights[c]){
                    noiseValues[c] = tgtNoiseRaw.noise[index];
                    nWeights++;
                    averageWeight += weights[c];
//                    if (refNoiseRaw.index === -1 && c ===1){
//                        // Green channel is from two pixels, so double its contribution
//                    }
                }
            }
            if (nWeights > 0){
                averageWeight /= nWeights;
                let weightStr = data.noiseWeightKeyword + ": " + averageWeight.toFixed(3);
                weightType = "NOISExx";
                console.writeln(weightStr);
                resultString += ", " + weightStr;
            } else {
                useRawNoise = false;
            }
        } 
        if (!useRawNoise){
            // Using corrected target image, so no need to apply a scale factor
            let useNoiseKSigma = true;
            if (refNoiseMRS[0] > 0 && refNoiseMRS[1] > 0.01){
                let tgtNoiseMRS = tgtImage.noiseMRS();
                if (tgtNoiseMRS[0] > 0 && tgtNoiseMRS[1] > 0.01){
                    useNoiseKSigma = false;
                    noiseValues[0] = tgtNoiseMRS[0];
                    signalToNoise[0] = refNoiseMRS[0] / tgtNoiseMRS[0];
                    weightType = "noiseMRS";
                }
            }
            if (useNoiseKSigma){
                let tgtNoiseKSigma = tgtImage.noiseKSigma();
                noiseValues[0] = tgtNoiseKSigma[0];
                if (tgtNoiseKSigma[0]){
                    signalToNoise[0] = refNoiseKSigma[0] / tgtNoiseKSigma[0];
                } else {
                    signalToNoise[0] = 0;
                }
                weightType = "noiseKSigma";
            }
            averageWeight = Math.pow( signalToNoise[0], 2 );
            for (let c=0; c<nChannels; c++){
                weights[c] = averageWeight;
            }
            let weightStr = data.noiseWeightKeyword + ": " + averageWeight.toFixed(3);
            console.writeln(weightStr);
            resultString += ", " + weightStr;
        }
        
        resultString += ", SNR: ";
        for (let c=0; c < signalToNoise.length; c++){
            if (c){
                resultString += ", ";
            }
            let snr = signalToNoise[c];
            if (snr){
                resultString += snr.toFixed(4);
            }
        }
        
        if (TEST){
            resultString += ", NSGS0 " + (1/scaleFactors[0].m).toFixed(3) + ", b=" + (scaleFactors[0].b).toPrecision(3);
        }
        
        let imageMaximum = tgtImage.maximum();
        fitsHeader.fitsHeaderRefFilename(keywords, File.extractName(refFile));
        fitsHeader.fitsHeaderGradient(keywords, data);
        fitsHeader.fitsHeaderNoise(keywords, noiseValues, weightType);
        fitsHeader.fitsHeaderMaxValue(keywords, imageMaximum);
        fitsHeader.fitsHeaderScale(keywords, scaleFactors);
        fitsHeader.fitsHeaderSignalToNoise(keywords, signalToNoise);
        fitsHeader.fitsHeaderNoiseWeight(keywords, data.noiseWeightKeyword, averageWeight, weightType);
                
        imageWindow.keywords = keywords;
        imageWindow.mainView.endProcess();
        let savedFile = saveFile(imageWindow, data, averageWeight, imageMaximum);
        savedFilename = savedFile.filename;
        savedFiles.push(savedFile);
        
        // Target image has changed (scaled / gradient) so cached data must be cleared
        data.cache.invalidateTgt();
        
        if (data.displayGradient){
            let gradientImage;
            let gradientPedestals = [];
            if (data.cache.isColor()){
                // Create new color image
                let smallWidth = gradientImages[0].smallImage.width;
                let smallHeight = gradientImages[0].smallImage.height;
                let smallRect = new Rect(smallWidth, smallHeight);
                gradientImage = new Image(smallWidth, smallHeight, 3, ColorSpace_RGB );
                for (let c=0; c<nChannels; c++){
                    // Reusing gradSamples typed array
                    let smallImage = gradientImages[c].smallImage;
                    smallImage.getSamples(gradSamples, smallRect, 0);
                    gradientImage.setSamples(gradSamples, smallRect, c);
                    gradientPedestals[c] = gradientImages[c].pedestal;
                    gradientImages[c].smallImage.free();
                } 
            }
            else {
                // Reuse existing grey smallImage
                gradientImage = gradientImages[0].smallImage;
                gradientPedestals[0] = gradientImages[0].pedestal;
            }
            let gradientResample = gradientImages[0].resample;
            saveGradientFile(gradientImage, savedFilename, "gradient", data, 
                    scaleFactors, weights, gradientResample, gradientPedestals);
            gradientImage.free();
        }
    } else {
        // Tgt was ref file
        let refImage = data.cache.getRefView().image;
        let imageMaximum = refImage.maximum();
        imageWindow.mainView.beginProcess(UndoFlag_NoSwapFile);
        averageWeight = 1.0;
        let weightStr = data.noiseWeightKeyword + ": 1.0";
        let scale = {m:1.0, b:0};
        let pedestals;
        let weights;
        if (imageWindow.mainView.image.isColor){
            scaleFactors = [scale, scale, scale];
            signalToNoise = [1, 1, 1];
            weights = [1, 1, 1];
            pedestals = [0, 0, 0];
        } else {
            scaleFactors = [scale];
            signalToNoise = [1];
            weights = [1];
            pedestals = [0];
        }
        console.writeln(weightStr);
        resultString += ", " + weightStr + ", SNR: 1.0";
        keywords.push(new FITSKeyword("HISTORY", "", "NSG.reference"));
        fitsHeader.fitsHeaderGradient(keywords, data);
        fitsHeader.fitsHeaderMaxValue(keywords, imageMaximum);
        fitsHeader.fitsHeaderScale(keywords, scaleFactors);
        fitsHeader.fitsHeaderSignalToNoise(keywords, signalToNoise);
        if (refNoiseRaw){
            let refNoiseValues = [];
            for (let i=0; i<refNoiseRaw.noise.length; i++){
                if (refNoiseRaw.noise[i]){
                    refNoiseValues.push(refNoiseRaw.noise[i]);
                }
            }
            fitsHeader.fitsHeaderNoise(keywords, refNoiseValues, "NOISExx");
        }
        let tmpArray = [];
        tmpArray[0] = refNoiseMRS[0];
        fitsHeader.fitsHeaderNoise(keywords, tmpArray, "noiseMRS");
        tmpArray[0] = refNoiseKSigma[0];
        fitsHeader.fitsHeaderNoise(keywords, tmpArray, "noiseKSigma");
        averageWeight = 1;
        fitsHeader.fitsHeaderNoiseWeight(keywords, data.noiseWeightKeyword, averageWeight, "");
        imageWindow.keywords = keywords;
        imageWindow.mainView.endProcess();
        let savedFile = saveFile(imageWindow, data, averageWeight, imageMaximum);
        savedFilename = savedFile.filename;
        savedFiles.push(savedFile);
        
        if (data.displayGradient){
            let gradientResample = 5;
            let w = Math.floor(refImage.width / gradientResample);
            let h = Math.floor(refImage.height / gradientResample);
            if (refImage.width % gradientResample){
                w++;
            }
            if (refImage.height % gradientResample){
                h++;
            }
            let colorSpace = data.cache.isColor() ? ColorSpace_RGB : ColorSpace_Gray;
            let gradientImage = new Image(w, h, nChannels, colorSpace);
            gradientImage.fill(0);
            saveGradientFile(gradientImage, savedFilename, "gradient", data, 
                    scaleFactors, weights, gradientResample, pedestals);
            gradientImage.free();
        }
    }
    
    gc(false);
    return new Result(resultString, savedFilename, averageWeight, scaleFactors, exposure, tgtFile, false);
}

function SavedFile(filename, dividedBy){
    this.filename = filename;
    this.dividedBy = dividedBy;
}

/**
 * @param {ImageWindow} imageWindow
 * @param {NsgData} data
 * @param {Number} averageWeight
 * @param {Number} imageMaximum
 * @returns {SavedFile} full filename, amount image was divided by to prevent truncation
 */
function saveFile(imageWindow, data, averageWeight, imageMaximum){
    let outputDir = data.outputDir;
    if ( !outputDir.endsWith( '/' ) ){
        outputDir += '/';
    }
    let tgtFilename = data.cache.getTgtFilename();
    let prefix;
    if (data.weightPrefix){
        let prefixPercent = Math.round(averageWeight * 100);
        if (prefixPercent < 10){
            prefix = "w00" + prefixPercent + "_";
        } else if (prefixPercent < 100){
            prefix = "w0" + prefixPercent + "_";
        } else {
            prefix = "w" + prefixPercent + "_";
        }
    } else {
        prefix = "";
    }
    let outputFile = outputDir + prefix + File.extractName(tgtFilename) + data.outputPostFix + File.extractSuffix(tgtFilename);
    if (File.exists(outputFile) && !data.overwrite){
        for ( let u = 1; ; ++u ){
            let tryFilePath = File.appendToName(outputFile, '_' + u.toString());
            if (!File.exists(tryFilePath)){
                outputFile = tryFilePath;
                break;
            }
        }
    }
    let image = imageWindow.mainView.image;
    let divideBy = data.rescale ? Math.max(imageMaximum, DIVIDE_BY) : 1.0;
    imageWindow.mainView.beginProcess(UndoFlag_NoSwapFile);
    if (data.rescale){
        image.apply( divideBy, ImageOp_Div );
    }
    image.truncate(); // remove any negative values
    imageWindow.mainView.endProcess();
    imageWindow.saveAs(outputFile, false/*queryOptions*/, false/*allowMessages*/,
              false/*strict*/, false/*verifyOverwrite*/, "no-warnings verbosity 0");
    return new SavedFile(outputFile, divideBy);
}

/**
 * Returned filename will start with prependText + colorLabel + "_"
 * If the filename is not unique and overwrite has not been specified, _N postfix is added.
 * The extension will be changed to .xisf
 * @param {String} filename
 * @param {String} prependText
 * @param {Boolean} overwrite
 * @returns {String} The new filename
 */
function prependFilename(filename, prependText, overwrite){
    let outputFile = File.prependToName( filename, prependText + "_" );
    outputFile = File.changeExtension( outputFile, ".xisf" );
    if (File.exists(outputFile) && !overwrite){
        for ( let u = 1; ; ++u ){
            let tryFilePath = File.appendToName(outputFile, '_' + u.toString());
            if (!File.exists(tryFilePath)){
                outputFile = tryFilePath;
                break;
            }
        }
    }
    return outputFile;
}

/**
 * Saves image as compressed .xisf file, with FITS headers for ref image and gradient smoothing.
 * Filename will start with prefix + "_"
 * If the filename is not unique and overwrite has not been specified, _N postfix is added.
 * The extension will be changed to .xisf
 * @param {Image} image
 * @param {String} filename
 * @param {String} prefix
 * @param {NsgData} data
 * @param {LinearFitData[]} scaleFactors
 * @param {Number[]} weights Weight for current channel
 * @param {Number} resample Gradient image needs to be resized by this factor
 * @param {Number[]} pedestals Added to gradient image to keep all pixels positive
 * @returns {String} The filename of the saved file
 */
function saveGradientFile(image, filename, prefix, data, scaleFactors, weights, resample, pedestals){
    let outputFile = prependFilename(filename, prefix, data.overwrite);
    let isColor = data.cache.getRefView().image.isColor;
    let nChannels = isColor ? 3 : 1;
    let gradWindow = new ImageWindow(1, 1, nChannels, 32, true, isColor);
    let weightKeyword = data.noiseWeightKeyword;
    gradWindow.mainView.beginProcess(UndoFlag_NoSwapFile);
    let testView = gradWindow.mainView;
    testView.image.assign(image);
    let keywords = gradWindow.keywords;
    let fitsHdr = new FitsHeader();
    fitsHdr.fitsHeaderGradient(keywords, data);
    fitsHdr.fitsHeaderDimension(keywords, data.cache.getRefView().image);
    keywords.push(new FITSKeyword("NSGPSTFX", data.outputPostFix, "NSG.postfix"));
    keywords.push(new FITSKeyword("NSGRESIZ", resample.toString(), "NSG.resize"));
    keywords.push(new FITSKeyword("NSGWKEY", weightKeyword, "NSG.weight keyword"));
    if (isColor && weightKeyword.length === 8){
        weightKeyword = weightKeyword.slice(0, 7); // The eight character will become 0, 1 or 2
    }
    for (let c=0; c<nChannels; c++){
        keywords.push(new FITSKeyword("NSGPLUS" + c, pedestals[c].toPrecision(9), "NSG.pedestal"));
        let inverse = 1 / scaleFactors[c].m;
        keywords.push(new FITSKeyword("NSGS" + c, inverse.toPrecision(5), "NSG.scale"));
        let wKey = isColor ? weightKeyword + c : weightKeyword;
        let weight = weights[c] ? weights[c] : 0;
        keywords.push(new FITSKeyword(wKey, weight.toPrecision(5), "NSG.weight"));
    }
    gradWindow.keywords = keywords;
    testView.endProcess();
    gradWindow.saveAs(outputFile, false/*queryOptions*/, false/*allowMessages*/,
              false/*strict*/, false/*verifyOverwrite*/, "compression-codec zlib");
    gradWindow.purge();
    gradWindow.forceClose();
    return outputFile;
}

/**
 * @param {Number} duration
 * @returns {String}
 */
function msToTime(duration) {
    let milliseconds = duration % 1000,
    seconds = Math.floor((duration / 1000) % 60),
    minutes = Math.floor((duration / (1000 * 60)) % 60),
    hours = Math.floor((duration / (1000 * 60 * 60)) % 24);
    if (hours > 0){
        return hours + " hours, " + minutes + " minutes";
    }
    if (minutes > 0){
        return minutes + " minutes " + seconds + " seconds";
    }
    if (seconds > 0){
        return seconds + " s";
    }
    return milliseconds + " ms";
}

/**
 * @param {Number} startTime
 * @returns {String}
 */
function getElapsedTime(startTime) {
    return msToTime(new Date().getTime() - startTime);  
}

/**
 * @param {Number[]} noise
 * @param {String[]} type
 * @param {Number} index Index to the defined noise entry, or -1 if RGB
 */
function Noise(noise, type, index){
    this.noise = noise;
    this.type = type;
    this.index = index;
    
    /**
     * @param {Noise} refNoise
     * @param {Noise} tgtNoise
     * @param {Number} index
     * @param {Number} scaleFactor
     * @returns {undefined|Number}
     */
    this.getSignalToNoise = function(refNoise, tgtNoise, index, scaleFactor){
        if (refNoise && tgtNoise && 
                refNoise.noise[index] && tgtNoise.noise[index] &&
                refNoise.type[index] === tgtNoise.type[index]){
            return refNoise.noise[index] / (tgtNoise.noise[index] * scaleFactor);
        } 
        return undefined;
    };
    /**
     * @param {Number} signalToNoise
     * @returns {undefined|Number}
     */
    this.getWeight = function(signalToNoise){
        if (signalToNoise){
            return Math.pow( signalToNoise, 2 );
        } 
        return undefined;
    };
}

/**
 * @param {String} resultString
 * @param {String} savedFilename
 * @param {Number} averageWeight
 * @param {Number} scaleFactors
 * @param {Number} exposure
 * @param {String} inputFilename 
 * @param {Boolean} failed 
 * @returns {Result}
 */
function Result(resultString, savedFilename, averageWeight, scaleFactors, exposure, inputFilename, failed){
    this.summary = resultString;
    this.file = savedFilename;
    this.weight = averageWeight;
    this.scaleFactors = scaleFactors;
    this.exposure = exposure;
    this.inputFile = inputFilename;
    this.failed = failed;
}

/**
 * Sort in ascending filename order (NSG input filename, full path)
 * @param {Result} a
 * @param {Result} b
 * @returns {Number}
 */
function compareResultFilename(a, b){
    if (a.inputFile < b.inputFile){
        return -1;
    }
    if (a.inputFile > b.inputFile){
        return 1;
    }
    return 0;
}

/**
 * Sort in descending weight order
 * @param {Result} a
 * @param {Result} b
 * @returns {Number}
 */
function compareResultWeight(a, b){
    return b.weight - a.weight;
}

function main() {
    // Create dialog, start looping
    let nsgDialog;
    let data = new NsgData();

    if (Parameters.isViewTarget) {
        // Perform the script on the target view (no user interface)
//        console.show();
//        data.loadParameters();
//        console.writeln("\n<b>", TITLE," ", VERSION, "</b>:");
//        // Process the target image.
//        normalizeScaleGradient(data, Parameters.targetView); 
//        return;
        data.loadParameters();  // temp solution. Need to implement view target instead of file target
    } else if (Parameters.isGlobalTarget) {
        data.loadParameters();
    } else {
        data.restoreSettings();
    }
    let exception;
    let maxWeightResult;
    let hasExposure = true;
    let referenceResult;
    let results = [];
    nsgDialog = new NsgDialog(data);
    data.setNsgDialog(nsgDialog);
    data.setPhotometryAutoValues(data.useAutoPhotometry, false);
    data.setSampleGenerationAutoValues(data.useAutoSampleGeneration, false);
    
    for (; ; ) {
        if (!nsgDialog.execute()){
            // Dialog cancelled.
            break;
        }
        console.show();
        // User must select a reference view and one or more targets
        if (!data.cache.getRefFilename()){
            (new MessageBox("WARNING: A reference file must be specified.", TITLE, StdIcon_Error, StdButton_Ok)).execute();
            continue;
        }
        if (!data.targetFiles.length) {
            (new MessageBox("WARNING: At least one target file must be specified.", TITLE, StdIcon_Error, StdButton_Ok)).execute();
            continue;
        }
        if (!data.outputDir || data.outputDir.trim().length < 2 || !File.directoryExists( data.outputDir )){
            new MessageBox("Please specify a valid output directory.", TITLE, StdIcon_Error, StdButton_Ok).execute();
            continue;
        }
//        if (data.gradientSmoothness < 0 && !data.manualRejectionCircles.length){
//            let msg = "<p>Warning: Low levels of gradient smoothness may result in artifacts.<p>" +
//                    "<p>To remove light pollution gradients, gradient smoothness should be set to " +
//                    "between 0.0 and 4.0</p>" +
//                    "<p>Lower levels of smoothing can be used to correct complex gradients but " +
//                    "check the normalized files and the 'curvature' files carefully.</p>" +
//                    "<p>See the <b>Quick Start Guide</b> and <b>Common Problems</b> sections within the documentation.</p>" +
//                    "<p>Select 'Ignore' to continue or 'Cancel' to return to the main dialog.</p>";
//            let messageBox = new MessageBox(msg, TITLE, StdIcon_Warning, StdButton_Ignore, StdButton_Cancel);
//            let response = messageBox.execute();
//            if (response === StdButton_Cancel){
//                continue;
//            }
//        }

        // Run the script
        try {
            let startTime = new Date().getTime();
            data.saveParameters();  // Save script parameters to the history.
            let refFilename = data.cache.getRefFilename();
            let fitsHeader = new FitsHeader();
            let keywords = data.cache.getRefImageWindow().keywords;
            console.noteln("\n", TITLE, " V", VERSION);
            console.noteln("Ref: '", refFilename, "'");
            let nFiles = data.targetFiles.length;
            let nNormalized = 0;
            let displayRemainTime = false;
            let refImage = data.cache.getRefView().image;
            let refFilter = fitsHeader.getFilter(keywords);
            let refNoiseRaw = fitsHeader.getNoise(keywords);
            let refNoiseMRS = refImage.noiseMRS();
            let refNoiseKSigma = refImage.noiseKSigma();
            let hasNewError = false;
            let aborted = false;
            let warnings = [];
            let errors = [];
            results = [];
            referenceResult = undefined;
            let savedFiles = [];
            for (let i=0; i<nFiles; i++){
                if (hasNewError){
                    hasNewError = false;
                    if (data.onErrorIndex === ASK_USER){
                        let dialog = new MessageBox(errors[errors.length-1], TITLE, StdIcon_Error, StdButton_Abort, StdButton_Ignore);
                        let buttonCode = dialog.execute();
                        aborted = buttonCode === StdButton_Abort;
                    }
                    if (aborted || data.onErrorIndex === ABORT){
                        aborted = true;
                        referenceResult = undefined;
                        results = [];
                        savedFiles = [];
                        break;
                    }
                }
                let targetFile = data.targetFiles[i];
                let tgtTime = new Date().getTime();
                if (console.abortRequested){
                    errors.push("Aborted ...");
                    break;
                }
                data.cache.setTgtFilename(targetFile);
                let tgtView = data.cache.getTgtView();
                if (tgtView.isNull){
                    hasNewError = true;
                    errors.push("ERROR [" + File.extractName(targetFile) + "] FAILED to load image.");
                    continue;
                }
                let tgtImage = tgtView.image;
                if (tgtImage.isColor !== refImage.isColor) {
                    hasNewError = true;
                    console.criticalln("ERROR [", targetFile, "] Color depth must match the reference image. Skipping target.");
                    errors.push("ERROR [" + File.extractName(targetFile) + "] Color depth must match.");
                    continue;
                }
                if (tgtImage.width !== refImage.width || tgtImage.height !== refImage.height) {
                    hasNewError = true;
                    console.criticalln("ERROR [", targetFile, "] Image dimensions must match the reference image. Skipping target.");
                    errors.push("ERROR [" + File.extractName(targetFile) + "] Image dimensions must match.");
                    continue;
                }
                let tgtKeywords = data.cache.getTgtImageWindow().keywords;
                let tgtFilter = fitsHeader.getFilter(tgtKeywords);
                if (tgtFilter !== refFilter){
                    let warning = "WARNING Filter mismatch: " + File.extractName(targetFile) + 
                            ": Reference '" + refFilter + "' Target '" + tgtFilter + "'";
                    console.warningln(warning);
                    warnings.push(warning);
                }
                if (refNoiseRaw){
                    let tgtNoiseRaw = fitsHeader.getNoise(tgtKeywords);
                    if (!tgtNoiseRaw){
                        // Ref image has NOISExx, but target image does not.
                        let warning = "WARNING: No 'NOISExx' in FITS header: " + File.extractName(targetFile);
                        console.warningln(warning);
                        warnings.push(warning);
                    }
                }
                let result = normalizeScaleGradient(data, targetFile, refNoiseRaw, refNoiseMRS, refNoiseKSigma, savedFiles);
                if (result && !result.failed){
                    nNormalized++;
                    if (targetFile !== data.cache.getRefFilename()){
                        results.push(result);
                        console.noteln("Normalized '", File.extractName(targetFile), 
                                "' [", i + 1, "/", nFiles, "] ", getElapsedTime(tgtTime));
                        let nRemain = nFiles - (i + 1);
                        if (displayRemainTime && nRemain){
                            let ms = new Date().getTime() - tgtTime;
                            console.noteln("Time remaining: ", msToTime(nRemain * ms), "\n");
                        } else {
                            console.writeln();
                        }
                        displayRemainTime = true;   // Accurate from second target
                    } else {
                        referenceResult = result;
                        console.noteln("Copied reference '", data.cache.getTgtView().fullId, 
                                "' [", i + 1, "/", nFiles, "] ", getElapsedTime(tgtTime), "\n");
                    }
                } else {
                    if (result && result.failed){
                        errors.push(result.summary);
                    } else {
                        errors.push("ERROR [" + File.extractName(targetFile) + "] FAILED");
                    }
                    hasNewError = true;
                    console.criticalln("Failed to normalize '", data.cache.getTgtView().fullId, 
                            "' [", i + 1, "/", nFiles, "] ", "\n");
                }
            }
            if (hasNewError){
                hasNewError = false;
                if (data.onErrorIndex === ASK_USER){
                    let dialog = new MessageBox(errors[errors.length-1], TITLE, StdIcon_Error, StdButton_Abort, StdButton_Ignore);
                    let buttonCode = dialog.execute();
                    aborted = buttonCode === StdButton_Abort;
                }
                if (aborted || data.onErrorIndex === ABORT){
                    aborted = true;
                    referenceResult = undefined;
                    results = [];
                    savedFiles = [];
                }
            }
            
            if (!aborted && data.rescale){
                // Rescale files. This is done to prevent truncation.
                let maxDividedBy = 0;
                for (let i=0; i<savedFiles.length; i++){
                    maxDividedBy = Math.max(savedFiles[i].dividedBy, maxDividedBy);
                }
                if (maxDividedBy > DIVIDE_BY){
                    for (let i=0; i<savedFiles.length; i++){
                        let filename = File.extractName(savedFiles[i].filename);
                        if (savedFiles[i].dividedBy !== maxDividedBy){
                            let imageWindows = ImageWindow.open(savedFiles[i].filename, "division", "no-warnings verbosity 0");
                            let imageWindow = imageWindows[0];
                            imageWindow.mainView.beginProcess(UndoFlag_NoSwapFile);
                            let image = imageWindow.mainView.image;
                            let dividedBy = savedFiles[i].dividedBy;
                            let rescale = maxDividedBy / dividedBy;
                            image.apply( rescale, ImageOp_Div );
                            imageWindow.mainView.endProcess();
                            imageWindow.saveAs(savedFiles[i].filename, false/*queryOptions*/, false/*allowMessages*/,
                                false/*strict*/, false/*verifyOverwrite*/, "no-warnings verbosity 0");
                            imageWindow.purge();
                            imageWindow.forceClose();
                            console.noteln(filename, ", Rescaled by ", dividedBy.toPrecision(5), " x ", rescale.toPrecision(5),
                                " = ", (dividedBy * rescale).toPrecision(5));
                        } else {
                            console.noteln(filename, ", Rescaled by ", maxDividedBy.toPrecision(5));
                        }
                    }
                }
            }
            
            if (data.sortByWeight){
                results.sort(compareResultWeight);
            } else {
                results.sort(compareResultFilename);
            }
            console.noteln("Normalized [", nNormalized, "/", nFiles, "] total time ", getElapsedTime(startTime));
            console.writeln("\n<b><u>Summary</u></b>");
            if (refNoiseRaw){
                console.write("Using noise estimates from FITS header:");
                if (refNoiseRaw.index === -1){
                    for (let c=0; c<refNoiseRaw.noise.length; c++){
                        console.write(" NOISE0", c);
                    }
                } else {
                    let noiseStr = "NOISE0" + refNoiseRaw.index;
                    console.write(" ", noiseStr);
                    if (refImage.isColor){
                        console.warningln("WARNING: Only found " + noiseStr + " in FITS header. " +
                            "Noise estimated from '" + noiseStr + "'. " +
                            "Noise estimates Noise00, Noise01 and Noise02 can be generated by WBPP or Debayer.");
                    }
                }
                console.writeln();
            } else {
                // No NOISExx headers
                console.warningln("WARNING: can't find noise estimate 'NOISExx' in FITS header. " +
                        "Noise estimated from registered images (less accurate). " +
                        "Noise estimates can be generated by WBPP, ImageCalibration or Debayer.");
            }
            let nth = 1;
            let maxResultNth;
            if (referenceResult){
                maxResultNth = nth;
                maxWeightResult = referenceResult;
                console.noteln("[",nth++,"], ", referenceResult.summary, ", Reference");
            }
            for (let i=0; i<results.length; i++){
                if (!maxWeightResult || results[i].weight > maxWeightResult.weight){
                    maxResultNth = nth;
                    maxWeightResult = results[i];
                }
                if (!results[i].exposure){
                    hasExposure = false;
                }
                console.writeln("[",nth++,"], ", results[i].summary);
            }
            for (let i=0; i<warnings.length; i++){
                console.warningln(warnings[i]);
            }
            for (let i=0; i<errors.length; i++){
                console.criticalln("[",nth++,"] ", errors[i]);
            }
            if (!hasExposure){
                console.warningln("Unknown exposure time for 1 or more images. " +
                        "Minimum weight test will assume all exposures are equal");
            }
            if (maxWeightResult){
                console.writeln("Max weight ", maxWeightResult.weight.toPrecision(5), ", image[", maxResultNth, "] ", File.extractName(maxWeightResult.file));
                let minimumWeight = maxWeightResult.weight * data.minimumWeight / 100;
                console.writeln("ImageIntegration: Enabling images with weights > ", 
                        maxWeightResult.weight.toPrecision(5), " x ", data.minimumWeight, "% = ", minimumWeight.toPrecision(4));
            }
            if (aborted){
                console.criticalln("Aborted ...");
            }
            if (results.length && data.autoExit){
                break;
            }
        } catch (e){
            exception = e;
            new MessageBox("" + e, TITLE, StdIcon_Error, StdButton_Ok).execute();
            break;
        }
        if (console.abortRequested){
            results = [];
            break;
        }
    }
    if (!exception){
        data.saveSettings(data);
    }
    
    if (data.cache !== undefined){
        data.cache.invalidate();
    }
    if (!exception){
        console.hide();
    } else {
        throw exception;
    }

    /**
     * Used to add images to ImageIntegration
     * @param {{enabled, path, drizzlePath, localNormalizationDataPath}[]} images
     * @param {summary: resultString, file: savedFilename, weight: weight, exposure: exposure} result
     * @param {Number} minimumWeight
     * @param {summary: resultString, file: savedFilename, weight: weight, exposure: exposure} maxWeightResult 
     * @param {Boolean} hasExposure
     * @return {Boolean} True if the added image will be enabled
     */
    function addImage(images, result, minimumWeight, maxWeightResult, hasExposure){
        let enabled;
        if (hasExposure){
            let exposureRatio = maxWeightResult.exposure / result.exposure;
            enabled = result.weight * exposureRatio > minimumWeight;
        } else {
            enabled = result.weight > minimumWeight;
        }
        // enabled, path, drizzlePath, localNormalizationDataPath
        images.push([enabled, result.file, "", ""]);
        return enabled;
    }
            
    if (exception){
        throw exception;
    } else {
        console.hide();
        if (data.imageIntegration && results.length > 0){
            if (data.sortByWeight){
                results.sort(compareResultWeight);
            } else {
                results.sort(compareResultFilename);
            }
            let minimumWeight = maxWeightResult.weight * data.minimumWeight / 100;
            let images = [];
            let nImages = 0;
            if (referenceResult){
                let enabled = addImage(images, referenceResult, minimumWeight, maxWeightResult, hasExposure);
                if (enabled) nImages++;
            }
            for (let i=0; i<results.length; i++){
                let enabled = addImage(images, results[i], minimumWeight, maxWeightResult, hasExposure);
                if (enabled) nImages++;
            }
            let P = null;
            if (data.imageIntegrationTemplateId !== ""){
                P = ProcessInstance.fromIcon(data.imageIntegrationTemplateId);
            }
            if (P===null || P.processId() !== "ImageIntegration") {
                P = new ImageIntegration;
                // User has not specified an ImageIntegration process icon.
                // Set rejection algorithm based on the number of images.
                if ( images.length === 2 ){
                    P.rejection = ImageIntegration.prototype.NoRejection;
                    // ImageIntegration requires at least 3 images. Add images twice.
                    images.push(images[0]);
                    images.push(images[1]);
                }
                else if ( nImages < 8 ){
                    P.rejection = ImageIntegration.prototype.PercentileClip;
                    P.pcClipLow = 0.2;
                    P.pcClipHigh = 0.1;
                } else if ( nImages <= 10 ){
                    P.rejection =  ImageIntegration.prototype.SigmaClip;
                    P.sigmaLow = 3.5;
                    P.sigmaHigh = 2.5;
                } else if ( nImages < 20 ){
                    P.rejection =  ImageIntegration.prototype.WinsorizedSigmaClip;
                    P.sigmaLow = 3.5;
                    P.sigmaHigh = 2.2;
                    P.winsorizationCutoff = 3.5;
                } else {
                    P.rejection = ImageIntegration.prototype.Rejection_ESD;
//                    P.esdOutliersFraction = 3.2 / nImages;
//                    P.esdAlpha = 0.05;
//                    P.esdLowRelaxation = 1.50;
                }
            }
            P.images = images;
            P.combination = ImageIntegration.prototype.Average;
            P.weightMode = ImageIntegration.prototype.KeywordWeight;
            P.weightKeyword = data.noiseWeightKeyword;
            P.normalization = ImageIntegration.prototype.NoNormalization;
            P.rejectionNormalization = ImageIntegration.prototype.NoRejectionNormalization;
            P.generateDrizzleData = false;
            P.launchInterface();
        }
    }
    return;
}

main();
