/* global ImageWindow, UndoFlag_NoSwapFile, LINEAR_RANGE, STAR_BKG_DELTA, File, MAX_TGT_STARS */

// Version 1.0 (c) John Murphy 7th-Apr-2021
//
// ======== #license ===============================================================
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, version 3 of the License.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.
// =================================================================================
//"use strict";

function Cache(){
    /**
     * ImageCache stores filename, imageWindow, detected stars, image bitmap
     * @returns {Cache.ImageCache}
     */
    function ImageCache(){
        let filename;
        let imageWindow;
        let starDetectionMask;
        /** Star detection sensitivity */
        let logStarSensitivity;
        let detectedStars;
        let bitmap;
        let self = this;

        /**
         * @param {String} fullPath
         * @return {Boolean} True if image exists and could be read
         */
        this.setFilename = function(fullPath){
            if (!fullPath){
                self.invalidate();
                return false;
            }
            if (!File.exists(fullPath)){
                console.criticalln("File does not exist: '", fullPath, "'");
                self.invalidate();
                return false;
            }
            if (fullPath === filename){
                return true;
            }
            self.invalidate();
            filename = fullPath;
            try {
                // Check the file header can be read
                getFITSKeywords( fullPath );
            } catch (exception){
                console.criticalln("Failed to read '", fullPath, "': ", exception);
                self.invalidate();
                return false;
            }
            return true;
        };

        /**
         * Purge and close image window if it exists
         */
        this.invalidate = function(){
            filename = undefined;
            if (imageWindow){
                imageWindow.purge();
                imageWindow.forceClose();
                imageWindow = undefined;
            }
            if (starDetectionMask){
                starDetectionMask.free();
                starDetectionMask = undefined;
            }
            logStarSensitivity = undefined;
            if (detectedStars){
                detectedStars.clear();
                detectedStars = undefined;
            }
            if (bitmap){
                bitmap.clear();
                bitmap = undefined;
            }
        };

        /**
         * @returns {fullPath}
         */
        this.getFilename = function(){
            return filename;
        };

        /**
         * @returns {ImageWindow|undefined}
         */
        this.getImageWindow = function(){
            if (!filename){
                return undefined;
            }
            if (!imageWindow){
                let imageWindows = ImageWindow.open(filename, "tmp", "no-warnings verbosity 0");
                imageWindow = imageWindows[0];
                if (!imageWindow.isFloatSample){
                    let keywords = imageWindow.keywords;
                    let view = imageWindow.mainView;
                    let image = view.image;
                    let nChannels = image.numberOfChannels;
                    let floatImageWindow = 
                        new ImageWindow(1, 1, nChannels, 32, true, nChannels > 1, view.fullId);
                    floatImageWindow.mainView.beginProcess(UndoFlag_NoSwapFile);
                    floatImageWindow.mainView.image.assign(image);
                    floatImageWindow.keywords = keywords;
                    floatImageWindow.mainView.endProcess();
                    imageWindow.purge();
                    imageWindow.forceClose();
                    imageWindow = floatImageWindow;
                }
            }
            return imageWindow;
        };

        /**
         * @returns {ImageWindow.mainView|undefined}
         */
        this.getView = function(){
            if (self.getImageWindow()){
                return imageWindow.mainView;
            }
            return new View();
        };
        
        /**
         * @returns {Bitmap}
         */
        this.getImageBitmap = function(){
            if (!bitmap){
                let img = self.getView().image;
                // Width, height, n channels, bitsPerSample, float, color, title
                let nChannels = img.isColor ? 3 : 1;
                let w = new ImageWindow(1, 1, nChannels, 16, false, img.isColor, "createBitmap");
                let view = w.mainView;
                view.beginProcess(UndoFlag_NoSwapFile);
                view.image.assign(img);
                view.endProcess();

                // Apply a Histogram Transformation based on the reference view's STF
                // before converting this temporary view into a bitmap
                let stf = STFAutoStretch(view, undefined, undefined, false);
                var HT = new HistogramTransformation;
                if (img.isColor){
                    HT.H = 
                        [[stf[0][0], stf[0][2], stf[0][1], 0, 1],
                        [stf[1][0], stf[1][2], stf[1][1], 0, 1],
                        [stf[2][0], stf[2][2], stf[2][1], 0, 1],
                        [0, 0.5, 1, 0, 1],
                        [0, 0.5, 1, 0, 1]];
                } else {
                    HT.H = 
                        [[0, 0.5, 1, 0, 1],
                        [0, 0.5, 1, 0, 1],
                        [0, 0.5, 1, 0, 1],
                        [stf[0][0], stf[0][2], stf[0][1], 0, 1],
                        [0, 0.5, 1, 0, 1]];
                }
                HT.executeOn(view, false); // no swap file

                bitmap = view.image.render();
                w.purge();
                w.close();
            }
            return bitmap;
        };

        /**
         * @param {Number} logSensitivity
         * @returns {DetectedStars}
         */
        this.hasDetectedStars = function(logSensitivity){
            return logStarSensitivity === logSensitivity;
        };

        /**
         * @param {Number} logSensitivity
         * @param {Number} marginX Detect stars within a limited area at image center
         * @param {Number} marginY Detect stars within a limited area at image center
         * @returns {DetectedStars}
         */
        function getDetectedStars (logSensitivity, marginX, marginY){
            let progressLabel = self.getView().id;
            if (logStarSensitivity !== logSensitivity){
                logStarSensitivity = logSensitivity;
                detectedStars = undefined;
            }
            if (!detectedStars){
                detectedStars = new DetectedStars(self.getView(), progressLabel, logStarSensitivity, marginX, marginY);
            }
            return detectedStars;
        };

        /**
         * @param {Number} logSensitivity
         * @param {Number} marginX Detect stars within a limited area at image center
         * @param {Number} marginY Detect stars within a limited area at image center
         * @returns {Star[]} Detected stars
         */
        this.getStars = function(logSensitivity, marginX, marginY){
            getDetectedStars(logSensitivity, marginX, marginY);
            return detectedStars.getStars();
        };
        
        /**
         * @param {Number} logSensitivity
         * @returns {Boolean} True if the stars have already been detected
         */
        this.hasCachedStars = function(logSensitivity){
            return (logStarSensitivity === logSensitivity) && (detectedStars !== undefined);
        };
        
        /**
         * @param {Number} logSensitivity
         * @returns {Number} Maximum star flux in image
         */
        this.getMaxStarFlux = function(logSensitivity){
            if (!detectedStars){
                getDetectedStars(logSensitivity, 0, 0).getStars();
            }
            return detectedStars.getMaxStarFlux();
        };
        
        /**
         * @returns {Number} LINEAR_RANGE * max pixel value in image to 3 decimal places
         */
        this.getLinearRange = function(){
            if (filename){
                let image = self.getView().image;
                return Math.round(1000 * image.maximum() * LINEAR_RANGE) / 1000;
            }
            return LINEAR_RANGE;
        };
        
    }
    
    // -------------------
    // REF Image Cache
    // -------------------
    let refImageCache = new ImageCache();
    
    /**
     * @param {String} fullPath
     * @return {Boolean} True if image exists and could be read
     */
    this.setRefFilename = function(fullPath){
        return refImageCache.setFilename(fullPath);
    };
    
    /**
     * @returns {String} fullPath
     */
    this.getRefFilename = function(){
        return refImageCache.getFilename();
    };
    
    /**
     * @param {Number} logSensitivity
     * @returns {Boolean} True if ref stars have already been detected
     */
    this.hasCachedRefStars = function(logSensitivity){
        return refImageCache.hasCachedStars(logSensitivity);
    };
    
    /**
     * @returns {ImageWindow|undefined}
     */
    this.getRefImageWindow = function(){
        return refImageCache.getImageWindow();
    };
    
    /**
     * @returns {ImageWindow.mainView|undefined}
     */
    this.getRefView = function(){
        return refImageCache.getView();
    };
    
    /**
     * @returns {Bitmap}
     */
    this.getRefImageBitmap = function(){
        return refImageCache.getImageBitmap();
    };
    
    /**
     * @param {Number} logSensitivity
     * @returns {Star[]} Detected stars
     */
    this.getRefStars = function(logSensitivity){
        return refImageCache.getStars(logSensitivity, 0, 0);
    };
    
    /**
     * @returns {Number} Calculated max safe linear value in ref image
     */
    this.getLinearRangeRef = function(){
        return refImageCache.getLinearRange();
    };
    
    /**
     * @param {Number} logSensitivity
     * @returns {Number|undefined} flux of the brightest star.
     */
    this.getMaxStarFlux = function (logSensitivity){
        return refImageCache.getMaxStarFlux(logSensitivity);
    };
    
    // -------------------
    // TGT Image Cache
    // -------------------
    let tgtImageCache = new ImageCache();
    
    /**
     * @param {String} fullPath
     * @return {Boolean} True if image exists and could be read
     */
    this.setTgtFilename = function(fullPath){
        return tgtImageCache.setFilename(fullPath);
    };
    
    /**
     * @returns {String} fullPath
     */
    this.getTgtFilename = function(){
        return tgtImageCache.getFilename();
    };
    
    /**
     * @returns {ImageWindow|undefined}
     */
    this.getTgtImageWindow = function(){
        return tgtImageCache.getImageWindow();
    };
    
    /**
     * @returns {ImageWindow.mainView|undefined}
     */
    this.getTgtView = function(){
        return tgtImageCache.getView();
    };
    
    /**
     * @returns {Bitmap}
     */
    this.getTgtImageBitmap = function(){
        return tgtImageCache.getImageBitmap();
    };
    
    /**
     * @param {NsgData} data
     * @param {Number} logSensitivity
     * @returns {Star[]} Detected stars
     */
    this.getTgtStars = function(data, logSensitivity){
        if (!tgtImageCache.hasDetectedStars(logSensitivity)){
            // Calculate size of central rectangle that contains roughly MAX_TGT_STARS reference stars
            let stars = refImageCache.getStars(logSensitivity, 0, 0);
            let areaFraction = stars.length <= MAX_TGT_STARS ? 1 : MAX_TGT_STARS/stars.length;
            if (areaFraction !== 1){
                let view = tgtImageCache.getView();
                let image = view.image;
                let marginFraction = (1 - Math.sqrt(areaFraction))/2;
                let marginX = Math.round(image.width * marginFraction);
                let marginY = Math.round(image.height * marginFraction);
                let mX, mY;
                for (let i = 100; i >= 0; i--){
                    mX = Math.round(marginX * i / 100);
                    mY = Math.round(marginY * i / 100);
                    let nStars = photometryCache.countStarsInRect(data, new Rect(mX, mY, 
                        image.width - mX, image.height - mY));
                    if (nStars >= MAX_TGT_STARS){
                        break;
                    }
                }
                return tgtImageCache.getStars(logSensitivity, mX, mY);
            }
        }
        return tgtImageCache.getStars(logSensitivity, 0, 0);
    };
    
    // --------------------
    // Photometry Cache
    // --------------------
    function PhotometryCache(){
        let refStarQuadTree_;
        let starMatchArray_;

        let refFilename_;
        let tgtFilename_;
        let logStarDetection_;
        let starSearchRadius_;
        let linearRange_;
        let fluxTolerance_;
        
        /**
         * @param {NsgData} data
         * @returns {BRQuadTree}
         */
        function getRefStarQuadTree(data){
            if (data.cache.getRefFilename() !== refFilename_ ||
                    data.logStarDetection !== logStarDetection_ ||
                    data.starSearchRadius !== starSearchRadius_ || 
                    data.linearRangeRef !== linearRange_){
                if (refStarQuadTree_){
                    refStarQuadTree_.clear();
                    refStarQuadTree_ = undefined;
                }
            }
            if (!refStarQuadTree_){
                refFilename_ = data.cache.getRefFilename();
                logStarDetection_ = data.logStarDetection;
                starSearchRadius_ = data.starSearchRadius;
                linearRange_ = data.linearRangeRef;
                
                let image = data.cache.getRefView().image;
                let imageRect = new Rect(image.width, image.height);
                let refStars = data.cache.getRefStars(logStarDetection_);
                refStarQuadTree_ = createQuadTree(refStars, starSearchRadius_, linearRange_, imageRect);
            }
            return refStarQuadTree_;
        };
        
        /**
         * @param {NsgData} data
         * @param {Rect} rect
         * @returns {Number}
         */
        this.countStarsInRect = function (data, rect){
            let quadTree = getRefStarQuadTree(data);
            let indexs = quadTree.search(rect);
            return indexs.length;
        };
        
        /**
         * @param {NsgData} data
         * @returns {StarMatch[]} StarMatch array, or undefined if ref or tgt not set
         */
        this.createStarMatchArray = function (data){
            if (data.cache.getRefFilename() === undefined || data.cache.getTgtFilename() === undefined){
                return undefined;
            }
            if (data.cache.getRefFilename() !== refFilename_ ||
                    data.cache.getTgtFilename() !== tgtFilename_ ||
                    data.logStarDetection !== logStarDetection_ ||
                    data.starSearchRadius !== starSearchRadius_ || 
                    data.linearRangeRef !== linearRange_ ||
                    data.starFluxTolerance !== fluxTolerance_){
                starMatchArray_ = undefined;
            }
            if (!starMatchArray_){
                tgtFilename_ = data.cache.getTgtFilename();
                let refQuadTree = getRefStarQuadTree(data);
                let tgtStars = data.cache.getTgtStars(data, logStarDetection_);
                starMatchArray_ = calcStarMatchArray(refQuadTree, tgtStars, starSearchRadius_, fluxTolerance_, linearRange_);
            }
            return starMatchArray_; 
        };
        
        /**
         * 
         * Call this method after the target image has been modified (e.g. scale and/or gradient subtraction)
         */
        this.clearTargetData = function(){
            starMatchArray_ = undefined;
            tgtFilename_ = undefined;
        };
        
        /**
         * Clears refStarQuadTree memory, and resets all module variables to undefined
         */
        this.invalidate = function(){
            if (refStarQuadTree_){
                refStarQuadTree_.clear();
                refStarQuadTree_ = undefined;
            }
            starMatchArray_ = undefined;
            refFilename_ = undefined;
            tgtFilename_ = undefined;
            logStarDetection_ = undefined;
            starSearchRadius_ = undefined;
            linearRange_ = undefined;
            fluxTolerance_ = undefined;
        };
    }
    
    let photometryCache = new PhotometryCache();
    
    /**
     * @param {NsgData} data
     * @returns {StarMatch[]}
     */
    this.getStarMatchArray = function(data){
        return photometryCache.createStarMatchArray(data);
    };
    
    // ---------------------
    // SampleGrid cache
    // ---------------------
    function SampleGridCache(){
        let refFilename_;
        let sampleSize_;
        let sampleGrid_;
        
        /**
         * The previously cached value is replaced if the refFile or sample size has changed
         * @param {NsgData} data
         * @returns {SampleGrid}
         */
        this.getSampleGrid = function (data){
            if (data.sampleSize !== sampleSize_ ||
                    data.cache.getRefFilename() !== refFilename_){
                sampleGrid_ = undefined;
            }
            if (!sampleGrid_){
                console.writeln("\n<b><u>Creating sample grid</u></b>");
                refFilename_ = data.cache.getRefFilename();
                sampleSize_ = data.sampleSize;
                console.writeln("Setting sample grid (Ref = ", refFilename_, ")");
                sampleGrid_ = new SampleGrid(data);
            }
            return sampleGrid_;
        };
        
        this.clearTargetData = function(){
            if (sampleGrid_){
                sampleGrid_.setTarget(undefined);
            }
        };
        
        /**
         * Reset all module variables to undefined
         */
        this.invalidate = function(){
            refFilename_ = undefined;
            sampleSize_ = undefined;
            sampleGrid_ = undefined;
        };
    }
    let sampleGridCache = new SampleGridCache();
    
    /**
     * 
     * @param {NsgData} data
     * @param {String} targetFilename Full file pathname
     * @returns {SampleGrid}
     */
    this.getSampleGrid = function(data, targetFilename){
        if (data.cache.setTgtFilename(targetFilename)){
            let sampleGrid = sampleGridCache.getSampleGrid(data);
            sampleGrid.setTarget(targetFilename);
            return sampleGrid;
        }
        return undefined;
    };
    
    /**
     * Return the number of bins in the sample grid. Returns undefined if there is no reference.
     * @param {NsgData} data
     * @returns {Number | undefined}
     */
    this.getSampleGridBinCount = function(data){
        let sampleGrid = sampleGridCache.getSampleGrid(data);
        return sampleGrid ? sampleGrid.getBinCount() : undefined;
    };
    
    /**
     * @returns {Boolean} True if ref image is color
     */
    this.isColor = function(){
        return refImageCache.getView().image.isColor;  
    };
    
    /**
     * Call invalidate on all four caches.
     * refImageCache, tgtImageCache, photometryCache, sampleGridCache
     */
    this.invalidate = function(){
        refImageCache.invalidate();
        tgtImageCache.invalidate();
        photometryCache.invalidate();
        sampleGridCache.invalidate();
    };
    
    /**
     * Call invalidate on all four caches.
     * refImageCache, tgtImageCache, photometryCache, sampleGridCache
     */
    this.invalidateTgt = function(){
        tgtImageCache.invalidate();
        photometryCache.clearTargetData();
        sampleGridCache.clearTargetData();
    };
}
