/* global TITLE */

// Version 1.0 (c) John Murphy 10th-March-2020
//
// ======== #license ===============================================================
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, version 3 of the License.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.
// =================================================================================
//"use strict";

// Map: filename -> FITSKeyword[]
let NSG_FILENAME_FITSKeywords_MAP = new Map();
/**
 * @param {String} fitsFilePath Full filename of image.
 * @throws {Error} If file cannot be read.
 * @returns {FITSKeyword[]}
 */
function getFITSKeywords( fitsFilePath ){
    if (NSG_FILENAME_FITSKeywords_MAP.has(fitsFilePath)){
        return NSG_FILENAME_FITSKeywords_MAP.get(fitsFilePath);
    }
    let keywords = LoadFITSKeywords( fitsFilePath );
    NSG_FILENAME_FITSKeywords_MAP.set(fitsFilePath, keywords);
    return keywords;
}

function FitsHeader(){
    let self = this;
    /**
     * If the fitsKeyword is a string, convert it into a number
     * @param {FITSKeyword} fitsKeyword
     * @returns {Number | undefined}
     */
    function getNumericValue(fitsKeyword){
        if (fitsKeyword.isNumeric){
            return fitsKeyword.numericValue;
        } else if (fitsKeyword.isString){
            let n = parseFloat(fitsKeyword.strippedValue);
            return isNaN(n) ? undefined : n;
        }
        return undefined;
    }
    
    /**
     * @param {FITSKeyword[]} fitsKeywords
     * @returns {Number} Altitude of object (or undefined)
     */
    this.getAltitude = function(fitsKeywords){
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "OBJCTALT")
                return getNumericValue(fitsKeyword);
        }
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "CENTALT")
                return getNumericValue(fitsKeyword);
        }
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "ALTITUDE")
                return getNumericValue(fitsKeyword);
        }
        return undefined;
    };
    
    /**
     * @param {FITSKeyword[]} fitsKeywords
     * @returns {Number} Airmass of object (or undefined)
     */
    this.getAirmass = function(fitsKeywords){
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "AIRMASS")
                return getNumericValue(fitsKeyword);
        }
        return undefined;
    };
    
    /**
     * @param {FITSKeyword[]} fitsKeywords
     * @returns {Number} Exposure time in seconds (or undefined)
     */
    this.getExposure = function(fitsKeywords){
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "EXPOSURE")
                return fitsKeyword.numericValue;
        }
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "EXPTIME")
                return fitsKeyword.numericValue;
        }
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "ELAPTIME")
                return fitsKeyword.numericValue;
        }
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "TELAPSE")
                return fitsKeyword.numericValue;
        }
        return undefined;
    };

    /**
     * @param {FITSKeyword[]} fitsKeywords
     * @returns {Number | undefined} Focal length in mm
     */
    this.getFocalLength = function(fitsKeywords){
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "FOCALLEN")
                return Math.round(fitsKeyword.numericValue);
        }
        return undefined;
    };

    /**
     * @param {FITSKeyword[]} fitsKeywords
     * @returns {Number | undefined} Pixel size in microns
     */
    this.getPixelSize = function(fitsKeywords){
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "XPIXSZ")
                return fitsKeyword.numericValue;
        }
        return undefined;
    };
    
    /**
     * @param {FITSKeyword[]} fitsKeywords
     * @returns {String} Filter name or "NONE"
     */
    this.getFilter = function(fitsKeywords){
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "FILTER")
                return fitsKeyword.strippedValue;
        }
        return "NONE";
    };
    
    /**
     * If color image, return value for channel zero.
     * @param {FITSKeyword[]} fitsKeywords
     * @returns {undefined|Number} PSFSGL / NOISExx
     */
    this.getPsfSignalWeight = function(fitsKeywords){
        let R = false;
        let G = false;
        let B = false;
        let index;
        let psfsgl = [];
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "PSFSGL00"){
                psfsgl[0] = fitsKeyword.numericValue;
                index = 0;
                R = true;
            }
            if (fitsKeyword.name === "PSFSGL01"){
                psfsgl[1] = fitsKeyword.numericValue;
                index = 1;
                G = true;
            }
            if (fitsKeyword.name === "PSFSGL02"){
                psfsgl[2] = fitsKeyword.numericValue;
                index = 2;
                B = true;
            }
        }
        if (R && G && B){
            // All three defined so RGB image. Return value for Red
            index = 0;
        }
        
        if (index !== undefined){
            let noiseData = self.getNoise(fitsKeywords);
            if (noiseData && noiseData.noise[index]){
                return psfsgl[index] / noiseData.noise[index];
            }
        }
        
        return undefined;
    };
    
    /**
     * @param {FITSKeyword[]} fitsKeywords
     * @returns {Noise} noise estimate or undefined
     */
    this.getNoise = function(fitsKeywords){
        let R = false;
        let G = false;
        let B = false;
        let index;
        let type = [];
        let noise = [];
        for (let fitsKeyword of fitsKeywords) {
            if (fitsKeyword.name === "NOISE00"){
                noise[0] = fitsKeyword.numericValue;
                index = 0;
                R = true;
            }
            if (fitsKeyword.name === "NOISE01"){
                noise[1] = fitsKeyword.numericValue;
                index = 1;
                G = true;
            }
            if (fitsKeyword.name === "NOISE02"){
                noise[2] = fitsKeyword.numericValue;
                index = 2;
                B = true;
            }
            
            if (fitsKeyword.name === "NOISEA00")
                type[0] = fitsKeyword.strippedValue;
            if (fitsKeyword.name === "NOISEA01")
                type[1] = fitsKeyword.strippedValue;
            if (fitsKeyword.name === "NOISEA02")
                type[2] = fitsKeyword.strippedValue;
        }
        
        if (R && G && B){
            // All three defined so RGB image
            index = -1;
        }
        
        if (index !== undefined)
            return new Noise(noise, type, index);
        
        return undefined;
    };

    /**
     * @param {FITSKeyword[]} keywords
     * @param {String} filename
     */
    this.fitsHeaderRefFilename = function (keywords, filename){
        keywords.push(new FITSKeyword("NSGREF", filename, "Normalized to this reference"));
    };
    
    /**
     * @param {FITSKeyword[]} keywords
     * @param {Number} image
     */
    this.fitsHeaderDimension = function (keywords, image){
        let w = image.width;
        let h = image.height;
        keywords.push(new FITSKeyword("NSGWIDTH", w.toString(), "NSG.width of reference image"));
        keywords.push(new FITSKeyword("NSGHIGHT", h.toString(), "NSG.height of reference image"));
    };

    /**
     * @param {FITSKeyword[]} keywords
     * @param {NsgData} data
     */
    this.fitsHeaderGradient = function (keywords, data){
        keywords.push(new FITSKeyword("HISTORY", "", 
            "NSG.gradientSmoothness: " + data.gradientSmoothness));
    };
    
    /**
     * @param {FITSKeyword[]} keywords
     * @param {Number} imageMaximum
     */
    this.fitsHeaderMaxValue = function(keywords, imageMaximum){
        let comment = "Higher value indicates fewer saturated stars";
        keywords.push(new FITSKeyword("NSGHIGH", imageMaximum.toPrecision(5), comment));
    };

    /**
     * @param {FITSKeyword[]} keywords
     * @param {LinearFitData[]} scaleFactors
     */
    this.fitsHeaderScale = function(keywords, scaleFactors){
        for (let c = 0; c < scaleFactors.length; c++){
            let inverse = 1 / scaleFactors[c].m;
            let comment = "NSG.scale[" + c + "]";
            keywords.push(new FITSKeyword("NSGS" + c, inverse.toPrecision(5), comment));
        }
    };
    
    /**
     * @param {FITSKeyword[]} keywords
     * @param {Number[]} signalToNoise
     */
    this.fitsHeaderSignalToNoise = function(keywords, signalToNoise){
        for (let c = 0; c < signalToNoise.length; c++){
            let comment = "NSG.SNR[" + c + "] Signal to noise ratio";
            let snr = signalToNoise[c];
            snr = snr ? snr : 0;    // snr is undefined if the noise estimate is zero
            keywords.push(new FITSKeyword("NSGSNR" + c, snr.toPrecision(5), comment));
        }
    };
    
    /**
     * Add noise MRS to a HISTORY comment
     * @param {FITSKeyword[]} keywords
     * @param {Number[]} noiseValues
     * @param {String} type NOISExx, noiseMRS, noiseKSigma
     */
    this.fitsHeaderNoise = function(keywords, noiseValues, type){
        if (noiseValues.length > 1){
            for (let c = 0; c < noiseValues.length; c++){
                if (noiseValues[c]){
                    let comment = "NSG." + type + "[" + c + "]: " + noiseValues[c].toPrecision(5);
                    keywords.push(new FITSKeyword("HISTORY", "", comment));
                }
            }
        } else if (noiseValues.length === 1 && noiseValues[0]){
            let comment = "NSG." + type + ": " + noiseValues[0].toPrecision(5);
            keywords.push(new FITSKeyword("HISTORY", "", comment));
        }
    };
    
    /**
     * Add weight based on noise evaluation to the FITS header
     * @param {FITSKeyword[]} keywords
     * @param {String} keyword
     * @param {Number} averageWeight average weight over all channels
     * @param {String} type NOISExx, noiseMRS, noiseKSigma 
     */
    this.fitsHeaderNoiseWeight = function(keywords, keyword, averageWeight, type){
        deleteKeyword(keywords, keyword);
        let comment = type.length > 0 ? " calculated from " + type : "";
        keywords.push(new FITSKeyword(keyword, averageWeight.toPrecision(5), 
                    "NSG.weight" + comment));
    };
    
}

/**
 * @param {FITSKeyword[]} keywords
 * @param {String} name Keyword to be deleted
 */
function deleteKeyword(keywords, name){
    for (let i=0; i<keywords.length; i++) {
        if (keywords[i].name === name){
            keywords.splice(i, 1);  // Delete one entry
            break;
        }
    }
}

/**
 * Calculate image scale in degrees per pixel from pixel size and focal length
 * @param {Number} pixelSize
 * @param {Number} focalLength
 * @returns {Number}
 */
function calcDegreesPerPixel(pixelSize, focalLength){
    return ((pixelSize * 1.0e-6) / (focalLength * 1.0e-3)) * (180 / Math.PI);
}
