/* global StdButton_Yes, UndoFlag_NoSwapFile */

// Version 1.0 (c) John Murphy 8th-Oct-2020
//
// ======== #license ===============================================================
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, version 3 of the License.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.
// =================================================================================
//"use strict";

/**
 * Calculates maximum and minimum values for the sample points
 * @param {SamplePair[][]} colorSamplePairs SamplePair[] for each channel
 * @param {Number} minRange The range will be at least this big
 * @param {Number} zoomFactor Zoom in by modifying minDif and maxDif (smaller
 * range produces a more zoomed in view)
 * @param {Number} selectedChannel R=0, G=1, B=2, All=3
 * @returns {SamplePairDifMinMax}
 */
function SamplePairDifMinMax(colorSamplePairs, minRange, zoomFactor, selectedChannel) {
    this.minDif = Number.POSITIVE_INFINITY;
    this.maxDif = Number.NEGATIVE_INFINITY;
    let values = [];
    for (let c=0; c<colorSamplePairs.length; c++) {
        if (selectedChannel === 3 || selectedChannel === c){
            let samplePairs = colorSamplePairs[c];
            for (let samplePair of samplePairs) {
                let dif = samplePair.getDifference();
                values.push(dif);
                this.minDif = Math.min(this.minDif, dif);
                this.maxDif = Math.max(this.maxDif, dif);
            }
        }
    }
    
    let dataRange = this.maxDif - this.minDif;
    let range = Math.max(dataRange, minRange) / zoomFactor;
    if (range > dataRange){
        // All points fit on the graph. Provide equal space above and below.
        let space = (range - dataRange) / 2;
        this.maxDif += space;
        this.minDif -= space;
    } else if (range < dataRange){
        // The points don't all fit on the graph. Center on median value.
        let median = Math.median(values);
        let max = median + range / 2;
        let min = median - range / 2;
        if (this.maxDif < max){
            let dif = max - this.maxDif;
            max -= dif;
            min -= dif;
        } else if (this.minDif > min){
            let dif = this.minDif - min;
            max += dif;
            min += dif;
        }
        this.maxDif = max;
        this.minDif = min;
    }
}

/**
 * 
 * @param {SamplePair[][]} samplePairsOnPath Sorted by distance along gradient path
 * @param {Number} minRange
 * @returns {Number}
 */
function getNoiseRange(samplePairsOnPath, minRange){
    let rangeArray = [];
    for (let c=0; c<samplePairsOnPath.length; c++) {
        let samplePairs = samplePairsOnPath[c];
        for (let i = 0; i < samplePairs.length - 10; i += 10){
            let max = samplePairs[i].getDifference();
            let min = max;
            for (let j = 1; j<10; j++){
                let difference = samplePairs[i+j].getDifference();
                max = Math.max(difference, max);
                min = Math.min(difference, min);
            }
            rangeArray.push(max - min);
        }
    }
    let range = rangeArray.length > 0 ? Math.max(Math.median(rangeArray), minRange) : minRange;
    return range;
}

/**
 * Returns the SamplePairs that are closest to the specified horizontal or vertical line
 * @param {SamplePair[][]} colorSamplePairs
 * @param {Number} maxDist If > 0 limit to samples less than this distance from gradient path
 * @param {Boolean} isHorizontal Horizontal or vertical line
 * @param {Number} lineEqualX Specifies vertical line position
 * @param {Number} lineEqualY Specifies horizontal line position
 * @returns {SamplePair[][]} SamplePair are sorted by distance along gradient path
 */
function getSamplePairsNearLine(colorSamplePairs, maxDist, isHorizontal, lineEqualX, lineEqualY){
    /**
     * @param {SamplePair} samplePair
     * @param {Boolean} isHorizontal Horizontal or vertical line
     * @param {Number} lineEqualX Specifies vertical line position
     * @param {Number} lineEqualY Specifies horizontal line position
     * @returns {GradientGraph.getSamplePairsNearLine.MapEntry} contains {samplePair, dif, key}
     */
    function MapEntry (samplePair, isHorizontal, lineEqualX, lineEqualY){
        this.samplePair = samplePair;
        this.dist = Number.POSITIVE_INFINITY;
        this.key = -1;
        if (isHorizontal){
            this.key = Math.round(samplePair.rect.center.x);
            this.dist = Math.abs(samplePair.rect.center.y - lineEqualY);
            
        } else {
            this.key = Math.round(samplePair.rect.center.y);
            this.dist = Math.abs(samplePair.rect.center.x - lineEqualX);
        }
    }

    let dataSamplePairs = [];
    let nChannels = colorSamplePairs.length;
    for (let c=0; c<nChannels; c++){
        dataSamplePairs[c] = [];
        let pathMap = new Map();
        for (let i=0; i<colorSamplePairs[c].length; i++){
            let samplePairs = colorSamplePairs[c];
            let value = new MapEntry(samplePairs[i], isHorizontal, lineEqualX, lineEqualY);
            if (maxDist <= 0 || value.dist < maxDist){
                let key = value.key;
                if (pathMap.has(key)){
                    let mapValue = pathMap.get(key);
                    if (value.dist < mapValue.dist){
                        // closer to path
                        pathMap.set(key, value);
                    }
                } else {
                    pathMap.set(key, value);
                }
            }
        }

        // Get values from map and convert to an array.
        for (let mapValue of pathMap.values()){
            dataSamplePairs[c].push(mapValue.samplePair);
        }
        if (isHorizontal){
            dataSamplePairs[c].sort((a, b) => a.rect.x0 - b.rect.x0);
        } else {
            dataSamplePairs[c].sort((a, b) => a.rect.y0 - b.rect.y0);
        }
        pathMap.clear();
    }

    return dataSamplePairs;
}

/**
 * Display graph of (difference between images) / (pixel distance across image)
 * @param {SamplePair[][]} colorSamplePairs The SamplePair points, corrected for scale
 * @param {NsgDialog} nsgDialog
 * @param {NsgData} data User settings used to create FITS header
 * @returns {undefined}
 */
function GradientGraph(colorSamplePairs, nsgDialog, data){
    let minScaleDif_ = 1e-9;
    let binnedColorSamplePairsCache_ = new Map();
    let surfaceSplinesCache_ = new Map();
    let graphBitmapLum;
    let graphBitmapRGB;
    let imageRect;
    
    function construct(){
        let title = "Gradient Graph";
        let image = data.cache.getRefView().image;
        imageRect = new Rect(image.width, image.height);
        // Display graph in script dialog
        let isColor = colorSamplePairs.length > 1;
        let graphDialog = new GradientGraphDialog(title, data, isColor, 
                createZoomedGradientGraph, nsgDialog);
        graphDialog.execute();
        
        // Dialog has closed. Clear cache
        graphDialog = null;
        for (let surfaceSpline of surfaceSplinesCache_.values()) {
            surfaceSpline.clear();
        }
        binnedColorSamplePairsCache_.clear();
        surfaceSplinesCache_.clear();
        if (graphBitmapLum){
            graphBitmapLum.clear();
            graphBitmapLum = null;
        }
        if (graphBitmapRGB){
            graphBitmapRGB.clear();
            graphBitmapRGB = null;
        }
    }
    
    /**
     * Callback function for GraphDialog to provide a zoomed graph.
     * GraphDialog uses Graph.getGraphBitmap() and the function pointer Graph.screenToWorld
     * @param {Number} zoomFactor
     * @param {Number} width
     * @param {Number} height
     * @param {Number} selectedChannel R=0, G=1, B=2, All=3
     * @param {Boolean} isHorizontal 
     * @param {Number} lineEqualX Specify path of vertical gradient line
     * @param {Number} lineEqualY Specify path of horizontal gradient line
     * @returns {Graph}
     */
    function createZoomedGradientGraph(zoomFactor, width, height, selectedChannel, isHorizontal,
            lineEqualX, lineEqualY){
        let firstMsg = true;
        let smoothness = data.gradientSmoothness;
        // === Binned sample pairs ===
        let bKey = calcMaxSamples(data);
        let bValue = binnedColorSamplePairsCache_.get(bKey);
        if (bValue === undefined){
            console.writeln("\n<b><u>Binning samples (max samples = " + bKey + ")</u></b> ");
            bValue = createColorBinnedSamplePairs(data, colorSamplePairs);
            binnedColorSamplePairsCache_.set(bKey, bValue);
            if (bValue[0].length < colorSamplePairs[0].length){
                console.writeln("Reduced samples from ", colorSamplePairs[0].length, " to ", bValue[0].length);
            } else {
                console.writeln("Using ", colorSamplePairs[0].length, " samples");
            }
        }
        let binnedColorSamplePairs = bValue;
        let nChannels = binnedColorSamplePairs.length;
        
        // === SurfaceSplines ===
        let surfaceSplines = [];
        for (let c = 0; c < nChannels; c++) {
            if (selectedChannel === c || selectedChannel === 3){
                let key = "_" + smoothness + "_" + c + "_";
                let value = surfaceSplinesCache_.get(key);
                if (value === undefined){
                    if (firstMsg){
                        console.writeln("\n<b><u>Calculating surface spline</u></b> ");
                        firstMsg = false;
                    }
                    let samplePairs = binnedColorSamplePairs[c];
                    value = calcSurfaceSpline(samplePairs, smoothness);
                    surfaceSplinesCache_.set(key, value);
                    console.writeln("Created surface spline [", c, "] smoothness ", 
                            smoothness, " from ", samplePairs.length, " samples");
                }
                surfaceSplines[c] = value;
            } else {
                surfaceSplines[c] = null;
            }
        }
        
        // === Create Graph
        // Using GradientGraph function call parameters
        let graph = createGraph(width, height, isHorizontal, surfaceSplines,
                imageRect, data, zoomFactor, selectedChannel, lineEqualX, lineEqualY);
        return graph;
    }
    
    /**
     * @param {Number} width
     * @param {Number} height
     * @param {Boolean} isHorizontal
     * @param {SurfaceSpline[]} surfaceSplines Difference between reference and target images
     * @param {Rect} imageRect 
     * @param {NsgData} data User settings used to create FITS header
     * @param {Number} zoomFactor Zoom factor for vertical axis only zooming.
     * @param {Number} selectedChannel R=0, G=1, B=2, All=3
     * @param {Number} lineEqualX Specify path of vertical gradient line
     * @param {Number} lineEqualY Specify path of horizontal gradient line
     * @returns {Graph}
     */
    function createGraph(width, height, isHorizontal, surfaceSplines,
                imageRect, data, zoomFactor, selectedChannel, lineEqualX, lineEqualY){
        let xLabel = isHorizontal ? "X-coordinate" : "Y-coordinate";
        let yLabel = "(" + data.cache.getTgtView().fullId + ") - (" + data.cache.getRefView().fullId + ")";
        
        // Get the SamplePairs that are closest to the line path
        let maxDist = data.sampleSize * 2.5;
        let dataSamplePairs = getSamplePairsNearLine(colorSamplePairs, maxDist, isHorizontal, lineEqualX, lineEqualY);
        // Graph scale
        let minScaleDif = 10 * getNoiseRange(dataSamplePairs, minScaleDif_);
        
        let yCoordinateRange = new SamplePairDifMinMax(dataSamplePairs, minScaleDif, zoomFactor, selectedChannel);
        
        return createAndDrawGraph(xLabel, yLabel, yCoordinateRange, width, height,
                isHorizontal, lineEqualX, lineEqualY, 
                surfaceSplines, imageRect, dataSamplePairs, selectedChannel);
    }
    
    /**
     * Draw gradient line and sample points for a single color channel.
     * @param {Graph} graph
     * @param {Boolean} isHorizontal
     * @param {Number[]} difArray Points to plot. Offset difference between ref and tgt
     * @param {Number} lineColor
     * @param {SamplePair[]} samplePairs
     * @param {Number} pointColor
     * @returns {undefined}
     */
    function drawLineAndPoints(graph, isHorizontal,
            difArray, lineColor, samplePairs, pointColor) {
                
        for (let samplePair of samplePairs) {
            // Draw the sample points
            let coord = isHorizontal ? samplePair.rect.center.x : samplePair.rect.center.y;
            graph.drawCross(coord, samplePair.getDifference(), pointColor);
        }
        graph.drawCurve(difArray, lineColor, false);
    }
    
    /**
     * 
     * @param {String} xLabel
     * @param {String} yLabel
     * @param {SamplePairDifMinMax} yCoordinateRange 
     * @param {Number} width 
     * @param {Number} height 
     * @param {Boolean} isHorizontal
     * @param {Number} lineEqualX Specifies the vertical line gradient path
     * @param {Number} lineEqualY Specifies the horizontal line gradient path
     * @param {SurfaceSpline[]} surfaceSplines
     * @param {Rect} imageRect
     * @param {SamplePair[][]} dataSamplePairs
     * @param {Number} selectedChannel R=0, G=1, B=2, All=3
     * @returns {Graph}
     */
    function createAndDrawGraph(xLabel, yLabel, yCoordinateRange, width, height, 
            isHorizontal, lineEqualX, lineEqualY,
            surfaceSplines, imageRect, dataSamplePairs, selectedChannel){
        let maxY = yCoordinateRange.maxDif;
        let minY = yCoordinateRange.minDif;
        let maxX;
        if (isHorizontal){
            maxX = imageRect.x1;
        } else {
            maxX = imageRect.y1;
        }
        if (!graphBitmapLum || graphBitmapLum.width !== width || graphBitmapLum.height !== height){
            if (graphBitmapLum){
                graphBitmapLum.clear();
            }
            graphBitmapLum = new Bitmap(width, height);
        }
        let graphDimensions = new GraphDimensions(0, minY, maxX, maxY, true);
        let graph = new Graph(graphDimensions, xLabel, yLabel, graphBitmapLum, 1, 1);
        
        let ssPath = [];
        let nAxisPixels = graphDimensions.getXAxisLength(width);
        let pixelWidth = 1 / graphDimensions.getXScale(width, 1);
        for (let xp=0; xp<nAxisPixels; xp++){
            // For each pixel along the graph x axis, append the corresponding image point
            let imageCoord = xp * pixelWidth;
            if (isHorizontal){
                ssPath.push(new Point(imageCoord, lineEqualY));
            } else {
                ssPath.push(new Point(lineEqualX, imageCoord));
            }
        }
        
        if (dataSamplePairs.length === 1){ // B&W
            let difArray = surfaceSplines[0].evaluate(ssPath).toArray();
            drawLineAndPoints(graph, isHorizontal,
                difArray, 0xFF990000, dataSamplePairs[0], 0xFFFFFFFF);
        } else {
            // Color. Need to create 3 graphs for r, g, b and then merge them (binary OR) so that
            // if three samples are on the same pixel we get white and not the last color drawn
            let lineColors = [0xFF990000, 0xFF009900, 0xFF000099]; // r, g, b
            let pointColors = [0xFFFF0000, 0xFF00FF00, 0xFF5555FF]; // r, g, b
            // Provided the saved bitmap is the same size, we can reuse it.
            // The Graph will fill the bitmap with zeros before using it.
            let bitmapSize = graph.getGraphAreaOnlySize();
            if (!graphBitmapRGB || 
                    graphBitmapRGB.width !== bitmapSize.width || 
                    graphBitmapRGB.height !== bitmapSize.height){
                if (graphBitmapRGB){
                    graphBitmapRGB.clear();
                }
                graphBitmapRGB = new Bitmap(bitmapSize.width, bitmapSize.height);
            }
            for (let c = 0; c < dataSamplePairs.length; c++){
                if (selectedChannel === 3 || selectedChannel === c){
                    let difArray = surfaceSplines[c].evaluate(ssPath).toArray();
                    let graphAreaOnly = graph.graphAreaOnlyFactory(graphBitmapRGB);
                    drawLineAndPoints(graphAreaOnly, isHorizontal,
                        difArray, lineColors[c], dataSamplePairs[c], pointColors[c]);
                    graph.mergeWithGraphAreaOnly(graphAreaOnly);
                }
            }
        }
        return graph;
    }
    
    construct();
}
