/* global UndoFlag_NoSwapFile, Dialog, StdButton_No, StdIcon_Question, StdButton_Cancel, StdButton_Yes */

// Version 1.0 (c) John Murphy 12th-Aug-2020
//
// ======== #license ===============================================================
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, version 3 of the License.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.
// =================================================================================
//"use strict";

/**
 * Create a dialog that displays a graph.
 * The Graph object returned from the supplied createZoomedGradientGraph(Number zoomFactor) 
 * function must include the methods:
 * Bitmap Graph.getGraphBitmap()
 * String Graph.screenToWorld(Number x, Number y)
 * The GraphDialog is initialised with the Graph returned from createZoomedGradientGraph, 
 * with a zoom factor of 1
 * @param {String} title Window title
 * @param {NsgData} data
 * @param {Boolean} isColor
 * @param {Graph function({Number} zoomFactor, {Number} width, {Number} height, {Number} channel)} createZoomedGradientGraph
 * Callback function used to create a zoomed graph
 * @param {NsgDialog} nsgDialog
 * @returns {GradientGraphDialog}
 */
function GradientGraphDialog(title, data, isColor, createZoomedGradientGraph, nsgDialog)
{
    this.__base__ = Dialog;
    this.__base__();
    let self = this;
    let zoom_ = 1;
    let selectedChannel_ = 3;
    let createZoomedGradientGraph_ = createZoomedGradientGraph;
    let graphHeight = data.smallScreen ? data.graphHeight - 300 : data.graphHeight;
    let width = this.logicalPixelsToPhysical(data.graphWidth);
    let height = this.logicalPixelsToPhysical(graphHeight);
    let image = data.cache.getRefView().image;
    let halfImageWidth = Math.round(image.width / 2);
    let halfImageHeight = Math.round(image.height / 2);
    let graph_ = createZoomedGradientGraph_(zoom_, width, height, selectedChannel_,
            data.isGradientLineHorizontal, 
            data.gradientLineX + halfImageWidth, 
            data.gradientLineY + halfImageHeight);
    let displayingGraph = true;
    let displayingRefImage = false;
    // Make sure bitmaps are in the cache
    data.cache.getTgtImageBitmap();
    data.cache.getRefImageBitmap();
    let allStars = data.cache.getRefStars(data.logStarDetection);
    
    /**
     * Converts bitmap (x,y) into graph coordinates.
     * @param {Number} x Bitmap x coordinate
     * @param {Number} y Bitmap y coordinate
     * @returns {String} Output string in format "( x, y )"
     */
    function displayXY(x, y){
        let xyString;
        if (displayingGraph){
            xyString = graph_.screenToWorld(x, y);
        } else {
            let imageScale = calcImageScale();
            let ix = Math.round((x - imageScale.dx)/imageScale.scale);
            let iy = Math.round((y - imageScale.dy)/imageScale.scale);
            xyString = "(" + ix + "," + iy + ")";
        }
        self.windowTitle = title + getZoomString() + "  " + xyString;
    };
    
    function calcImageScale(){
        let bitmap = data.cache.getTgtImageBitmap();
        let scaleX = (bitmapControl.width - 4) / bitmap.width;
        let scaleY = (bitmapControl.height - 4) / bitmap.height;
        let scale = Math.min(scaleX, scaleY);
        let dx = Math.round((bitmapControl.width - bitmap.width * scale) / 2);
        let dy = Math.round((bitmapControl.height - bitmap.height * scale) / 2);
        return {scale: scale, dx: dx, dy: dy};
    }
    
    // Draw bitmap into this component
    let bitmapControl = new Control(this);
    
    bitmapControl.onPaint = function (){
        let g;
        try {
            g = new Graphics(bitmapControl);
            g.clipRect = new Rect(0, 0, bitmapControl.width, bitmapControl.height);
            if (displayingGraph){
                g.drawBitmap(0, 0, graph_.getGraphBitmap());
            } else {
                let bitmap;
                if (displayingRefImage){
                    bitmap = data.cache.getRefImageBitmap();
                } else {
                    bitmap = data.cache.getTgtImageBitmap();
                }
                let imageScale = calcImageScale();
                g.fillRect(0, 0, bitmapControl.width, bitmapControl.height, new Brush());
                g.translateTransformation( imageScale.dx, imageScale.dy );
                g.scaleTransformation( imageScale.scale );
                g.drawBitmap(0, 0, bitmap);
                g.antialiasing = false;
                let x = data.gradientLineX + halfImageWidth;
                let y = data.gradientLineY + halfImageHeight;
                g.pen = data.isGradientLineHorizontal ? new Pen(0xff00ff00, 0.0) : new Pen(0xff008800, 0.0);
                g.drawLine( 0, y, bitmap.width - 1, y );
                g.pen = data.isGradientLineHorizontal ? new Pen(0xff008800, 0.0) : new Pen(0xff00ff00, 0.0);
                g.drawLine( x, 0, x, bitmap.height - 1);
                drawRejectionCircles(g, 0, data, allStars, -1);
            }
        } catch (e) {
            console.criticalln("GradientGraphDialog bitmapControl.onPaint() error: " + e);
        } finally {
            g.end();
        }
    };
    
    bitmapControl.onMouseDoubleClick = function ( clickX, clickY, buttonState, modifiers ){
        if (!displayingGraph){
            let imageScale = calcImageScale();
            let x = Math.round((clickX - imageScale.dx)/imageScale.scale);
            let y = Math.round((clickY - imageScale.dy)/imageScale.scale);
            x = Math.min(x, control_X.upperBound);
            x = Math.max(x, control_X.lowerBound);
            y = Math.min(y, control_Y.upperBound);
            y = Math.max(y, control_Y.lowerBound);
            data.gradientLineX = x - halfImageWidth;
            data.gradientLineY = y - halfImageHeight;
            control_X.setValue(x);
            control_Y.setValue(y);
            update(bitmapControl.width, bitmapControl.height);
        }
    };
    
    
    bitmapControl.onMousePress = function ( x, y, button, buttonState, modifiers ){
        // Display graph coordinates in title bar
        displayXY(x, y);
    };
    
    bitmapControl.onMouseMove = function ( x, y, buttonState, modifiers ){
        // When dragging mouse, display graph coordinates in title bar
        displayXY(x, y);
        // TODO create pan mode using space bar (modifiers = 8)
    };
    
    bitmapControl.onMouseWheel = function ( x, y, delta, buttonState, modifiers ){
        if (displayingGraph){
            if (delta < 0){
                updateZoom( zoom_ + 1);
            } else {
                updateZoom( zoom_ - 1);
            }
        }
    };
    
    bitmapControl.onResize = function (wNew, hNew, wOld, hOld) {
        update(wNew, hNew);
    };
    
    /**
     * @param {Number} zoom
     */
    function updateZoom (zoom) {
        if (zoom < 101 && zoom > -99){
            zoom_ = zoom;
            update(bitmapControl.width, bitmapControl.height);
            self.windowTitle = title + getZoomString();   // display zoom factor in title bar
        }
    }
    
    /**
     * @param {Number} width Graph bitmap width
     * @param {Number} height Graph bitmap height
     */
    function update(width, height){
        try {
            if (displayingGraph){
                graph_ = createZoomedGradientGraph_(getZoomFactor(), width, height, selectedChannel_, 
                    data.isGradientLineHorizontal, 
                    data.gradientLineX + halfImageWidth, data.gradientLineY + halfImageHeight);
            }
            bitmapControl.repaint();    // display the zoomed graph bitmap
        } catch (e) {
            console.criticalln("Graph update error: " + e);
        }
    }
    
    /**
     * If zoom_ is positive, return zoom_ (1 to 100)
     * If zoom_ is zero or negative, then:
     * 0 -> 1/2
     * -1 -> 1/3
     * -2 -> 1/4
     * -98 -> 1/100
     * @returns {Number} Zoom factor
     */
    function getZoomFactor(){
        return zoom_ > 0 ? zoom_ : 1 / (2 - zoom_);
    }
    
    /**
     * @returns {String} Zoom string (e.g. " 1:2")
     */
    function getZoomString(){
        let zoomFactor = getZoomFactor();
        if (zoomFactor < 1){
            return " 1:" + Math.round(1/zoomFactor);
        } else {
            return " " + zoomFactor + ":1";
        }
    }
    
    bitmapControl.toolTip = "Left click: Display (x,y) in title bar";
    
    // ========================================
    // User controls
    // ========================================
    let minHeight = bitmapControl.minHeight;
    
    this.onToggleSection = function(bar, beginToggle){
        if (beginToggle){
            if (bar.isExpanded()){
                bitmapControl.setMinHeight(bitmapControl.height + bar.section.height + 2);
            } else {
                bitmapControl.setMinHeight(bitmapControl.height - bar.section.height - 2);
            }
            this.adjustToContents();
        }  else {
            bitmapControl.setMinHeight(minHeight);
            let maxDialogHeight = self.logicalPixelsToPhysical(1150);
            if (self.height > maxDialogHeight)
                self.resize(self.width, maxDialogHeight);
        }
    };
    
    /**
     * When a slider is dragged, only fast draw operations are performed.
     * When the drag has finished (or after the user has finished editing in the textbox)
     * this method is called to perform all calculations.
     */
    function finalUpdateFunction(){
        self.enabled = false;
        processEvents();
        update(bitmapControl.width, bitmapControl.height);
        self.enabled = true;
        processEvents();
    }
    
    // Gradient controls
    let labelMinLength = this.font.width(" Horizontal gradient line: y =");
    let gradientControls = new NsgGradientControls();
    let smoothnessControl = gradientControls.createGradientSmoothnessControl(this, data, 0);
    smoothnessControl.label.minWidth = labelMinLength;
    smoothnessControl.onValueUpdated = function (value) {
        data.gradientSmoothness = value;
        nsgDialog.gradientSmoothness_Control.setValue(value);
    };
    smoothnessControl.slider.minWidth = 300;
    addFinalUpdateListener(smoothnessControl, finalUpdateFunction);

    let directionToggle = new CheckBox(this);
    directionToggle.text = "Horizontal";
    directionToggle.toolTip = "<p>If selected, display the gradient along a horizontal line.</p>" +
            "<p>If not selected, display the gradient along a vertical line.</p>";
    directionToggle.checked = data.isGradientLineHorizontal;
    directionToggle.onClick = function (checked) {
        data.isGradientLineHorizontal = checked;
        self.enabled = false;
        processEvents();
        update(bitmapControl.width, bitmapControl.height);
        self.enabled = true;
        processEvents();
        control_Y.enabled = data.isGradientLineHorizontal;
        control_X.enabled = !data.isGradientLineHorizontal;
    };
    
    let imageCheckBox = new CheckBox(this);
    imageCheckBox.text = "Image";
    imageCheckBox.toolTip = "Display the image instead of the graph.";
    imageCheckBox.checked = !displayingGraph;
    imageCheckBox.onClick = function (checked) {
        self.enableGraphSection(!checked);
        update(bitmapControl.width, bitmapControl.height);
    };
    
    let refCheckBox = new CheckBox(this);
    refCheckBox.text = "Reference";
    refCheckBox.toolTip = "Display either reference or target image.";
    refCheckBox.checked = displayingRefImage;
    refCheckBox.onClick = function (checked) {
        displayingRefImage = checked;
        update(bitmapControl.width, bitmapControl.height);
    };
    
    // ===========================
    // Color toggles
    // ===========================
    let redRadioButton = new RadioButton(this);
    redRadioButton.text = "Red";
    redRadioButton.toolTip = "<p>Display the red channel gradient</p>" + 
            "<p>This is only used to declutter the display. " +
            "The 'Smoothness' setting will be applied to all color channels.</p>";
    redRadioButton.checked = false;
    redRadioButton.onClick = function (checked) {
        selectedChannel_ = 0;
        self.enabled = false;
        processEvents();
        update(bitmapControl.width, bitmapControl.height);
        self.enabled = true;
        processEvents();
    };
    
    let greenRadioButton = new RadioButton(this);
    greenRadioButton.text = "Green";
    greenRadioButton.toolTip = "<p>Display the green channel gradient</p>" + 
            "<p>This is only used to declutter the display. " +
            "The 'Smoothness' setting will be applied to all color channels.</p>";
    greenRadioButton.checked = false;
    greenRadioButton.onClick = function (checked) {
        selectedChannel_ = 1;
        self.enabled = false;
        processEvents();
        update(bitmapControl.width, bitmapControl.height);
        self.enabled = true;
        processEvents();
    };
    
    let blueRadioButton = new RadioButton(this);
    blueRadioButton.text = "Blue";
    blueRadioButton.toolTip = "<p>Display the blue channel gradient</p>" + 
            "<p>This is only used to declutter the display. " +
            "The 'Smoothness' setting will be applied to all color channels.</p>";
    blueRadioButton.checked = false;
    blueRadioButton.onClick = function (checked) {
        selectedChannel_ = 2;
        self.enabled = false;
        processEvents();
        update(bitmapControl.width, bitmapControl.height);
        self.enabled = true;
        processEvents();
    };
    
    let allRadioButton = new RadioButton(this);
    allRadioButton.text = "All";
    allRadioButton.toolTip = "Display the gradient for all channels";
    allRadioButton.checked = true;
    allRadioButton.onClick = function (checked) {
        selectedChannel_ = 3;
        self.enabled = false;
        processEvents();
        update(bitmapControl.width, bitmapControl.height);
        self.enabled = true;
        processEvents();
    };
    
    if (!isColor){
        redRadioButton.enabled = false;
        greenRadioButton.enabled = false;
        blueRadioButton.enabled = false;
    }

    let optionsSizer = new HorizontalSizer(this);
    optionsSizer.margin = 0;
    optionsSizer.spacing = 10;
    optionsSizer.addSpacing(4);

    optionsSizer.add(directionToggle);
    optionsSizer.add(imageCheckBox);
    optionsSizer.add(refCheckBox);
    optionsSizer.addSpacing(20);
    optionsSizer.add(redRadioButton);
    optionsSizer.add(greenRadioButton);
    optionsSizer.add(blueRadioButton);
    optionsSizer.add(allRadioButton);
    optionsSizer.addStretch();
    
    // =================================
    // SectionBar Gradient line position
    // =================================
    let control_Y = new NumericControl(this);
    control_Y.real = false;
    control_Y.label.text = "Horizontal gradient line: y =";
    control_Y.label.minWidth = labelMinLength;
    control_Y.toolTip = "<p>Display the gradient along the horizontal line at this y coordinate</p>";
    control_Y.setRange(0, image.height);
    control_Y.slider.setRange(0, image.height);
    control_Y.maxWidth = Math.max(this.logicalPixelsToPhysical(1000));
    control_Y.enabled = data.isGradientLineHorizontal;
    control_Y.setValue(data.gradientLineY + halfImageHeight);
    control_Y.onValueUpdated = function (value) {
        data.gradientLineY = value - halfImageHeight;
        update(bitmapControl.width, bitmapControl.height);
    };
    addFinalUpdateListener(control_Y, finalUpdateFunction);
    
    let control_X = new NumericControl(this);
    control_X.real = false;
    control_X.label.text = "Vertical gradient line: x =";
    control_X.label.minWidth = labelMinLength;
    control_X.toolTip = "<p>Display the gradient along the vertical line at this x coordinate</p>";
    control_X.setRange(0, image.width);
    control_X.slider.setRange(0, image.width);
    control_X.maxWidth = Math.max(this.logicalPixelsToPhysical(1000));
    control_X.enabled = !data.isGradientLineHorizontal;
    control_X.setValue(data.gradientLineX + halfImageWidth);
    control_X.onValueUpdated = function (value) {
        data.gradientLineX = value - halfImageWidth;
        update(bitmapControl.width, bitmapControl.height);
    };
    addFinalUpdateListener(control_X, finalUpdateFunction);
    
    this.enableGraphSection = function(checked){
        displayingGraph = checked;
        refCheckBox.enabled = !checked;
        redRadioButton.enabled = checked && isColor;
        greenRadioButton.enabled = checked && isColor;
        blueRadioButton.enabled = checked && isColor;
        allRadioButton.enabled = checked;
        smoothnessControl.enabled = checked;
        zoomIn_Button.enabled = checked;
        zoomOut_Button.enabled = checked;
        if (!checked){
            zoom_ = 1;
            self.windowTitle = title + getZoomString();   // display zoom factor in title bar
        }
    };
    
    let linePositionSection = new Control(this);
    linePositionSection.sizer = new VerticalSizer;
    linePositionSection.sizer.spacing = 2;
    linePositionSection.sizer.add(control_Y);
    linePositionSection.sizer.add(control_X);
    linePositionSection.sizer.addSpacing(5);
    let linePositionBar = new SectionBar(this, "Gradient Path");
    linePositionBar.setSection(linePositionSection);
    linePositionBar.onToggleSection = this.onToggleSection;
    linePositionBar.toolTip = 
        "<p>Specifies a line across the gradient surface spline. " +
        "The gradient along this line will be displayed in the graph.</p>";
    
    let graphSection = new Control(this);
    graphSection.sizer = new HorizontalSizer;
    graphSection.sizer.add(smoothnessControl);
    graphSection.sizer.addStretch();
    let graphBar = new SectionBar(this, "Gradient Graph");
    graphBar.setSection(graphSection);
    graphBar.onToggleSection = this.onToggleSection;
    graphBar.toolTip = "<p>Specify the level of smoothing applied to the gradient correction.</p>";
    
    // ===========================
    // Zoom controls and OK button
    // ===========================
    let zoomIn_Button = new ToolButton(this);
    zoomIn_Button.icon = this.scaledResource(":/icons/zoom-in.png");
    zoomIn_Button.setScaledFixedSize(24, 24);
    zoomIn_Button.toolTip = "Zoom In";
    zoomIn_Button.onMousePress = function (){
        updateZoom( zoom_ + 1);
    };

    let zoomOut_Button = new ToolButton(this);
    zoomOut_Button.icon = this.scaledResource(":/icons/zoom-out.png");
    zoomOut_Button.setScaledFixedSize(24, 24);
    zoomOut_Button.toolTip = "Zoom Out";
    zoomOut_Button.onMousePress = function (){
        updateZoom( zoom_ - 1);
    };

    let zoom11_Button = new ToolButton(this);
    zoom11_Button.icon = this.scaledResource(":/icons/zoom-1-1.png");
    zoom11_Button.setScaledFixedSize(24, 24);
    zoom11_Button.toolTip = "Zoom 1:1";
    zoom11_Button.onMousePress = function (){
        updateZoom( 1 );
    };
    
    let update_Button = new PushButton();
    update_Button.text = "Refresh";
    update_Button.toolTip = "<p>Forces the display to refresh.</p>" +
            "<p>(The display usually refreshes automatically)</p>";
    update_Button.defaultButton = true;
    update_Button.onClick = function(){
        finalUpdateFunction();
    };
    
    let ok_Button = new PushButton(this);
    ok_Button.text = "OK";
    ok_Button.icon = this.scaledResource( ":/icons/ok.png" );
    ok_Button.onClick = function(){
        self.ok();
    };

    let zoomButton_Sizer = new HorizontalSizer(this);
    zoomButton_Sizer.margin = 0;
    zoomButton_Sizer.spacing = 4;
    zoomButton_Sizer.add(zoomIn_Button);
    zoomButton_Sizer.add(zoomOut_Button);
    zoomButton_Sizer.add(zoom11_Button);
    zoomButton_Sizer.addSpacing(10);
    zoomButton_Sizer.add(update_Button);
    zoomButton_Sizer.addStretch();
    zoomButton_Sizer.add(ok_Button);
    zoomButton_Sizer.addSpacing(10);
    
    //-------------
    // Global sizer
    //-------------
    this.sizer = new VerticalSizer(this);
    this.sizer.margin = 2;
    this.sizer.spacing = 2;
    this.sizer.add(bitmapControl, 100);
    this.sizer.add(optionsSizer);
    this.sizer.add(linePositionBar);
    this.sizer.add(linePositionSection);
    this.sizer.add(graphBar);
    this.sizer.add(graphSection);
    this.sizer.add(zoomButton_Sizer);
    
    this.userResizable = true;
    let preferredWidth = width + this.sizer.margin * 2;
    let preferredHeight = height + this.sizer.margin * 2 + this.sizer.spacing * 3 + 
           zoomIn_Button.height * 3 + 4;
    this.resize(preferredWidth, preferredHeight);
    
    this.setScaledMinSize(300, 300);
    this.windowTitle = title + " 1:1";
    this.enableGraphSection(displayingGraph);
}

GradientGraphDialog.prototype = new Dialog;