/* global Parameters, View, APERTURE_GROWTH, APERTURE_ADD, APERTURE_GAP, APERTURE_BKG_DELTA, APERTURE_GROWTH_REJECTION, APERTURE_GROWTH_TARGET, LINEAR_RANGE, File, DataType_Boolean, Settings, KEYPREFIX, DataType_Float, DataType_String, DataType_Int32, DataType_UCString, DEFAULT_PIXEL_SIZE, DEFAULT_FOCAL_LENGTH, DEFAULT_INT_MIN_WEIGHT */

// Version 1.0 (c) John Murphy 5th-Apr-2021
//
// ======== #license ===============================================================
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, version 3 of the License.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.
// =================================================================================

#define APERTURE_ADD 1
#define APERTURE_GROWTH 0.25
#define APERTURE_GROWTH_REJECTION 1.0
#define APERTURE_GAP 2
#define APERTURE_BKG_DELTA 10
#define LINEAR_RANGE 0.7
#define DEFAULT_PIXEL_SIZE 6
#define DEFAULT_FOCAL_LENGTH 1000
#define DEFAULT_INT_MIN_WEIGHT 25.0
#define KEYPREFIX "NormalizeScaleGradient"

function NsgData(){
    
    let nsgDialog;
    let self = this;
    
    /**
     * If not running in ViewTarget mode, supply the dialog so that it can be
     * updated when the data changes. 
     * @param {NsgDialog} dialog
     */
    this.setNsgDialog = function(dialog){
        nsgDialog = dialog;
    };
    
    /**
     * Set ref filename, read the image and set some auto values
     * @param {String} filename
     * @returns {Boolean} True if the image was read (not cached)
     */
    this.setRefFilename = function(filename){
        if (nsgDialog){
            nsgDialog.enabled = false;
            processEvents();
        }
        let loadedFile = false;
        let origFilename = this.cache.getRefFilename();
        if (this.cache.setRefFilename(filename)){
            let hasChanged = origFilename !== this.cache.getRefFilename();
            if (hasChanged){
                this.manualRejectionCircles = [];
                this.cache.setTgtFilename(undefined);
            }
            this.setPhotometryAutoValues(this.useAutoPhotometry, true);
            this.setSampleGenerationAutoValues(this.useAutoSampleGeneration, true);
            loadedFile = true;
        }
        if (nsgDialog){
            nsgDialog.enabled = true;
            processEvents();
        }
        return loadedFile;
    };
    
    /**
     * If auto photometry, calc and set aperture gap
     */
    function setApertureGapAutoValue(){
        if (self.useAutoPhotometry && self.cache.getRefFilename()){
            let gap = calcDefaultApertureGap(self);
            let apertureGapControls = new NsgPhotometryControls();
            let upperBound = apertureGapControls.apertureGap.range.max;
            self.apertureGap = Math.min(gap, upperBound);
            if (nsgDialog){
                nsgDialog.apertureGap_Control.setValue(self.apertureGap);
            }
        } 
    }
    
    /**
     * If auto photometry, calc and set aperture bg delta
     */
    function setApertureBgDeltaAutoValue(){
        if (self.useAutoPhotometry && self.cache.getRefFilename()){
            let bgDelta = calcDefaultApertureBgDelta(self);
            let apertureBgDeltaControls = new NsgPhotometryControls();
            let upperBound = apertureBgDeltaControls.apertureBgDelta.range.max;
            self.apertureBgDelta = Math.min(bgDelta, upperBound);
            if (nsgDialog){
                nsgDialog.apertureBgDelta_Control.setValue(self.apertureBgDelta);
            }
        } 
    }
    
    /**
     * If auto photometry, calc and set max linear value
     */
    function setLinearRangeAutoValue(){
        if (self.useAutoPhotometry && self.cache.getRefFilename()){
            self.linearRangeRef = self.cache.getLinearRangeRef();
            if (nsgDialog){
                nsgDialog.linearRangeRef_Control.setValue(self.linearRangeRef);
            }
        } 
    }
    
    /**
     * If auto photometry, set aperture add to APERTURE_ADD
     */
    function setApertureAddAutoValue(){
        if (self.useAutoPhotometry){
            self.apertureAdd = APERTURE_ADD;
            if (nsgDialog){
                nsgDialog.apertureAdd_Control.setValue(self.apertureAdd);
            }
        } 
    }
    
    /**
     * If auto photometry, calc and set aperture growth rate
     */
    function setApertureGrowthRateAutoValue(){
        if (self.useAutoPhotometry && self.cache.getRefFilename()){
            let maxStarFlux = self.cache.getMaxStarFlux(self.logStarDetection);
            let limit = Math.round(calcDefaultGrowthLimit(self));
            limit = Math.max(limit, 2);
            self.apertureGrowthRate = calcStarGrowthRate(maxStarFlux, APERTURE_GROWTH, limit);
            if (nsgDialog){
                nsgDialog.apertureGrowthRate_Control.setValue(self.apertureGrowthRate);
            }
        } 
    }
    
    /**
     * If auto photometry, set outliers to 2% of matched stars, but don't exceed outlier control's range.
     */
    function setOutlierRemovalAutoValue(){
        if (self.useAutoPhotometry && self.cache.getRefFilename() && self.cache.getTgtFilename()){
            let starMatchArray = self.cache.getStarMatchArray(self);
            // 2% of stars, up to a max of 50
            let max = (new NsgPhotometryControls()).outlierRemoval.range.max;
            self.outlierRemoval = Math.floor(Math.min(max, starMatchArray.length / 50));
            if (nsgDialog){
                nsgDialog.outlierRemoval_Control.setValue(self.outlierRemoval);
            }
        }
    }
    
    /**
     * If auto photometry, set limit stars to 100%
     */
    function setLimitPhotoStarsPercentAutoValue(){
        if (self.useAutoPhotometry){
            self.limitPhotoStarsPercent = 100;
            if (nsgDialog){
                nsgDialog.limitPhotoStarsPercent_Control.setValue(self.limitPhotoStarsPercent);
            }
        } 
    }
    
    /**
     * @param {Boolean} checked If true disable controls and set them to calculated / default values
     * @param {Boolean} updateAll Force stars to be detected if they are not already cached.
     */
    this.setPhotometryAutoValues = function (checked, updateAll){
        this.useAutoPhotometry = checked;
        if (nsgDialog){
            nsgDialog.autoPhotometryCheckBox.checked = checked;
            nsgDialog.apertureAdd_Control.enabled = !checked;
            nsgDialog.apertureGrowthRate_Control.enabled = !checked;
            nsgDialog.apertureGap_Control.enabled = !checked;
            nsgDialog.apertureBgDelta_Control.enabled = !checked;
            nsgDialog.limitPhotoStarsPercent_Control.enabled = !checked;
            nsgDialog.outlierRemoval_Control.enabled = !checked;
            nsgDialog.linearRangeRef_Control.enabled = !checked;
        }
        if (checked){
            setApertureAddAutoValue();
            setLimitPhotoStarsPercentAutoValue();
            setApertureGapAutoValue();
            setApertureBgDeltaAutoValue();
            if (updateAll || this.cache.hasCachedRefStars(this.logStarDetection)){
                setApertureGrowthRateAutoValue();
                setLinearRangeAutoValue();
                setOutlierRemovalAutoValue();
            }
        }
    };
    
    /**
     * @param {Boolean} checked If true disable controls and set them to calculated / default values
     * @param {Boolean} updateAll Force stars to be detected if they are not already cached.
     */
    this.setSampleGenerationAutoValues = function (checked, updateAll){
        this.useAutoSampleGeneration = checked;
        if (nsgDialog){
            nsgDialog.autoSampleGenerationCheckBox.checked = this.useAutoSampleGeneration;
            nsgDialog.sampleStarGrowthRate_Control.enabled = !checked;
            nsgDialog.limitSampleStarsPercent_Control.enabled = !checked;
            nsgDialog.sampleSize_Control.enabled = !checked;
        }
        if (checked){
            setSampleSizeAutoValue();
            if (updateAll || this.cache.hasCachedRefStars(this.logStarDetection)){
                setLimitSampleStarsPercentAutoValue();
                setSampleStarGrowthRateAutoValue();
            }
        }
    };
    
    /**
     * If auto sample generation, set sample size based on image scale
     */
    function setSampleSizeAutoValue(){
        if (self.useAutoSampleGeneration && self.cache.getRefFilename()){
            let sampleControls = new NsgSampleControls();
            let lowerBound = sampleControls.sampleSize.range.min;
            let upperBound = sampleControls.sampleSize.range.max;
            let pixelAngle = calcDegreesPerPixel(self.pixelSize, self.focalLength);
            // Make sure we sample at least 100 x 100 microns on the sensor
            let minSampleSize = Math.round(100 / self.pixelSize);
            minSampleSize = Math.max(minSampleSize, lowerBound);
            // 0.005 deg = 18 arcsec (18 >> than than 4 arcsecond seeing)
            minSampleSize = Math.max(minSampleSize, Math.round(0.0075 / pixelAngle));
            // Try to have 4000 samples
            let w = self.cache.getRefView().image.width;
            let h = self.cache.getRefView().image.height;
            let size = Math.floor(Math.sqrt(w * h / 4500));
            size = Math.max(minSampleSize, size);
            size = Math.min(size, upperBound);
            self.sampleSize = size;
            if (nsgDialog){
                nsgDialog.sampleSize_Control.setValue(self.sampleSize);
            }
        } 
    }
    
    /**
     * If auto sample generation, limit 'grid rejection stars' to 10% of sample bins
     */
    function setLimitSampleStarsPercentAutoValue(){
        if (self.useAutoSampleGeneration && self.cache.getRefFilename()){
            let starCount = self.cache.getRefStars(self.logStarDetection).length;
            let binCount = self.cache.getSampleGridBinCount(self);
            let nStars = binCount / 10;
            self.limitSampleStarsPercent = Math.min(100, Math.round(10000 * nStars / starCount) / 100);
            if (nsgDialog){
                nsgDialog.limitSampleStarsPercent_Control.setValue(self.limitSampleStarsPercent);
            }
        }
    }
    
    /**
     * If auto sample generation, calculate star rejection circle growth rate.
     * This needs the ref stars to be detected (or cached) to determing max star flux.
     */
    function setSampleStarGrowthRateAutoValue(){
        if (self.useAutoSampleGeneration && self.cache.getRefFilename()){
            let maxStarFlux = self.cache.getMaxStarFlux(self.logStarDetection);
            let limit = Math.round(calcDefaultTargetGrowthLimit(self));
            limit = Math.max(limit, 2);
            self.sampleStarGrowthRate = calcStarGrowthRate(maxStarFlux, APERTURE_GROWTH_REJECTION, limit);
            if (nsgDialog){
                nsgDialog.sampleStarGrowthRate_Control.setValue(self.sampleStarGrowthRate);
            }
        }
    }
    
    /**
     * Used for process history and to populate the contents of a process icon.
     */
    this.saveParameters = function () {
        Parameters.clear();

        let referenceFile = this.cache.getRefFilename();
        if (referenceFile){
            Parameters.set("referenceFile", referenceFile);
        }
        
        if (this.targetFiles.length){
            for (let i=0; i < this.targetFiles.length; i++){
                Parameters.set("file" + i, this.targetFiles[i]);
            }
        }
        if (nsgDialog){
            let selectedTarget = nsgDialog.getSelectedTargetFilename(false);
            if (selectedTarget){
                Parameters.set("selectedTarget", selectedTarget);
            }
        }
        Parameters.set("pixelSize", this.pixelSize);
        Parameters.set("focalLength", this.focalLength);
        if (this.outputDir)
            Parameters.set("outputDir", this.outputDir);
        Parameters.set("outputPostFix", this.outputPostFix);
        Parameters.set("useFullPath", this.useFullPath);
        Parameters.set("overwrite", this.overwrite);
        Parameters.set("rescale", this.rescale);
        Parameters.set("displayGradient", this.displayGradient);
        Parameters.set("noiseWeightKeyword", this.noiseWeightKeyword);
        Parameters.set("weightPrefix", this.weightPrefix);
        Parameters.set("onErrorIndex", this.onErrorIndex);

        // Star Detection
        Parameters.set("logStarDetection", this.logStarDetection);
        
        // Photometric Star Search
        Parameters.set("starFluxTolerance", this.starFluxTolerance);
        Parameters.set("starSearchRadius", this.starSearchRadius);
        
        // Photometric Scale
        Parameters.set("apertureGrowthRate", this.apertureGrowthRate);
        Parameters.set("apertureAdd", this.apertureAdd);
        Parameters.set("apertureGap", this.apertureGap);
        Parameters.set("apertureBgDelta", this.apertureBgDelta);
        Parameters.set("limitPhotoStarsPercent", this.limitPhotoStarsPercent);
        Parameters.set("linearRangeRef", this.linearRangeRef);
        Parameters.set("outlierRemoval", this.outlierRemoval);
        Parameters.set("useAutoPhotometry", this.useAutoPhotometry);
        
        // Gradient Sample Generation
        Parameters.set("sampleStarGrowthRate", this.sampleStarGrowthRate);
        Parameters.set("limitSampleStarsPercent", this.limitSampleStarsPercent);
        Parameters.set("sampleSize", this.sampleSize);
        Parameters.set("maxSamples", this.maxSamples);
        Parameters.set("useAutoSampleGeneration", this.useAutoSampleGeneration);
        for (let i=0; i<this.manualRejectionCircles.length; i++){
            let mrc = this.manualRejectionCircles[i];
            Parameters.set("manualRejectionCircle" + i + "_x", mrc.x);
            Parameters.set("manualRejectionCircle" + i + "_y", mrc.y);
            Parameters.set("manualRejectionCircle" + i + "_radius", mrc.radius);
        }
        
        // Gradient Correction
        Parameters.set("gradientSmoothness", this.gradientSmoothness);
        Parameters.set("gradientLineX", this.gradientLineX);
        Parameters.set("gradientLineY", this.gradientLineY);
        Parameters.set("isGradientLineHorizontal", this.isGradientLineHorizontal);
        
        // ImageIntegration
        Parameters.set("imageIntegration", this.imageIntegration);
        Parameters.set("minimumWeight", this.minimumWeight);
        Parameters.set("sortByWeight", this.sortByWeight);
        Parameters.set("imageIntegrationTemplateId", this.imageIntegrationTemplateId);
        
        Parameters.set("smallScreen", this.smallScreen);
        Parameters.set("autoExit", this.autoExit);
        Parameters.set("graphWidth", this.graphWidth);
        Parameters.set("graphHeight", this.graphHeight);
        Parameters.set("extraControls", NSG_EXTRA_CONTROLS);
    };

    /**
     * Reload our script's data from a process icon
     */
    this.loadParameters = function () {
        if (Parameters.has("referenceFile")) {
            let filename = Parameters.getString("referenceFile");
            if (File.exists(filename)){
                this.cache.setRefFilename(filename);
            } else {
                console.warningln("** Warning: Reference file does not exist: '", filename, "'");
            }
        }
        let nthFile = 0;
        this.targetFiles = [];
        while (Parameters.has("file" + nthFile)){
            let filename = Parameters.getString("file" + nthFile);
            if (filename && File.exists(filename)){
                this.targetFiles.push(filename);
            } else {
                console.warningln("** Warning: Target file does not exist: '", filename, "'");
            }
            nthFile++;
        }
        if (!this.targetFiles.length){
            // backward compatability
            if (Parameters.has("targetFiles")) {
                let files = (Parameters.getString("targetFiles")).split(',');
                this.targetFiles = [];
                for (let filename of files){
                    if (File.exists(filename)){
                        this.targetFiles.push(filename);
                    } else {
                        console.warningln("Target file does not exist: '", filename, "'");
                    }
                }
            }
        }
        if (Parameters.has("selectedTarget")){
            this.savedSelectedTarget = Parameters.getString("selectedTarget");
        }
        if (Parameters.has("pixelSize")) {
            this.pixelSize = (Parameters.getReal("pixelSize"));
        }
        if (Parameters.has("focalLength")) {
            this.focalLength = (Parameters.getInteger("focalLength"));
        }
        if (Parameters.has("outputDir")) {
            let dir = Parameters.getString("outputDir");
            if (dir && File.directoryExists(dir)){
                this.outputDir = dir;
            } else {
                console.warningln("** Warning: Output directory does not exist: '", dir, "'");
            }
        }
        if (Parameters.has("outputPostFix")) {
            this.outputPostFix = (Parameters.getString("outputPostFix"));
        }
        if (Parameters.has("useFullPath")) {
            this.useFullPath = (Parameters.getBoolean("useFullPath"));
        }
        if (Parameters.has("overwrite")) {
            this.overwrite = (Parameters.getBoolean("overwrite"));
        }
        if (Parameters.has("rescale")) {
            this.rescale = (Parameters.getBoolean("rescale"));
        }
        if (Parameters.has("displayGradient")) {
            this.displayGradient = (Parameters.getBoolean("displayGradient"));
        }
        if (Parameters.has("weightPrefix")) {
            this.weightPrefix = (Parameters.getBoolean("weightPrefix"));
        }
        if (Parameters.has("noiseWeightKeyword")) {
            this.noiseWeightKeyword = (Parameters.getString("noiseWeightKeyword"));
        }
        if (Parameters.has("onErrorIndex")) {
            this.onErrorIndex = (Parameters.getInteger("onErrorIndex"));
        }

        // Star Detection
        if (Parameters.has("logStarDetection"))
            this.logStarDetection = Parameters.getReal("logStarDetection");
        
        // Photometric Star Search
        if (Parameters.has("starFluxTolerance"))
            this.starFluxTolerance = Parameters.getReal("starFluxTolerance");
        if (Parameters.has("starSearchRadius"))
            this.starSearchRadius = Parameters.getReal("starSearchRadius");
        
        // Photometric Scale
        if (Parameters.has("apertureGrowthRate"))
            this.apertureGrowthRate = Parameters.getReal("apertureGrowthRate");
        if (Parameters.has("apertureAdd"))
            this.apertureAdd = Parameters.getInteger("apertureAdd");
        if (Parameters.has("apertureGap"))
            this.apertureGap = Parameters.getInteger("apertureGap");
        if (Parameters.has("apertureBgDelta"))
            this.apertureBgDelta = Parameters.getInteger("apertureBgDelta");
        if (Parameters.has("limitPhotoStarsPercent"))
            this.limitPhotoStarsPercent = Parameters.getReal("limitPhotoStarsPercent");
        if (Parameters.has("linearRangeRef"))
            this.linearRangeRef = Parameters.getReal("linearRangeRef");
        if (Parameters.has("outlierRemoval"))
            this.outlierRemoval = Parameters.getInteger("outlierRemoval");
        if (Parameters.has("useAutoPhotometry"))
            this.useAutoPhotometry = Parameters.getBoolean("useAutoPhotometry");
        
        // Gradient Sample Generation
        if (Parameters.has("sampleStarGrowthRate"))
            this.sampleStarGrowthRate = Parameters.getReal("sampleStarGrowthRate");
        if (Parameters.has("limitSampleStarsPercent"))
            this.limitSampleStarsPercent = Parameters.getReal("limitSampleStarsPercent");
        if (Parameters.has("sampleSize"))
            this.sampleSize = Parameters.getInteger("sampleSize");
        if (Parameters.has("maxSamples"))
            this.maxSamples = Parameters.getInteger("maxSamples");
        if (Parameters.has("useAutoSampleGeneration"))
            this.useAutoSampleGeneration = Parameters.getBoolean("useAutoSampleGeneration");
        for (let i=0; ; i++){
            if (Parameters.has("manualRejectionCircle" + i + "_x") &&
                    Parameters.has("manualRejectionCircle" + i + "_y") &&
                    Parameters.has("manualRejectionCircle" + i + "_radius")){
                
                let x = Parameters.getReal("manualRejectionCircle" + i + "_x");
                let y = Parameters.getReal("manualRejectionCircle" + i + "_y");
                let r = Parameters.getInteger("manualRejectionCircle" + i + "_radius");
                this.manualRejectionCircles.push(new ManualRejectionCircle(x, y, r));
            } else {
                break;
            }
        }
        
        // Gradient Correction
        if (Parameters.has("gradientSmoothness"))
            this.gradientSmoothness = Parameters.getReal("gradientSmoothness");
        if (Parameters.has("gradientLineX"))
            this.gradientLineX = Parameters.getInteger("gradientLineX");
        if (Parameters.has("gradientLineY"))
            this.gradientLineY = Parameters.getInteger("gradientLineY");
        if (Parameters.has("isGradientLineHorizontal"))
            this.isGradientLineHorizontal = Parameters.getBoolean("isGradientLineHorizontal");
                
        // ImageIntegration
        if (Parameters.has("minimumWeight"))
            this.minimumWeight = Parameters.getReal("minimumWeight");
        if (Parameters.has("sortByWeight")) {
            this.sortByWeight = (Parameters.getBoolean("sortByWeight"));
        }
        if (Parameters.has("imageIntegrationTemplateId"))
            this.imageIntegrationTemplateId = Parameters.getString("imageIntegrationTemplateId");
        
        if (Parameters.has("smallScreen"))
            this.smallScreen = Parameters.getBoolean("smallScreen");
        if (Parameters.has("autoExit"))
            this.autoExit = Parameters.getBoolean("autoExit");
        
        if (Parameters.has("imageIntegration"))
            this.imageIntegration = Parameters.getBoolean("imageIntegration");
        if (Parameters.has("graphWidth"))
            this.graphWidth = Parameters.getInteger("graphWidth");
        if (Parameters.has("graphHeight"))
            this.graphHeight = Parameters.getInteger("graphHeight");
        
        if (Parameters.has("extraControls")){
            NSG_EXTRA_CONTROLS = Parameters.getBoolean("extraControls");
        }

    };

    /**
     * Initialise the scripts data
     */
    this.setParameters = function () {
        this.pixelSize = DEFAULT_PIXEL_SIZE;
        this.focalLength = DEFAULT_FOCAL_LENGTH;
        this.targetFiles = [];
        this.savedSelectedTarget = "";
        this.outputDir = "";
        this.outputPostFix = "_nsg";
        this.useFullPath = false;
        this.overwrite = false;
        this.weightPrefix = true;
        this.rescale = true;
        this.noiseWeightKeyword = "NWEIGHT";
        this.onErrorIndex = 0;
        
        // Star Detection
        this.logStarDetection = -1;
        
        // Photometric Star Search
        this.starFluxTolerance = 2.0;
        this.starSearchRadius = 2.0;
        
        // Photometric Scale
        this.apertureGrowthRate = APERTURE_GROWTH;
        this.apertureAdd = APERTURE_ADD;
        this.apertureGap = APERTURE_GAP;
        this.apertureBgDelta = APERTURE_BKG_DELTA;
        this.limitPhotoStarsPercent = 100;
        this.outlierRemoval = 0;
        this.useAutoPhotometry = true;
        
        // Gradient Sample Generation
        this.sampleStarGrowthRate = APERTURE_GROWTH_REJECTION;
        this.limitSampleStarsPercent = 35;
        this.sampleSize = 20;
        this.maxSamples = 3000;
        this.useAutoSampleGeneration = true;
        this.manualRejectionCircles = [];
        
        // Gradient Correction
        this.gradientSmoothness = 2.0;
        this.gradientLineX = 0;
        this.gradientLineY = 0;
        this.isGradientLineHorizontal = true;
        
        // ImageIntegration
        this.imageIntegration = true;
        this.minimumWeight = DEFAULT_INT_MIN_WEIGHT;
        this.sortByWeight = true;
        this.imageIntegrationTemplateId = "";
        
        this.graphWidth = 1200; // gradient and photometry graph width
        this.graphHeight = 800; // gradient and photometry graph height
        
        this.displayGradient = false;
        this.smallScreen = false;
        this.autoExit = false;
        
        if (this.cache !== undefined){
            this.cache.invalidate();
        }
        this.cache = new Cache();
        this.linearRangeRef = LINEAR_RANGE;
        
        this.viewFlag = 0;
    };
    
    this.saveSettings = function(){
        self.resetSettings();
        let referenceFile = this.cache.getRefFilename();
        if (referenceFile){
            Settings.write( KEYPREFIX+"/referenceFile", DataType_UCString, referenceFile);
        }
        if (this.targetFiles.length){
            for (let i=0; i < this.targetFiles.length; i++){
                Settings.write( KEYPREFIX+"/file" + i, DataType_UCString, this.targetFiles[i]);
            }
        }
        if (nsgDialog){
            let selectedTarget = nsgDialog.getSelectedTargetFilename(false);
            if (selectedTarget){
                Settings.write( KEYPREFIX+"/selectedTarget", DataType_UCString, selectedTarget);
            }
        }
        
        Settings.write( KEYPREFIX+"/pixelSize", DataType_Float, this.pixelSize );
        Settings.write( KEYPREFIX+"/focalLength", DataType_Int32, this.focalLength );
        
        // Output files
        if (this.outputDir)
            Settings.write( KEYPREFIX+"/outputDir", DataType_UCString, this.outputDir);
        Settings.write( KEYPREFIX+"/outputPostFix", DataType_String, this.outputPostFix );
        Settings.write( KEYPREFIX+"/useFullPath", DataType_Boolean, this.useFullPath );
        Settings.write( KEYPREFIX+"/overwrite", DataType_Boolean, this.overwrite );
        Settings.write( KEYPREFIX+"/rescale", DataType_Boolean, this.rescale );
        Settings.write( KEYPREFIX+"/displayGradient", DataType_Boolean, this.displayGradient );
        Settings.write( KEYPREFIX+"/noiseWeightKeyword", DataType_String, this.noiseWeightKeyword );
        Settings.write( KEYPREFIX+"/weightPrefix", DataType_Boolean, this.weightPrefix );
        Settings.write( KEYPREFIX+"/onErrorIndex", DataType_Int32, this.onErrorIndex );

        // Star Detection
        Settings.write( KEYPREFIX+"/logStarDetection", DataType_Float, this.logStarDetection );

        // Photometric Star Search
        Settings.write( KEYPREFIX+"/starFluxTolerance", DataType_Float, this.starFluxTolerance );
        Settings.write( KEYPREFIX+"/starSearchRadius", DataType_Float, this.starSearchRadius );

        // Photometric Scale
        Settings.write( KEYPREFIX+"/apertureGrowthRate", DataType_Float, this.apertureGrowthRate );
        Settings.write( KEYPREFIX+"/apertureAdd", DataType_Int32, this.apertureAdd );
        Settings.write( KEYPREFIX+"/apertureGap", DataType_Int32, this.apertureGap );
        Settings.write( KEYPREFIX+"/apertureBgDelta", DataType_Int32, this.apertureBgDelta );
        Settings.write( KEYPREFIX+"/limitPhotoStarsPercent", DataType_Float, this.limitPhotoStarsPercent );
        Settings.write( KEYPREFIX+"/linearRangeRef", DataType_Float, this.linearRangeRef );
        Settings.write( KEYPREFIX+"/outlierRemoval", DataType_Int32, this.outlierRemoval );
        Settings.write( KEYPREFIX+"/useAutoPhotometry", DataType_Boolean, this.useAutoPhotometry );

        // Gradient Sample Generation
        Settings.write( KEYPREFIX+"/sampleStarGrowthRate", DataType_Float, this.sampleStarGrowthRate );
        Settings.write( KEYPREFIX+"/limitSampleStarsPercent", DataType_Float, this.limitSampleStarsPercent );
        Settings.write( KEYPREFIX+"/sampleSize", DataType_Int32, this.sampleSize );
        Settings.write( KEYPREFIX+"/extraControls", DataType_Boolean, NSG_EXTRA_CONTROLS );
        if (NSG_EXTRA_CONTROLS){
            Settings.write( KEYPREFIX+"/maxSamples", DataType_Int32, this.maxSamples );
        }
        Settings.write( KEYPREFIX+"/useAutoSampleGeneration", DataType_Boolean, this.useAutoSampleGeneration );
        for (let i=0; i<this.manualRejectionCircles.length; i++){
            let mrc = this.manualRejectionCircles[i];
            Settings.write( KEYPREFIX+"/manualRejectionCircle" + i + "_x", DataType_Float, mrc.x);
            Settings.write( KEYPREFIX+"/manualRejectionCircle" + i + "_y", DataType_Float, mrc.y);
            Settings.write( KEYPREFIX+"/manualRejectionCircle" + i + "_radius", DataType_Int32, mrc.radius);
        }

        // Gradient correction
        Settings.write( KEYPREFIX+"/gradientSmoothness", DataType_Float, this.gradientSmoothness );
        Settings.write( KEYPREFIX+"/gradientLineX", DataType_Int32, this.gradientLineX );
        Settings.write( KEYPREFIX+"/gradientLineY", DataType_Int32, this.gradientLineY );
        Settings.write( KEYPREFIX+"/isGradientLineHorizontal", DataType_Boolean, this.isGradientLineHorizontal );
        
        // ImageIntegration
        Settings.write( KEYPREFIX+"/imageIntegration", DataType_Boolean, this.imageIntegration );
        Settings.write( KEYPREFIX+"/minimumWeight", DataType_Float, this.minimumWeight );
        Settings.write( KEYPREFIX+"/sortByWeight", DataType_Boolean, this.sortByWeight );
        Settings.write( KEYPREFIX+"/imageIntegrationTemplateId", DataType_UCString, this.imageIntegrationTemplateId );

        Settings.write( KEYPREFIX+"/smallScreen", DataType_Boolean, this.smallScreen );
        Settings.write( KEYPREFIX+"/autoExit", DataType_Boolean, this.autoExit );
    };
    
    this.restoreSettings = function(){
        let keyValue = Settings.read( KEYPREFIX+"/referenceFile", DataType_UCString );
        if ( Settings.lastReadOK ){
            if (keyValue && File.exists(keyValue)){
                this.cache.setRefFilename(keyValue);
            }
        }
    
        let nthFile = 0;
        this.targetFiles = [];
        do {
            keyValue = Settings.read( KEYPREFIX+"/file" + nthFile, DataType_UCString );
            if ( Settings.lastReadOK ){
                if (keyValue && File.exists(keyValue)){
                    this.targetFiles.push(keyValue);
                }
            }
            nthFile++;
        } while (Settings.lastReadOK);
        
        keyValue = Settings.read( KEYPREFIX+"/selectedTarget", DataType_UCString );
        if ( Settings.lastReadOK ){
            if (keyValue && File.exists(keyValue)){
                this.savedSelectedTarget = keyValue;
            }
        }
        
        keyValue = Settings.read( KEYPREFIX+"/pixelSize", DataType_Float );
        if ( Settings.lastReadOK )
            this.pixelSize = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/focalLength", DataType_Int32 );
        if ( Settings.lastReadOK )
            this.focalLength = keyValue;
        
        // Output files
        keyValue = Settings.read( KEYPREFIX+"/outputDir", DataType_UCString );
        if ( Settings.lastReadOK ){
            if (keyValue && File.directoryExists(keyValue)){
                this.outputDir = keyValue;
            }
        }
        keyValue = Settings.read( KEYPREFIX+"/outputPostFix", DataType_String );
        if ( Settings.lastReadOK )
            this.outputPostFix = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/useFullPath", DataType_Boolean );
        if ( Settings.lastReadOK )
            this.useFullPath = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/overwrite", DataType_Boolean );
        if ( Settings.lastReadOK )
            this.overwrite = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/rescale", DataType_Boolean );
        if ( Settings.lastReadOK )
            this.rescale = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/displayGradient", DataType_Boolean );
        if ( Settings.lastReadOK )
            this.displayGradient = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/noiseWeightKeyword", DataType_String );
        if ( Settings.lastReadOK )
            this.noiseWeightKeyword = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/weightPrefix", DataType_Boolean );
        if ( Settings.lastReadOK )
            this.weightPrefix = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/onErrorIndex", DataType_Int32 );
        if ( Settings.lastReadOK )
            this.onErrorIndex = keyValue;
        
        // Star Detection
        keyValue = Settings.read( KEYPREFIX+"/logStarDetection", DataType_Float );
        if ( Settings.lastReadOK )
            this.logStarDetection = keyValue;
        
        // Photometric Star Search
        keyValue = Settings.read( KEYPREFIX+"/starFluxTolerance", DataType_Float );
        if ( Settings.lastReadOK )
            this.starFluxTolerance = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/starSearchRadius", DataType_Float );
        if ( Settings.lastReadOK )
            this.starSearchRadius = keyValue;
        
        // Photometric Scale
        keyValue = Settings.read( KEYPREFIX+"/apertureGrowthRate", DataType_Float );
        if ( Settings.lastReadOK )
            this.apertureGrowthRate = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/apertureAdd", DataType_Int32 );
        if ( Settings.lastReadOK )
            this.apertureAdd = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/apertureGap", DataType_Int32 );
        if ( Settings.lastReadOK )
            this.apertureGap = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/apertureBgDelta", DataType_Int32 );
        if ( Settings.lastReadOK )
            this.apertureBgDelta = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/limitPhotoStarsPercent", DataType_Float );
        if ( Settings.lastReadOK )
            this.limitPhotoStarsPercent = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/linearRangeRef", DataType_Float );
        if ( Settings.lastReadOK )
            this.linearRangeRef = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/outlierRemoval", DataType_Int32 );
        if ( Settings.lastReadOK )
            this.outlierRemoval = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/useAutoPhotometry", DataType_Boolean );
        if ( Settings.lastReadOK )
            this.useAutoPhotometry = keyValue;
        
        // Gradient Sample Generation
        keyValue = Settings.read( KEYPREFIX+"/sampleStarGrowthRate", DataType_Float );
        if ( Settings.lastReadOK )
            this.sampleStarGrowthRate = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/sampleStarGrowthRateTarget", DataType_Float );
        if ( Settings.lastReadOK )
            this.sampleStarGrowthRateTarget = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/limitSampleStarsPercent", DataType_Float );
        if ( Settings.lastReadOK )
            this.limitSampleStarsPercent = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/sampleSize", DataType_Int32 );
        if ( Settings.lastReadOK )
            this.sampleSize = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/extraControls", DataType_Boolean );
        if ( Settings.lastReadOK )
            NSG_EXTRA_CONTROLS = keyValue;
        if (NSG_EXTRA_CONTROLS){
            keyValue = Settings.read( KEYPREFIX+"/maxSamples", DataType_Int32 );
            if ( Settings.lastReadOK )
                this.maxSamples = keyValue;
        }
        keyValue = Settings.read( KEYPREFIX+"/useAutoSampleGeneration", DataType_Boolean );
        if ( Settings.lastReadOK )
            this.useAutoSampleGeneration = keyValue;
        for (let i=0; ; i++){
            let x, y, radius;
            keyValue = Settings.read( KEYPREFIX+"/manualRejectionCircle" + i + "_x", DataType_Float );
            if ( Settings.lastReadOK )
                x = keyValue;
            keyValue = Settings.read( KEYPREFIX+"/manualRejectionCircle" + i + "_y", DataType_Float );
            if ( Settings.lastReadOK )
                y = keyValue;
            keyValue = Settings.read( KEYPREFIX+"/manualRejectionCircle" + i + "_Radius", DataType_Int32 );
            if ( Settings.lastReadOK )
                radius = keyValue;

            if (x && y && radius){
                this.manualRejectionCircles.push(new ManualRejectionCircle(x, y, radius));
            } else {
                break;
            }
        }
        
        // Gradient correction
        keyValue = Settings.read( KEYPREFIX+"/gradientSmoothness", DataType_Float );
        if ( Settings.lastReadOK )
            this.gradientSmoothness = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/gradientLineX", DataType_Int32 );
        if ( Settings.lastReadOK )
            this.gradientLineX = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/gradientLineY", DataType_Int32 );
        if ( Settings.lastReadOK )
            this.gradientLineY = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/isGradientLineHorizontal", DataType_Boolean );
        if ( Settings.lastReadOK )
            this.isGradientLineHorizontal = keyValue;
        
        // ImageIntegration
        keyValue = Settings.read( KEYPREFIX+"/imageIntegration", DataType_Boolean );
        if ( Settings.lastReadOK )
            this.imageIntegration = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/minimumWeight", DataType_Float );
        if ( Settings.lastReadOK )
            this.minimumWeight = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/sortByWeight", DataType_Boolean );
        if ( Settings.lastReadOK )
            this.sortByWeight = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/imageIntegrationTemplateId", DataType_UCString );
        if ( Settings.lastReadOK )
            this.imageIntegrationTemplateId = keyValue;
        
        keyValue = Settings.read( KEYPREFIX+"/smallScreen", DataType_Boolean );
        if ( Settings.lastReadOK )
            this.smallScreen = keyValue;
        keyValue = Settings.read( KEYPREFIX+"/autoExit", DataType_Boolean );
        if ( Settings.lastReadOK )
            this.autoExit = keyValue;
        
    };
    
    this.resetSettings = function(){
        Settings.remove( KEYPREFIX );
    };
    
    this.setParameters();
}
