/* global Dialog, FrameStyle_Sunken, VERSION, TITLE, TextAlign_Right, TextAlign_VertCenter, APERTURE_GROWTH_REJECTION, APERTURE_ADD, APERTURE_GROWTH, DataType_Boolean, StdIcon_Question, StdButton_Ok, StdButton_Cancel, Settings, KEYPREFIX, StdIcon_Error, StdDialogCode_Ok, REF_TEXT_COLOR, Parameters, ProcessInstance, File, DEFAULT_PIXEL_SIZE, DEFAULT_FOCAL_LENGTH, DEFAULT_INT_MIN_WEIGHT, COL_FULL_FILENAME, COL_FILENAME, COL_ALT, COL_EXPOSURE, COL_FILTER, COL_AIRMASS, COL_PSFSGL, COL_NOISE */

// Version 1.0 (c) John Murphy 5th-Apr-2021
//
// ======== #license ===============================================================
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, version 3 of the License.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.
// =================================================================================

#include <pjsr/DataType.jsh>
#include <pjsr/StdDialogCode.jsh>
#include <pjsr/FrameStyle.jsh>
#include <pjsr/NumericControl.jsh>
#include <pjsr/Sizer.jsh>
#include <pjsr/TextAlign.jsh>
#include <pjsr/StdButton.jsh>
#include <pjsr/StdIcon.jsh>
#include <pjsr/StdCursor.jsh>
#include <pjsr/SectionBar.jsh>
#include "ImageScaleDialog.js"
#include "HelpDialog.js"

#define REF_TEXT_COLOR 0xFF55AA55
#define COL_FULL_FILENAME 0
#define COL_FILENAME 1
#define COL_PSFSGL 2
#define COL_NOISE 3
#define COL_ALT 4
#define COL_AIRMASS 5
#define COL_EXPOSURE 6
#define COL_FILTER 7

let NSG_EXTRA_CONTROLS = false;

function NsgDialog(data){
    this.__base__ = Dialog;
    this.__base__();

    let self = this;
    
    this.onToggleSection = function(bar, beginToggle){
        if (beginToggle){
            this.dialog.setVariableSize();
        } else {
//            bar.updateSection();
            this.dialog.setFixedSize();
        }
    };
    
    // =======================================
    // SectionBar: "Quick Start Guide"
    // =======================================
    // Create the Program Description at the top
    let titleLabel = new Label(this);
    titleLabel.frameStyle = FrameStyle_Sunken;
    titleLabel.margin = 4;
    titleLabel.wordWrapping = true;
    titleLabel.useRichText = true;
    titleLabel.text = "<b>Normalizes the target images to the reference. " +
        "Applies a scale factor and removes the relative gradient.</b><br />" +
        "Use on registered images before stacking.<br />" +
        "(1) Read help sections: <i>Prerequisites</i>, <i>Quick Start Guide</i>.<br />" +
        "(2) After 27,000 lines of code, I would be extremely grateful for a coffee <b><u>https://ko-fi.com/jmurphy</u></b> Thanks!<br />" +
        "Copyright &copy; 2019-2021 John Murphy. <i>I would also like to thank Adam Block for his advice and ideas.</i>";
    titleLabel.toolTip = "https://ko-fi.com/jmurphy";
    titleLabel.onMousePress = function( x, y, button, buttonState, modifiers ){
        (new HelpDialog()).execute();
    };
    let titleSection = new Control(this);
    titleSection.sizer = new VerticalSizer;
    titleSection.sizer.add(titleLabel);
    let titleBar = new SectionBar(this, "Normalize Scale Gradient V" + VERSION);
    titleBar.setSection(titleSection);
    titleBar.onToggleSection = this.onToggleSection;
    // SectionBar "Quick Start Guide" End
    
    // =======================================
    // SectionBar: "Reference Image"
    // =======================================
    this.refTextBox = new Edit( this );
    this.refTextBox.onEditCompleted = function(){
        // Do not allow the user to edit this field
        let refFilename = data.cache.getRefFilename();
        self.refTextBox.text = refFilename ? refFilename : "";
    };
    
    function setReferenceFilename(filename){
        let hdrEntries = self.getHdrEntries(filename);
        if (hdrEntries.exception){
            (new MessageBox("Failed to read '" + filename + "'")).execute();
            self.refTextBox.text = data.cache.getRefFilename();
            return;
        }
        if (!hdrEntries.pixelSize || !hdrEntries.focalLength){
            let imageScaleDialog = new ImageScaleDialog(hdrEntries, KEYPREFIX + "Save", false );
            imageScaleDialog.execute();
            hdrEntries.pixelSize = imageScaleDialog.getPixelSize();
            hdrEntries.focalLength = imageScaleDialog.getFocalLength();
        }
        self.enabled = false;
        processEvents();
        self.setPixelScaleFields(hdrEntries.pixelSize, hdrEntries.focalLength, hdrEntries.filter);
        data.setRefFilename(filename);
        self.refTextBox.text = data.cache.getRefFilename();
        self.updateTargetTextColor(data);
        data.manualRejectionCircles = [];
        data.gradientLineX = 0;
        data.gradientLineY = 0;
        data.isGradientLineHorizontal = true;
        self.enabled = true;
        processEvents();
    }
    
    /**
     * Reads PixelSize, FocalLength and Filter from FITS header.
     * If PixelSize or FocalLength are not found, they are set to undefined.
     * If Filter is not found it is set to 'NONE'.
     * @param {String} filename
     * @returns {pixelSize, focalLength, filter}
     */
    this.getHdrEntries = function(filename){
        let pixelSize;
        let focalLength;
        let filter = "NONE";
        let exception;
        try {
            let keywords = getFITSKeywords( filename );
            let fitsHdr = new FitsHeader();
            pixelSize = fitsHdr.getPixelSize(keywords);
            focalLength = fitsHdr.getFocalLength(keywords);
            filter = fitsHdr.getFilter(keywords);
        } catch (e){
            console.criticalln(e);
            exception = e;
        }        
        return {pixelSize: pixelSize, focalLength: focalLength, filter: filter, exception: exception};
    };
    
    /**
     * @param {Number|undefined} pixelSize
     * @param {Number|undefined} focalLength
     * @param {String} filterName
     */
    this.setPixelScaleFields = function(pixelSize, focalLength, filterName){
        data.pixelSize = pixelSize ? pixelSize : DEFAULT_PIXEL_SIZE;
        data.focalLength = focalLength ? focalLength : DEFAULT_FOCAL_LENGTH;
        pixelSizeControl.setValue(data.pixelSize);
        focalLengthControl.setValue(data.focalLength);
        if (filterName)
            filterTextBox.text = filterName;
        pixelScaleControl.setValue(calcDegreesPerPixel(data.pixelSize, data.focalLength) * 3600);
        processEvents();
    };
    
    let setRef_Button = new PushButton( this );
    setRef_Button.text = "Set reference";
    setRef_Button.toolTip = "<p>Set the reference to the selected target image.</p>";
    setRef_Button.onClick = function(){
        let filename = self.getSelectedTargetFilename(true);
        if (filename){
            setReferenceFilename(filename);
            let nodes = self.files_TreeBox.selectedNodes;
            nodes[0].selected = false;
            console.noteln("Loaded reference image. Ready...");
        }
    };
    
    let refFileSizer = new HorizontalSizer(this);
    refFileSizer.spacing = 4;
    refFileSizer.add(this.refTextBox, 100);
    refFileSizer.add(setRef_Button);
    
    let focalLengthControl = new NumericEdit(this);
    focalLengthControl.real = false;
    focalLengthControl.label.text = "Focal length:";
    focalLengthControl.toolTip = "Focal length in mm";
    focalLengthControl.setRange(0, 99999);
    focalLengthControl.enabled = false;
    
    let pixelSizeControl = new NumericEdit(this);
    pixelSizeControl.real = true;
    pixelSizeControl.label.text = "Pixel size:";
    pixelSizeControl.toolTip = "Pixel size, including binning, in microns";
    pixelSizeControl.setRange(0, 999);
    pixelSizeControl.setPrecision(2);
    pixelSizeControl.enabled = false;
    
    let pixelScaleControl = new NumericEdit(this);
    pixelScaleControl.real = true;
    pixelScaleControl.label.text = "Pixel scale:";
    pixelScaleControl.toolTip = "Pixel scale in arcseconds";
    pixelScaleControl.setRange(0, 999);
    pixelScaleControl.setPrecision(2);
    pixelScaleControl.enabled = false;
    
    let filterTextBox = new Edit( this );
    filterTextBox.setFixedWidth(this.font.width("XXXXXXXXXXXXXXXXX"));
    filterTextBox.readOnly = true;
    filterTextBox.toolTip = "<p>Reference filter.</p>" +
            "<p>All target images must match the reference filter.</p>";
    
    let imageScaleButton = new PushButton(this);
    imageScaleButton.text = "Image scale";
    imageScaleButton.toolTip =
            "<p>Manually set the focal length and pixel size.</p>";
    imageScaleButton.onClick = function () {
        let hdrEntries = {pixelSize: data.pixelSize, focalLength: data.focalLength};
        let imageScaleDialog = new ImageScaleDialog( hdrEntries, KEYPREFIX + "Save", true);
        if (StdDialogCode_Ok === imageScaleDialog.execute()){
            data.pixelSize = imageScaleDialog.getPixelSize();
            data.focalLength = imageScaleDialog.getFocalLength();
            self.setPixelScaleFields(data.pixelSize, data.focalLength, undefined);
            data.setPhotometryAutoValues(data.useAutoPhotometry, true);
            data.setSampleGenerationAutoValues(data.useAutoSampleGeneration, true);
        }
    };
    
    let filterGroupBox = new GroupBox(this);
    filterGroupBox.title = "Filter";
    filterGroupBox.sizer = new HorizontalSizer;
    filterGroupBox.sizer.margin = 2;
    filterGroupBox.sizer.add(filterTextBox);
    
    let imageScaleGroupBox = new GroupBox(this);
    imageScaleGroupBox.title = "Image scale";
    imageScaleGroupBox.sizer = new HorizontalSizer;
    imageScaleGroupBox.sizer.margin = 2;
    imageScaleGroupBox.sizer.add(focalLengthControl);
    imageScaleGroupBox.sizer.addSpacing(10);
    imageScaleGroupBox.sizer.add(pixelSizeControl);
    imageScaleGroupBox.sizer.addSpacing(10);
    imageScaleGroupBox.sizer.add(pixelScaleControl);
    imageScaleGroupBox.sizer.addStretch();
    imageScaleGroupBox.sizer.add(imageScaleButton);
    imageScaleGroupBox.sizer.addSpacing(2);
    
    let filterScaleSizer = new HorizontalSizer;
    filterScaleSizer.spacing = 12;
    filterScaleSizer.add(filterGroupBox);
    filterScaleSizer.add(imageScaleGroupBox, 100);
    
    let refSection = new Control(this);
    refSection.sizer = new VerticalSizer;
    refSection.sizer.spacing = 4;
    refSection.sizer.add(refFileSizer);
    refSection.sizer.add(filterScaleSizer);
    let refBar = new SectionBar(this, "Reference Image");
    refBar.setSection(refSection);
    refBar.onToggleSection = this.onToggleSection;
    // SectionBar "Reference file" End
    
    // =======================================
    // SectionBar: "Target Images"
    // =======================================
    this.files_TreeBox = new TreeBox( this );
    this.files_TreeBox.multipleSelection = true;
    this.files_TreeBox.rootDecoration = false;
    this.files_TreeBox.alternateRowColor = true;
    this.files_TreeBox.setScaledMinSize( 825, 300 );
    this.files_TreeBox.headerVisible = true;
    this.files_TreeBox.headerSorting = true;
    this.files_TreeBox.horizontalAutoScroll = true;
    this.files_TreeBox.setHeaderText( COL_FULL_FILENAME, "FullFilename" );
    this.files_TreeBox.hideColumn( COL_FULL_FILENAME, true );
    this.files_TreeBox.setHeaderText( COL_FILENAME, "Filename" );
    this.files_TreeBox.setHeaderText( COL_PSFSGL, "PSF Weight" );
    this.files_TreeBox.setHeaderText( COL_NOISE, "Noise" );
    this.files_TreeBox.setHeaderText( COL_ALT, "ALT" );
    this.files_TreeBox.setHeaderText( COL_AIRMASS, "Airmass" );
    this.files_TreeBox.setHeaderText( COL_EXPOSURE, "Time" );
    this.files_TreeBox.setHeaderText( COL_FILTER, "Filter" );
    this.files_TreeBox.onNodeDoubleClicked = function ( treeBoxNode, columnIndex ){
        let filename = treeBoxNode.text(COL_FULL_FILENAME)
        console.noteln("Opening ", filename);
        let imageWindows = ImageWindow.open(filename, "view", "no-warnings verbosity 0");
        let imageWindow = imageWindows[0];
        let view = imageWindow.mainView;
        let stf = STFAutoStretch(view, undefined, undefined, false);
        var HT = new HistogramTransformation;
        if (view.image.isColor){
            HT.H = 
                [[stf[0][0], stf[0][2], stf[0][1], 0, 1],
                [stf[1][0], stf[1][2], stf[1][1], 0, 1],
                [stf[2][0], stf[2][2], stf[2][1], 0, 1],
                [0, 0.5, 1, 0, 1],
                [0, 0.5, 1, 0, 1]];
        } else {
            HT.H = 
                [[0, 0.5, 1, 0, 1],
                [0, 0.5, 1, 0, 1],
                [0, 0.5, 1, 0, 1],
                [stf[0][0], stf[0][2], stf[0][1], 0, 1],
                [0, 0.5, 1, 0, 1]];
        }
        HT.executeOn(view, false); // no swap file
        imageWindow.zoomToFit();
        console.hide();
        imageWindow.show();
        imageWindow.bringToFront();
    };
    
    /**
     * Sets reference text color to green. All other rows use black.
     * @param {NsgData} data
     */
    this.updateTargetTextColor = function(data){
        let fitsHeader = new FitsHeader();
        let refFilename = data.cache.getRefFilename();
        let refFilter = filterTextBox.text;
        for ( let i = 0; i < self.files_TreeBox.numberOfChildren; ++i ){
            let node = self.files_TreeBox.child( i );
            let font = node.font( COL_FILENAME );
            let filename = node.text( COL_FULL_FILENAME );
            let filter;
            try {
                let keywords = getFITSKeywords( filename );
                filter = fitsHeader.getFilter(keywords);
            } catch (exception){
                filter = "ERROR";
            }
            let color;
            if (filename === refFilename){
                font.italic = true;
                color = REF_TEXT_COLOR;
            } else if (refFilename && filter !== refFilter){
                font.italic = false;
                color = 0xFFFF0000; 
            } else {
                font.italic = false;
                color = 0xFF000000; 
            }
            let nColumns = self.files_TreeBox.numberOfColumns;
            for (let j = 0; j < nColumns; j++){
                node.setFont( j, font );
                node.setTextColor(j, color); 
            }
        }
        adjustFilenameColumnWidth();
    };
    
    this.fullPathCheckBox = new CheckBox();
    this.fullPathCheckBox.text = "Full paths";
    this.fullPathCheckBox.toolTip = "<p>Show full path for target images.</p>";
    this.fullPathCheckBox.onClick = function( checked ){
        data.useFullPath = checked;
        data.savedSelectedTarget = self.getSelectedTargetFilename(false);
        updateTargetImagesList();
        setSelectedTarget();
    };
    
    this.updateTargetImageCount = function(){
        nImagesControl.setValue(self.files_TreeBox.numberOfChildren);
    };
    
    let nImagesControl = new NumericEdit(this);
    nImagesControl.real = false;
    nImagesControl.label.text = "Images:";
    nImagesControl.toolTip = "Number of target images";
    nImagesControl.setRange(0, 99999);
    nImagesControl.enabled = false;
    
    function updateTargetImagesList(){
        self.files_TreeBox.clear();
        let maxPSFSGL = getMaxPSFSGL(data.targetFiles);
        let maxNoise = getMaxDisplayNoise(data.targetFiles);
        for ( let i = 0; i < data.targetFiles.length; i++ ){
            try {
                addTargetFileNode(data.targetFiles[i], maxPSFSGL, maxNoise);
            } catch (exception){
                console.criticalln("Failed to read FITS header '", data.targetFiles[i], "': ", exception);
            }
        }
        self.updateTargetTextColor(data);
        self.updateTargetImageCount();
    }
    
    let filesAdd_Button = new PushButton( this );
    filesAdd_Button.text = "Add";
    filesAdd_Button.icon = this.scaledResource( ":/icons/add.png" );
    filesAdd_Button.toolTip = "<p>Add target files.</p>";
    filesAdd_Button.onClick = function(){
        let openFileDialog = new OpenFileDialog;
        openFileDialog.multipleSelections = true;
        openFileDialog.caption = "Select Target Images";
        openFileDialog.loadImageFilters();

        if ( openFileDialog.execute() ){
            self.enabled = false;
            processEvents();
            for ( let i = 0; i < openFileDialog.fileNames.length; ++i ){
                let tgtFile = openFileDialog.fileNames[i];
                try {
                    getFITSKeywords(tgtFile); // Check file is readable
                    data.targetFiles.push(tgtFile);
                } catch (exception){
                    console.criticalln("Failed to read '", tgtFile, "': ", exception);
                }
            }
            updateTargetImagesList();
            self.enabled = true;
            processEvents();
        }
    };

    let filesClear_Button = new PushButton( this );
    filesClear_Button.text = "Clear";
    filesClear_Button.icon = this.scaledResource( ":/icons/clear.png" );
    filesClear_Button.toolTip = "<p>Remove all target files from the list.</p>";
    filesClear_Button.onClick = function(){
        self.files_TreeBox.clear();
        data.targetFiles.length = 0;
        data.cache.setTgtFilename(undefined);
        adjustFilenameColumnWidth();
        self.updateTargetImageCount();
        data.setRefFilename(undefined);
        self.refTextBox.text = "";
    };

    let filesInvert_Button = new PushButton( this );
    filesInvert_Button.text = "Invert selection";
    filesInvert_Button.icon = this.scaledResource( ":/icons/select-invert.png" );
    filesInvert_Button.toolTip = "<p>Invert the current selection of target files.</p>";
    filesInvert_Button.onClick = function(){
        for ( let i = 0; i < self.files_TreeBox.numberOfChildren; ++i ){
            self.files_TreeBox.child( i ).selected = !self.files_TreeBox.child( i ).selected;
        }
        data.cache.setTgtFilename(undefined);
    };

    let filesRemove_Button = new PushButton( this );
    filesRemove_Button.text = "Remove selected";
    filesRemove_Button.icon = this.scaledResource( ":/icons/delete.png" );
    filesRemove_Button.toolTip = "<p>Remove all selected files from the list.</p>";
    filesRemove_Button.onClick = function(){
        let refFilename = data.cache.getRefFilename();
        for ( let i = self.files_TreeBox.numberOfChildren - 1; i >= 0; --i){
            if ( self.files_TreeBox.child( i ).selected ){
                let node = self.files_TreeBox.child( i );
                if (node.text(COL_FULL_FILENAME) === refFilename){
                    data.setRefFilename(undefined);
                    self.refTextBox.text = "";
                }
                self.files_TreeBox.remove( i );       
            }
        }
        data.targetFiles.length = 0;
        for ( let i = 0; i < self.files_TreeBox.numberOfChildren; ++i ){
            data.targetFiles.push(self.files_TreeBox.child( i ).text( COL_FULL_FILENAME ));
        }
        data.cache.setTgtFilename(undefined);
        adjustFilenameColumnWidth();
        self.updateTargetImageCount();
    };

    let filesButtons_Sizer = new HorizontalSizer;
    filesButtons_Sizer.spacing = 4;
    filesButtons_Sizer.add( nImagesControl );
    filesButtons_Sizer.addStretch();
    filesButtons_Sizer.add( this.fullPathCheckBox );
    filesButtons_Sizer.addStretch();
    filesButtons_Sizer.add( filesAdd_Button );
    filesButtons_Sizer.addStretch();
    filesButtons_Sizer.add( filesClear_Button );
    filesButtons_Sizer.addStretch();
    filesButtons_Sizer.add( filesInvert_Button );
    filesButtons_Sizer.add( filesRemove_Button );
    
    let filesSection = new Control(this);
    filesSection.sizer = new VerticalSizer;
    filesSection.sizer.add( this.files_TreeBox );
    filesSection.sizer.addSpacing(4);
    filesSection.sizer.add( filesButtons_Sizer );
    let filesBar = new SectionBar(this, "Target Images");
    filesBar.setSection(filesSection);
    filesBar.onToggleSection = this.onToggleSection;
    // SectionBar "Target Images" End
    
    // =======================================
    // SectionBar: "Output Images"
    // =======================================
    let labelWidth = this.font.width( "Output Directory:" );
    let outputDir_Label = new Label( this );
    outputDir_Label.text = "Output Directory:";
    outputDir_Label.minWidth = labelWidth;
    outputDir_Label.textAlignment = TextAlign_Right | TextAlign_VertCenter;
   
    this.outputDirTextBox = new Edit( this );
    this.outputDirTextBox.onEditCompleted = function(){
        if (self.outputDirTextBox.text && File.directoryExists( self.outputDirTextBox.text )){
            data.outputDir = self.outputDirTextBox.text;
        } else {
            self.outputDirTextBox.text = data.outputDir;
        }
    };
    
    let outputDirButton = new ToolButton( this );
    outputDirButton.icon = this.scaledResource( ":/browser/select-file.png" );
    outputDirButton.setScaledFixedSize( 20, 20 );
    outputDirButton.toolTip = "<p>Select the output directory.</p>";
    outputDirButton.onClick = function(){
        let openFolderDialog = new GetDirectoryDialog();
        openFolderDialog.caption = "Select Output Directory";
//        openFolderDialog.initialPath = data.outputDir;
        if ( openFolderDialog.execute() ){
            data.outputDir = openFolderDialog.directory;
            self.outputDirTextBox.text = data.outputDir;
        }
    };
    let outputDirSizer = new HorizontalSizer(this);
    outputDirSizer.spacing = 4;
    outputDirSizer.add(outputDir_Label);
    outputDirSizer.add(this.outputDirTextBox, 100);
    outputDirSizer.add(outputDirButton);
    
    let postfix_Label = new Label( this );
    postfix_Label.text = "Postfix:";
    postfix_Label.minWidth = labelWidth;
    postfix_Label.textAlignment = TextAlign_Right | TextAlign_VertCenter;
    this.postFixTextBox = new Edit( this );
    this.postFixTextBox.setFixedWidth(this.font.width("___NSGWEIGHT___"));
    this.postFixTextBox.onTextUpdated = function (value){
        data.outputPostFix = value.trim();
    };
    
    this.overwriteCheckBox = new CheckBox( this );
    this.overwriteCheckBox.text = "Overwrite existing files";
    this.overwriteCheckBox.toolTip =
        "<p>Allow overwriting of existing image files.</p>" +
        "<p><b>Warning</b> This option may lead to irreversible data loss - enable it at your own risk.</p>";
    this.overwriteCheckBox.onClick = function( checked ){
        data.overwrite = checked;
    };
    
    this.rescaleCheckBox = new CheckBox( this );
    this.rescaleCheckBox.text = "Rescale result";
    this.rescaleCheckBox.toolTip =
        "<p>After normalization, some images can exceed PixInsight's 0.0 to 1.0 range. " +
        "These images are truncated to the 0.0 to 1.0 range.</p>" +
        "<p>If this option is selected, a single scale factor is applied to all corrected images " +
        "to prevent truncation. <b>All images must be normalized in one run.</b></p>";
    this.rescaleCheckBox.onClick = function( checked ){
        data.rescale = checked;
    };
    
    this.displayGradientToggle = new CheckBox(this);
    this.displayGradientToggle.text = "Gradient files";
    this.displayGradientToggle.toolTip = "<p>Save images that show the relative gradient." +
            "<Filenames start with 'gradient'/p>" +
            "<p>If a gradient peak or trough corresponds to a bright star, " +
            "increase the Sample Rejection: Star growth rate (see the Sample Generation section). " +
            "If a very bright saturated star does not have a rejection circle, " +
            "use Manual Sample Rejection to add one.</p>" +
            "<p>If any signs of a galaxy / nebula " +
            "is visible, the gradient smoothness should be increased.</p>";
    this.displayGradientToggle.onCheck = function (checked) {
        data.displayGradient = checked;
    };
    this.displayGradientToggle.checked = data.displayGradient;
    
    let noiseWeightToolTip = 
            "<p>NWEIGHT is the square of the signal to noise ratio. " +
            "The signal is measured using stellar photometry. " +
            "The noise was calculated by PixInsight during calibration, " +
            "and is stored in the NOISExx FITS header.</p>" +
            "<p>NWEIGHT can be used by ImageIntegration to specify the weight of each image. " +
            "Use this text box to specify the FITS header name. </p>" +
            "<p>In ImageIntegration, set<br/>" +
            "<b>Weights:</b> to 'FITS keyword'<br/>" +
            "<b>Weight keyword:</b> to this value.</p>";
    let noiseWeight_Label = new Label( this );
    noiseWeight_Label.text = "Weight keyword:";
    noiseWeight_Label.minWidth = labelWidth;
    noiseWeight_Label.textAlignment = TextAlign_Right | TextAlign_VertCenter;
    noiseWeight_Label.toolTip = noiseWeightToolTip;
    this.noiseWeightTextBox = new Edit( this );
    this.noiseWeightTextBox.setFixedWidth(this.font.width("___NSGWEIGHT___"));
    this.noiseWeightTextBox.toolTip = noiseWeightToolTip;        
    this.noiseWeightTextBox.onTextUpdated = function (value){
        let keyword = value.trim();
        // Allow 8 characters if last digit is '0'. Otherwise limit to 7.
        if (keyword.length >= 8){
            let maxLen = keyword[7] === '0' ? 8 : 7;
            keyword = keyword.substring(0, maxLen);
            self.noiseWeightTextBox.text = keyword;
        }
        data.noiseWeightKeyword = keyword.length ? keyword : "NWEIGHT";
    };
    
    this.weightPrefixCheckBox = new CheckBox( this );
    this.weightPrefixCheckBox.text = "Weight prefix";
    this.weightPrefixCheckBox.toolTip =
        "<p>Prefix the output files with their weight. " +
        "Files can then be ordered by weight within a file browser.</p>" +
        "<p>Example filename: 'w098_filename_nsg' for a file with weight 0.98</p>";
    this.weightPrefixCheckBox.onClick = function( checked ){
        data.weightPrefix = checked;
    };
    
    let onError_Label = new Label( this );
    onError_Label.text = "On error:";
    onError_Label.minWidth = labelWidth;
    onError_Label.textAlignment = TextAlign_Right | TextAlign_VertCenter;
    this.onErrorComboBox = new ComboBox( this );
    this.onErrorComboBox.addItem("Continue");
    this.onErrorComboBox.addItem("Abort");
    this.onErrorComboBox.addItem("Ask user");
    this.onErrorComboBox.toolTip = "<p>Note that if a run is aborted, " +
            "the 'Rescale result' option is also aborted.</p>" +
            "<p>The images that have already been normalized will not be rescaled.</p>";
    this.onErrorComboBox.onItemSelected = function ( itemIndex ){
        data.onErrorIndex = itemIndex;
    };
    
    let postFixSizer = new HorizontalSizer(this);
    postFixSizer.spacing = 4;
    postFixSizer.add(postfix_Label);
    postFixSizer.add(this.postFixTextBox);
    postFixSizer.addSpacing(20);
    postFixSizer.add(this.overwriteCheckBox);
    postFixSizer.addSpacing(20);
    postFixSizer.add(this.displayGradientToggle);
    postFixSizer.addStretch();
    
    let weightSizer = new HorizontalSizer(this);
    weightSizer.spacing = 4;
    weightSizer.add(noiseWeight_Label);
    weightSizer.add(this.noiseWeightTextBox);
    weightSizer.addSpacing(20);
    weightSizer.add(this.weightPrefixCheckBox);
    weightSizer.addSpacing(20);
    weightSizer.add(this.rescaleCheckBox);
    weightSizer.addStretch();
    weightSizer.add(onError_Label);
    weightSizer.add(this.onErrorComboBox);
    
    let outputDirSection = new Control(this);
    outputDirSection.sizer = new VerticalSizer;
    outputDirSection.sizer.spacing = 4;
    outputDirSection.sizer.add(outputDirSizer);
    outputDirSection.sizer.add(postFixSizer);
    outputDirSection.sizer.add(weightSizer);
    let outputDirBar = new SectionBar(this, "Output Images");
    outputDirBar.setSection(outputDirSection);
    outputDirBar.onToggleSection = this.onToggleSection;
    // SectionBar "Reference file" End

    // =======================================
    // SectionBar: "Star Detection"
    // =======================================
    let starDetectionControls = new NsgStarDetectionControls();
    this.refLogStarDetection_Control = starDetectionControls.createRefLogStarDetect_Control(this, data, 0);
    
    let detectedStarsButton = new PushButton(this);
    detectedStarsButton.text = "Detected stars ";
    detectedStarsButton.toolTip =
            "<p>Displays the detected stars and provides star detection controls.</p>";
    detectedStarsButton.onClick = function () {
        let tgtFile = getSelectedTarget(true);
        if (tgtFile){
            let refFile = getReference();
            if (refFile){
                if (data.cache.setTgtFilename(tgtFile)){
                    (new DetectedStarsDialog("Detected Stars", data, self)).execute();
                    data.setPhotometryAutoValues(data.useAutoPhotometry, true);
                    data.setSampleGenerationAutoValues(data.useAutoSampleGeneration, true);
                } else {
                    (new MessageBox("Failed to read '" + tgtFile + "'")).execute();
                }
            }
        }
    };
    
    let starDetectionSection = new Control(this);
    starDetectionSection.sizer = new HorizontalSizer;
    starDetectionSection.sizer.spacing = 20;
    starDetectionSection.sizer.add(this.refLogStarDetection_Control, 100);
    starDetectionSection.sizer.add(detectedStarsButton, 0);
    let starDetectionBar = new SectionBar(this, "Star Detection");
    starDetectionBar.setSection(starDetectionSection);
    starDetectionBar.onToggleSection = this.onToggleSection;
    starDetectionBar.toolTip = "<p>Star detection sensitivity.</p>";
    // SectionBar "Star Detection" End
    
    let photometrySearchSection;
    let photometrySearchBar;

    // =======================================
    // SectionBar: "Photometric Star Search"
    // =======================================
    let labelWidth = Math.max(
            this.font.width("Star flux tolerance:"), 
            this.font.width("Star search radius:"));
    this.starFluxTolerance_Control = new NumericControl(this);
    this.starFluxTolerance_Control.real = true;
    this.starFluxTolerance_Control.label.text = "Star flux tolerance:";
    this.starFluxTolerance_Control.toolTip =
            "<p>Star flux tolerance is used to prevent invalid target to reference " +
            "star matches. Smaller values reject more matches.</p>" +
            "<p>Star matches are rejected if the difference in star flux " +
            "is larger than expected. The algorithm first calculates the average scale difference, " +
            "and then rejects matches if their brightness ratio is greater than " +
            "(expected ratio * tolerance) or smaller than (expected ratio / tolerance)</p>" +
            "<p>1.0 implies the star flux ratio must exactly match the expected ratio.</p>" +
            "<p>2.0 implies that the ratio can be double or half the expected ratio.</p>" +
            "<p>You usually don't need to modify this parameter.</p>";
    this.starFluxTolerance_Control.label.minWidth = labelWidth;
    this.starFluxTolerance_Control.setRange(1.01, 2);
    this.starFluxTolerance_Control.slider.setRange(100, 200);
    this.starFluxTolerance_Control.setPrecision(2);
    this.starFluxTolerance_Control.slider.minWidth = 100;
    this.starFluxTolerance_Control.onValueUpdated = function (value) {
        data.starFluxTolerance = value;
    };

    this.starSearchRadius_Control = new NumericControl(this);
    this.starSearchRadius_Control.real = true;
    this.starSearchRadius_Control.label.text = "Star search radius:";
    this.starSearchRadius_Control.toolTip =
            "<p>Search radius used to match the reference and target stars. " +
            "Larger values find more photometric stars but at the risk of matching " +
            "the wrong star.</p>" +
            "<p>You only need to modify this parameter if your images contain distortions.</p>";

    this.starSearchRadius_Control.label.minWidth = labelWidth;
    this.starSearchRadius_Control.setRange(0.1, 20);
    this.starSearchRadius_Control.slider.setRange(1, 200);
    this.starSearchRadius_Control.setPrecision(1);
    this.starSearchRadius_Control.slider.minWidth = 100;
    this.starSearchRadius_Control.onValueUpdated = function (value) {
        data.starSearchRadius = value;
    };

    photometrySearchSection = new Control(this);
    photometrySearchSection.sizer = new VerticalSizer;
    photometrySearchSection.sizer.spacing = 4;
    photometrySearchSection.sizer.add(this.starFluxTolerance_Control);
    photometrySearchSection.sizer.add(this.starSearchRadius_Control);
    photometrySearchBar = new SectionBar(this, "Photometry Star Search");
    photometrySearchBar.setSection(photometrySearchSection);
    photometrySearchBar.onToggleSection = this.onToggleSection;
    photometrySearchBar.toolTip = "<p>Search criteria used to match reference and target stars.</p>" +
            "<p>The default settings usually work well.</p>";
    // SectionBar: "Photometric Star Search" End
    
    // =======================================
    // SectionBar: "Photometry"
    // =======================================
    let photometryControls = new NsgPhotometryControls();
    this.apertureGrowthRate_Control = photometryControls.createApertureGrowthRateEdit(this, data);
    this.apertureGrowthRate_Control.onValueUpdated = function (value){
        data.apertureGrowthRate = value;
    };
    this.apertureAdd_Control = photometryControls.createApertureAddEdit(this, data);
    this.apertureAdd_Control.onValueUpdated = function (value){
        data.apertureAdd = value;
    };
    this.apertureGap_Control = photometryControls.createApertureGapEdit(this, data);
    this.apertureGap_Control.onValueUpdated = function (value){
        data.apertureGap = value;
    };
    this.apertureBgDelta_Control = photometryControls.createApertureBgDeltaEdit(this, data);
    this.apertureBgDelta_Control.onValueUpdated = function (value){
        data.apertureBgDelta = value;
    };
    let photometryStarsButton = new PushButton(this);
    photometryStarsButton.text = "Photometry stars ";
    photometryStarsButton.toolTip =
            "<p>Displays the photometry stars.</p>" + 
            "<p>Provides all the photometry controls.</p>";
    photometryStarsButton.onClick = function () {
        let tgtFile = getSelectedTarget(true);
        if (tgtFile){
            let refFile = getReference();
            if (refFile){
                if (data.cache.setTgtFilename(tgtFile)){
                    data.setPhotometryAutoValues(data.useAutoPhotometry, true);
                    (new PhotometryStarsDialog("Photometry Stars", data, self)).execute();
                } else {
                    (new MessageBox("Failed to read '" + tgtFile + "'")).execute();
                }
            }
        }
    };

    let apertureGroupBox = new GroupBox(this);
    apertureGroupBox.title = "Star aperture size";
    apertureGroupBox.sizer = new HorizontalSizer();
    apertureGroupBox.sizer.margin = 2;
    apertureGroupBox.sizer.spacing = 10;
    apertureGroupBox.sizer.add(this.apertureAdd_Control);
    apertureGroupBox.sizer.add(this.apertureGrowthRate_Control);
    apertureGroupBox.sizer.add(this.apertureGap_Control);
    apertureGroupBox.sizer.add(this.apertureBgDelta_Control);
    apertureGroupBox.sizer.addStretch();
    
    this.limitPhotoStarsPercent_Control = 
            photometryControls.createLimitPhotoStarsPercentEdit(this, data);
    this.limitPhotoStarsPercent_Control.onValueUpdated = function (value){
        data.limitPhotoStarsPercent = value;
    };
    
    this.linearRangeRef_Control = photometryControls.createLinearRangeRefEdit(this, data);
    this.linearRangeRef_Control.onValueUpdated = function (value){
        data.linearRangeRef = value;
    };
    
    this.outlierRemoval_Control = 
            photometryControls.createOutlierRemovalEdit(this, data);
    this.outlierRemoval_Control.onValueUpdated = function (value){
        data.outlierRemoval = value;
    };
 
    let photometryGraphButton = new PushButton(this);
    photometryGraphButton.text = "Photometry graph";
    photometryGraphButton.toolTip =
            "<p>Displays the photometry graph. " +
            "For each star, the flux measured in the reference image is plotted " +
            "against the flux measured in the target image. " +
            "A best fit line is drawn through these points. " +
            "The gradient provides the brightness scale factor.</p>" +
            "<p>Provides all the photometry controls.</p>";
    photometryGraphButton.onClick = function () {
        let tgtFile = getSelectedTarget(true);
        if (tgtFile){
            let refFile = getReference();
            if (refFile){
                if (data.cache.setTgtFilename(tgtFile)){
                    data.setPhotometryAutoValues(data.useAutoPhotometry, true);
                    displayStarGraph(data.cache.getRefView(), data.cache.getTgtView(), data, self);
                } else {
                    (new MessageBox("Failed to read '" + tgtFile + "'")).execute();
                }
            }
        }
    };

    let filterGroupBox = new GroupBox(this);
    filterGroupBox.title = "Filter photometry stars";
    filterGroupBox.sizer = new HorizontalSizer(filterGroupBox);
    filterGroupBox.sizer.margin = 2;
    filterGroupBox.sizer.spacing = 10;
    filterGroupBox.sizer.add(this.limitPhotoStarsPercent_Control);
    filterGroupBox.sizer.add(this.outlierRemoval_Control);
    filterGroupBox.sizer.addStretch();
    
    let linearGroupBox = new GroupBox(this);
    linearGroupBox.title = "Linear range";
    linearGroupBox.sizer = new HorizontalSizer(linearGroupBox);
    linearGroupBox.sizer.margin = 2;
    linearGroupBox.sizer.spacing = 10;
    linearGroupBox.sizer.add(this.linearRangeRef_Control);
    linearGroupBox.sizer.addStretch();
    
    let starButtonGroupBox = new GroupBox(this);
    starButtonGroupBox.title = "Edit / Display";
    starButtonGroupBox.sizer = new HorizontalSizer(starButtonGroupBox);
    starButtonGroupBox.sizer.margin = 2;
    starButtonGroupBox.sizer.addSpacing(2);
    starButtonGroupBox.sizer.add(photometryStarsButton);
    starButtonGroupBox.sizer.addSpacing(2);
    
    let graphButtonGroupBox = new GroupBox(this);
    graphButtonGroupBox.title = "Edit / Display";
    graphButtonGroupBox.sizer = new HorizontalSizer(starButtonGroupBox);
    graphButtonGroupBox.sizer.margin = 2;
    graphButtonGroupBox.sizer.addSpacing(2);
    graphButtonGroupBox.sizer.add(photometryGraphButton);
    graphButtonGroupBox.sizer.addSpacing(2);
 
    this.autoPhotometryCheckBox = new CheckBox(this);
    this.autoPhotometryCheckBox.text = "Auto";
    this.autoPhotometryCheckBox.toolTip = 
            "<p>Sets all controls to calculated values.</p>";
    this.autoPhotometryCheckBox.onCheck = function (checked){
        if (checked){
            self.enabled = false;
            processEvents();
            data.cache.setTgtFilename(getSelectedTarget(false));
        }
        data.setPhotometryAutoValues(checked, true);
        self.enabled = true;
        processEvents();
    };

    let photometryAutoGroupBox = new GroupBox(this);
    photometryAutoGroupBox.sizer = new HorizontalSizer();
    photometryAutoGroupBox.sizer.margin = 2;
    photometryAutoGroupBox.sizer.addSpacing(10);
    photometryAutoGroupBox.sizer.add(this.autoPhotometryCheckBox);
    photometryAutoGroupBox.sizer.addSpacing(10);
    
    let apertureHorizSizer = new HorizontalSizer();
    apertureHorizSizer.spacing = 12;
    apertureHorizSizer.add(apertureGroupBox, 100);
    apertureHorizSizer.add(photometryAutoGroupBox);
    apertureHorizSizer.add(starButtonGroupBox);
    
    let filterHorizSizer = new HorizontalSizer();
    filterHorizSizer.spacing = 12;
    filterHorizSizer.add(filterGroupBox, 100);
    filterHorizSizer.add(linearGroupBox, 100);
    filterHorizSizer.add(graphButtonGroupBox);

    let photometrySection = new Control(this);
    photometrySection.sizer = new VerticalSizer();
    photometrySection.sizer.spacing = 4;
    photometrySection.sizer.add(apertureHorizSizer);
    photometrySection.sizer.add(filterHorizSizer);
    let photometryBar = new SectionBar(this, "Photometry");
    photometryBar.setSection(photometrySection);
    photometryBar.onToggleSection = this.onToggleSection;
    photometryBar.toolTip = "<p>Specifies photometry parameters. These are used " +
            " to calculate the brightness scale factor.</p>";
    // SectionBar: "Photometric Scale" End

    // =======================================
    // SectionBar: "Sample Generation"
    // =======================================
    const sampleGenerationStrLen = this.font.width("Multiply star radius:");
    let sampleControls = new NsgSampleControls;

    this.limitSampleStarsPercent_Control = sampleControls.createLimitSampleStarsPercentEdit(this, data);       
    this.limitSampleStarsPercent_Control.onValueUpdated = function (value) {
        data.limitSampleStarsPercent = value;
    };
    
    let filterSampleStarsGroupBox = new GroupBox(this);
    filterSampleStarsGroupBox.title = "Filter stars";
    filterSampleStarsGroupBox.sizer = new HorizontalSizer();
    filterSampleStarsGroupBox.sizer.margin = 2;
    filterSampleStarsGroupBox.sizer.spacing = 10;
    filterSampleStarsGroupBox.sizer.add(this.limitSampleStarsPercent_Control);
    filterSampleStarsGroupBox.sizer.addStretch();
    
    this.sampleStarGrowthRate_Control = sampleControls.createSampleStarGrowthRateEdit(this, data);    
    this.sampleStarGrowthRate_Control.onValueUpdated = function (value){
        data.sampleStarGrowthRate = value;
    };
    
    this.autoSampleGenerationCheckBox = new CheckBox(this);
    this.autoSampleGenerationCheckBox.text = "Auto";
    this.autoSampleGenerationCheckBox.toolTip = 
            "<p>Calculates default values for most of the Sample Generation parameters.</p>" +
            "<p>These are calculated from the headers:" +
            "<ul><li><b>'XPIXSZ'</b> (Pixel size, including binning, in microns)</li>" +
            "<li><b>'FOCALLEN'</b> (Focal length in mm).</li></p>";
    this.autoSampleGenerationCheckBox.onCheck = function (checked){
        if (checked){
            self.enabled = false;
            processEvents();
        }
        data.setSampleGenerationAutoValues(checked, true);
        self.enabled = true;
        processEvents();
    };

    let sampleAutoGroupBox = new GroupBox(this);
    sampleAutoGroupBox.sizer = new HorizontalSizer();
    sampleAutoGroupBox.sizer.margin = 2;
    sampleAutoGroupBox.sizer.addSpacing(10);
    sampleAutoGroupBox.sizer.add(this.autoSampleGenerationCheckBox);
    sampleAutoGroupBox.sizer.addSpacing(10);

    let sampleStarRejectRadiusGroupBox = new GroupBox(this);
    sampleStarRejectRadiusGroupBox.title = "Sample rejection";
    sampleStarRejectRadiusGroupBox.toolTip = "<p>This section determines which " +
            "samples are used to create the relative gradient model that will be " +
            "used to correct the target image.</p>" +
            "<p>The aim is to reject samples that cover any light from bright stars. " +
            "This includes diffraction spikes, filter halos " +
            "and the scattered light around bright stars.</p>";
    sampleStarRejectRadiusGroupBox.sizer = new HorizontalSizer();
    sampleStarRejectRadiusGroupBox.sizer.margin = 2;
    sampleStarRejectRadiusGroupBox.sizer.spacing = 10;
    sampleStarRejectRadiusGroupBox.sizer.add(this.sampleStarGrowthRate_Control);
    sampleStarRejectRadiusGroupBox.sizer.addStretch();
    
    this.sampleSize_Control = sampleControls.createSampleSizeEdit(this, data);
    this.sampleSize_Control.onValueUpdated = function (value) {
        data.sampleSize = value;
    };

    let sampleSizeGroupBox = new GroupBox(this);
    sampleSizeGroupBox.title = "Samples";
    sampleSizeGroupBox.sizer = new HorizontalSizer();
    sampleSizeGroupBox.sizer.margin = 2;
    sampleSizeGroupBox.sizer.spacing = 0;
    sampleSizeGroupBox.sizer.add(this.sampleSize_Control);
    if (NSG_EXTRA_CONTROLS){
        this.maxSamples_Control = new NumericEdit(this);
        this.maxSamples_Control.real = false;
        this.maxSamples_Control.label.text = "";
        this.maxSamples_Control.toolTip =
            "<p>Limits the number of samples used to create the surface spline. " +
            "If the number of samples exceed this limit, they are combined " +
            "(binned) to create super samples.</p>" +
            "<p>A larger number of samples increases the " +
            "theoretical maximum resolution of the surface spline. However, " +
            "small unbinned samples are noisier and require more smoothing. " +
            "The default value is usually a good compromise.</p>" +
            "<p>The time required to initialize the surface spline approximately " +
            "doubles every 1300 samples.</p>";
        this.maxSamples_Control.setRange(2000, 5000);
        this.maxSamples_Control.setValue(data.maxSamples);
        this.maxSamples_Control.enabled = false;

        let displayBinnedSamplesButton = new PushButton(this);
        displayBinnedSamplesButton.text = "Max samples";
        displayBinnedSamplesButton.toolTip =
                "<p>Displays the binned samples used to construct the surface spline " +
                "that models the relative gradient between the reference and target images.</p>" +
                "<p>Samples are binned to improve performance if the number of " +
                "samples exceed the specified limit.</p>" +
                "<p>The area of each binned sample represents the number of samples " +
                "it was created from.</p>" +
                "<p>Each binned sample's center is calculated from " +
                "the center of mass of the samples it was created from.</p>" +
                "<p>To see which of the unbinned samples were rejected due to stars, " +
                "use 'Sample grid'.</p>";
        displayBinnedSamplesButton.onClick = function () {
            let tgtFile = getSelectedTarget(true);
            if (tgtFile){
                let refFile = getReference();
                if (refFile){
                    if (data.cache.setTgtFilename(tgtFile)){
                        data.setPhotometryAutoValues(data.useAutoPhotometry, true);
                        data.setSampleGenerationAutoValues(data.useAutoSampleGeneration, true);

                        // Scale factors
                        let nChannels = data.cache.isColor() ? 3 : 1;
                        let colorStarPairs = getColorStarPairs(nChannels, data);
                        let scaleFactors = getScaleFactors(colorStarPairs, data);

                        // === Sample pairs ===
                        let sampleGrid = data.cache.getSampleGrid(data, tgtFile);
                        let stars = data.cache.getRefStars(data.logStarDetection);
                        let colorSamplePairs = sampleGrid.createColorScaledSamplePairs(stars, data, scaleFactors);
                        let dialog = new BinnedSampleGridDialog("Binned Samples", colorSamplePairs[0], data, self);
                        dialog.execute();
                        dialog = null;
                    } else {
                        (new MessageBox("Failed to read '" + tgtFile + "'")).execute();
                    }
                }
            }
        };
        sampleSizeGroupBox.sizer.addSpacing(10);
        sampleSizeGroupBox.sizer.add(displayBinnedSamplesButton);
        sampleSizeGroupBox.sizer.add(this.maxSamples_Control);
    }
    sampleSizeGroupBox.sizer.addStretch();

    let displaySamplesButton = new PushButton(this);
    displaySamplesButton.text = "Sample generation";
    displaySamplesButton.toolTip =
            "<p>Displays the generated samples and the stars used to reject samples.</p>" +
            "<p>Provides edit sliders for all 'Sample Generation' section parameters.</p>" +
            "<p>Samples are rejected if they: " +
            "<ul><li>Contain one or more zero pixels in either image.</li>" +
            "<li>Are too close to a star included in the 'Limit stars %' range.</li></ul>" +
            "The surviving samples are drawn as squares. The stars used to " +
            "reject samples are indicated by circles.</p>" +
            "<p>A surface spline is constructed from the generated samples " +
            "to model the relative gradient between the reference and target images.</p>";
    displaySamplesButton.onClick = function () {
        let tgtFile = getSelectedTarget(true);
        if (tgtFile){
            let refFile = getReference();
            if (refFile){
                if (data.cache.setTgtFilename(tgtFile)){
                    data.setPhotometryAutoValues(data.useAutoPhotometry, true);
                    data.setSampleGenerationAutoValues(data.useAutoSampleGeneration, true);
                    let dialog = new SampleGridDialog("Sample Generation", data, tgtFile, self);
                    dialog.execute();
                    dialog = null;
                } else {
                    (new MessageBox("Failed to read '" + tgtFile + "'")).execute();
                }
            }
        }
    };
    
    let editDisplayGroupBox = new GroupBox(this);
    editDisplayGroupBox.title = "Edit / Display";
    editDisplayGroupBox.sizer = new HorizontalSizer();
    editDisplayGroupBox.sizer.margin = 2;
    editDisplayGroupBox.sizer.addSpacing(2);
    editDisplayGroupBox.sizer.add(displaySamplesButton);
    editDisplayGroupBox.sizer.addSpacing(2);
    
    let generateSamplesHorizSizer = new HorizontalSizer();
    generateSamplesHorizSizer.spacing = 12;
    generateSamplesHorizSizer.add(filterSampleStarsGroupBox);
    generateSamplesHorizSizer.add(sampleSizeGroupBox);
    generateSamplesHorizSizer.add(sampleStarRejectRadiusGroupBox);
    generateSamplesHorizSizer.add(sampleAutoGroupBox);
    generateSamplesHorizSizer.add(editDisplayGroupBox);
    
    let sampleGenerationSection = new Control(this);
    sampleGenerationSection.sizer = new VerticalSizer;
    sampleGenerationSection.sizer.spacing = 4;
    sampleGenerationSection.sizer.add(generateSamplesHorizSizer);
    let sampleGenerationBar = new SectionBar(this, "Sample Generation");
    sampleGenerationBar.setSection(sampleGenerationSection);
    sampleGenerationBar.onToggleSection = this.onToggleSection;
    sampleGenerationBar.toolTip = 
            "<p>This section generates samples used to model " +
            "the relative gradient between the reference and target images.</p>" +
            "<p>The image is divided up into a grid of sample squares. " +
            "A sample's value is the median of the pixels it contains.</p>" +
            "<p>Samples are rejected if they contain one or more zero pixels in " +
            "either image or if they are too close to a bright star.</p>";

    // SectionBar: "Gradient Sample Generation" End
    
    // ===============================================================
    // SectionBar: "Gradient Correction" : Group box
    // ===============================================================
    // Gradient controls
    let gradientControls = new NsgGradientControls();
    this.gradientSmoothness_Control = 
            gradientControls.createGradientSmoothnessControl(this, data, 0);
    this.gradientSmoothness_Control.onValueUpdated = function (value) {
        data.gradientSmoothness = value;
    };
    
    let gradientGraphButton = new PushButton(this);
    gradientGraphButton.text = "Gradient graph";
    gradientGraphButton.toolTip =
        "<p>Edit the 'Smoothness' parameter and view the gradient.</p>" +
        "<p>The vertical axis represents the difference between the two images, " +
        "the horizontal axis the distance across the image.</p>";
    gradientGraphButton.onClick = function () {
        let tgtFile = getSelectedTarget(true);
        if (tgtFile){
            let refFile = getReference();
            if (refFile){
                if (data.cache.setTgtFilename(tgtFile)){
                    data.setPhotometryAutoValues(data.useAutoPhotometry, true);
                    data.setSampleGenerationAutoValues(data.useAutoSampleGeneration, true);

                    let nChannels = data.cache.isColor() ? 3 : 1;

                    // Scale factors
                    let colorStarPairs = getColorStarPairs(nChannels, data);
                    let scaleFactors = getScaleFactors(colorStarPairs, data);

                    // === Sample pairs ===
                    let sampleGrid = data.cache.getSampleGrid(data, tgtFile);
                    let stars = data.cache.getRefStars(data.logStarDetection);
                    let colorSamplePairs = sampleGrid.createColorScaledSamplePairs(stars, data, scaleFactors);

                    for (let c=0; c<nChannels; c++){
                        if (colorSamplePairs[c].length < 3) {
                            new MessageBox("Error: Too few samples to create a Surface Spline.", TITLE, StdIcon_Error, StdButton_Ok).execute();
                            return;
                        }
                    }

                    // === Gradient graph ===
                    GradientGraph(colorSamplePairs, self, data);
                } else {
                    (new MessageBox("Failed to read '" + tgtFile + "'")).execute();
                }
            }
        }
        processEvents();
    };
    
    let gradientSection = new Control(this);
    gradientSection.sizer = new HorizontalSizer(this);
    gradientSection.sizer.margin = 2;
    gradientSection.sizer.spacing = 20;
    gradientSection.sizer.add(this.gradientSmoothness_Control, 100);
    gradientSection.sizer.add(gradientGraphButton);
    let gradientBar = new SectionBar(this, "Gradient Correction");
    gradientBar.setSection(gradientSection);
    gradientBar.onToggleSection = this.onToggleSection;
    
    //------------------------------------------
    // ImageIntegration
    //------------------------------------------
    this.minimumWeight_Control = new NumericControl(this);
    this.minimumWeight_Control.real = true;
    this.minimumWeight_Control.label.text = "Minimum weight %:";
    this.minimumWeight_Control.toolTip =
            "<p>ImageIntegration is populated with all the normalized images, but " +
            "if an image's weight is too low, the image entry is disabled. " +
            "This can be used to reject images taken through thin clouds or from a bright sky.</p>" +
            "<p>This control specifies the minimum weight for an enabled image " +
            "as a percentage of the highest image weight. " +
            "For example, if 'minimum weight %' is set to 50%, and the maximum image weight is 1.2, " +
            "all images with a weight less than 0.6 will be disabled.</p>" +
            "<p>The minimum weight criteria is relative to the exposure time. " +
            "For example, if 'minimum weight %' is 50%, and the maximum weight is 1.0 from a 10 minute exposure, " +
            "then a 5 minute exposure will only be disabled if its weight is less than 0.25</p>";
    this.minimumWeight_Control.setRange(0, 100);
    this.minimumWeight_Control.slider.setRange(0, 200);
    this.minimumWeight_Control.setPrecision(2);
    this.minimumWeight_Control.slider.minWidth = 100;
    this.minimumWeight_Control.onValueUpdated = function (value) {
        data.minimumWeight = value;
    };
    
    this.sortByWeightCheckBox = new CheckBox( this );
    this.sortByWeightCheckBox.text = "Sort by weight";
    this.sortByWeightCheckBox.toolTip =
        "<p>Determines the ImageIntegration input file order.</p>" +
        "<p>The first file is always the reference file. " +
        "All other files are either sorted by weight or filename.</p>" +
        "<p>When sorting by filename, the NormalizeScaleGradient input filename is used. This includes the full path.</p>";
    this.sortByWeightCheckBox.onClick = function( checked ){
        data.sortByWeight = checked;
    };
    
    let weightSizer = new HorizontalSizer;
    weightSizer.spacing = 20;
    weightSizer.add( this.minimumWeight_Control, 100 );
    weightSizer.add( this.sortByWeightCheckBox );
    
    let templateIconIdToolTip = 
            "<p>Initialize ImageIntegration from a process icon. " +
            "<p>Use &lt;Auto Rejection Algorithm&gt; to Initialize ImageIntegration " +
            "with a rejection algorithm based on the number of images.</p>" +
            "<p>The required NormalizeScaleGradient settings " +
            "(No normalization, weight keyword, etc) will override these initial settings.</p>";

    let templateIconIdLabel = new Label( this );
    templateIconIdLabel.text = "Template icon:";
    templateIconIdLabel.minWidth = this.font.width( "Minimum weight %:" );
    templateIconIdLabel.toolTip = templateIconIdToolTip;
    templateIconIdLabel.textAlignment = TextAlign_Right | TextAlign_VertCenter;

    this.templateIconIdComboBox = new ComboBox( this );
    this.templateIconIdComboBox.toolTip = templateIconIdToolTip;
    this.templateIconIdComboBox.addItem( "<Auto Rejection Algorithm>" );
    let icons = ProcessInstance.iconsByProcessId( "ImageIntegration" );
    for ( let i = 0; i < icons.length; ++i ){
        this.templateIconIdComboBox.addItem( icons[ i ] );
    }
    this.templateIconIdComboBox.onItemSelected = function( item ){
        if ( this.itemText( item ) === "<Auto Rejection Algorithm>" )
            data.imageIntegrationTemplateId = "";
        else
            data.imageIntegrationTemplateId = this.itemText( item );
    };
    
    let templateIconIdSizer = new HorizontalSizer;
    templateIconIdSizer.spacing = 4;
    templateIconIdSizer.add( templateIconIdLabel );
    templateIconIdSizer.add( this.templateIconIdComboBox, 100 );
    
    this.imageIntegrationSection = new Control(this);
    this.imageIntegrationSection.sizer = new VerticalSizer(this);
    this.imageIntegrationSection.sizer.margin = 2;
    this.imageIntegrationSection.sizer.spacing = 4;
    this.imageIntegrationSection.sizer.add(weightSizer);
    this.imageIntegrationSection.sizer.add(templateIconIdSizer);
    this.imageIntegrationBar = new SectionBar(this, "ImageIntegration");
    this.imageIntegrationBar.enableCheckBox();
    this.imageIntegrationBar.checkBox.onCheck = function (checked) {
        data.imageIntegration = checked;
    };
    this.imageIntegrationBar.setSection(this.imageIntegrationSection);
    this.imageIntegrationBar.onToggleSection = this.onToggleSection;
    this.imageIntegrationBar.toolTip = 
            "<p>If images have been normalized, the ImageIntegration process is displayed on exit.</p>" +
            "<p>ImageIntegration will be initialized with the normalized images, and with " +
            "the correct settings for normalized images. All other settings must be set by the user.</p>" +
            "<p>ImageIntegration is populated with all the normalized images, " +
            "but if an image's weight is too low, the image entry is disabled. " +
            "This can be used to reject images taken through thin clouds or from a bright sky.</p>";
    
    //------------------------------------------
    // Button bar
    //------------------------------------------
    let newInstance_Button = new ToolButton(this);
    newInstance_Button.icon = this.scaledResource(":/process-interface/new-instance.png");
    newInstance_Button.setScaledFixedSize(24, 24);
    newInstance_Button.toolTip = "Drag & Drop to desktop to create a Process Icon";
    newInstance_Button.onMousePress = function () {
        this.hasFocus = true;
        data.saveParameters();
        this.pushed = false;
        self.newInstance();
    };
    
    let browseDocumentationButton = new ToolButton(this);
    browseDocumentationButton.icon = this.scaledResource(":/process-interface/browse-documentation.png");
    browseDocumentationButton.setScaledFixedSize(24, 24);
    browseDocumentationButton.toolTip =
            "<p>Opens a browser to view the script's documentation.</p>";
    browseDocumentationButton.onClick = function () {
            Dialog.browseScriptDocumentation("NormalizeScaleGradient");
            return;
    };
    
    let resetButton = new ToolButton(this);
    resetButton.icon = this.scaledResource(":/process-interface/reset.png");
    resetButton.setScaledFixedSize(24, 24);
    resetButton.toolTip =
            "<p>Resets the dialog's parameters.</p>" +
            "<p>Saved settings are also cleared.</p>";
    resetButton.onClick = function () {
        data.setParameters();
        data.resetSettings();
        setDialogValues();
    };
    
    let ok_Button = new PushButton(this);
    ok_Button.defaultButton = false;
    ok_Button.text = "Run";
    ok_Button.toolTip = "<p>Normalize the target images to the reference.</p>";
    ok_Button.icon = this.scaledResource( ":/icons/power.png" );
    ok_Button.onClick = function () {
        self.ok();
    };
    
    let cancel_Button = new PushButton(this);
    cancel_Button.text = "Exit";
    cancel_Button.icon = this.scaledResource( ":/icons/close.png" );
    cancel_Button.onClick = function () {
        self.cancel();
    };

    this.smallScreenToggle = new CheckBox(this);
    this.smallScreenToggle.text = "Small screen";
    this.smallScreenToggle.toolTip = "<p>Restrict dialog window heights to less than 900 pixels</p>";
    this.smallScreenToggle.onCheck = function (checked) {
        data.smallScreen = checked;
    };
    this.smallScreenToggle.checked = data.smallScreen;
    
    this.autoExitToggle = new CheckBox(this);
    this.autoExitToggle.text = "Auto exit";
    this.autoExitToggle.toolTip = "<p>After normalization completes, exit.</p>" +
            "<p>If specified, ImageIntegration will then be invoked.</p>";
    this.autoExitToggle.onCheck = function (checked) {
        data.autoExit = checked;
    };
    this.autoExitToggle.checked = data.autoExit;
    
    let buttons_Sizer = new HorizontalSizer(this);
    buttons_Sizer.spacing = 6;
    buttons_Sizer.add(newInstance_Button);
    buttons_Sizer.add(browseDocumentationButton);
    buttons_Sizer.add(resetButton);
    buttons_Sizer.addSpacing(20);
    buttons_Sizer.add(this.smallScreenToggle);
    buttons_Sizer.addSpacing(20);
    buttons_Sizer.add(this.autoExitToggle);
    buttons_Sizer.addStretch();
    buttons_Sizer.add(ok_Button);
    buttons_Sizer.add(cancel_Button);
    
    //---------------------------------------------------------------
    // Vertically stack all the SectionBars and OK/Cancel button bar
    //---------------------------------------------------------------
    this.sizer = new VerticalSizer(this);
    this.sizer.margin = 6;
    this.sizer.spacing = 4;
    this.sizer.add(titleBar);
    this.sizer.add(titleSection);
    this.sizer.add(filesBar);
    this.sizer.add(filesSection);
    this.sizer.add(refBar);
    this.sizer.add(refSection);
    this.sizer.add(outputDirBar);
    this.sizer.add(outputDirSection);
    this.sizer.add(starDetectionBar);
    this.sizer.add(starDetectionSection);
    this.sizer.add(photometrySearchBar);
    this.sizer.add(photometrySearchSection);
    this.sizer.add(photometryBar);
    this.sizer.add(photometrySection);
    this.sizer.add(sampleGenerationBar);
    this.sizer.add(sampleGenerationSection);
    this.sizer.add(gradientBar);
    this.sizer.add(gradientSection);
    this.sizer.add(this.imageIntegrationBar);
    this.sizer.add(this.imageIntegrationSection);
    this.sizer.addSpacing(5);
    this.sizer.add(buttons_Sizer);
    
    starDetectionSection.hide();
    photometrySearchSection.hide();
    if (data.useAutoPhotometry)
        photometrySection.hide();
    if (data.useAutoSampleGeneration && data.manualRejectionCircles.length === 0)
        sampleGenerationSection.hide();
    if (data.minimumWeight === DEFAULT_INT_MIN_WEIGHT && data.sortByWeight && !data.imageIntegrationTemplateId)
        this.imageIntegrationSection.hide();
    
    //-------------------------------------------------------
    // Set all the window data
    //-------------------------------------------------------
    this.windowTitle = TITLE;
    this.adjustToContents();
    this.setFixedSize();
    let originalFilenameColumnWidth = self.files_TreeBox.columnWidth(COL_FILENAME);
    setDialogValues();
    setSelectedTarget();
    
    /**
     * @param {Boolean} displayWarning Display warning if a single target file is not selected
     * @return {String} selected target filename
     */
    this.getSelectedTargetFilename = function(displayWarning){
        let nodes = self.files_TreeBox.selectedNodes;
        if (nodes.length !== 1){
            if (displayWarning){
                (new MessageBox("Select a single target file", TITLE, StdIcon_Error, StdButton_Ok)).execute();
            }
            return undefined;
        }
        return nodes[0].text(COL_FULL_FILENAME);
    };
    
    /**
     * Read and cache the target image. Check it is compatible with ref image.
     * Return the full target filename
     * @param {Boolean} displayWarning Display warning if a single target file is not selected
     * @return {String} selected target filename
     */
    function getSelectedTarget(displayWarning){
        let targetFile = self.getSelectedTargetFilename(displayWarning);
        if (!targetFile){
            return undefined;
        }
        data.cache.setTgtFilename(targetFile);
        let tgtImage = data.cache.getTgtView().image;
        let refView = data.cache.getRefView();
        let refImage = refView.image;
        if (!refView.isNull && !refImage.isEmpty){
            if (tgtImage.isColor !== refImage.isColor){
                (new MessageBox("ERROR: Color depth must match the reference image.\n" + targetFile,
                        TITLE, StdIcon_Error, StdButton_Ok)).execute();
                return undefined;
            }
            if (tgtImage.width !== refImage.width || tgtImage.height !== refImage.height) {
                (new MessageBox("ERROR: Image dimensions must match the reference image.\n" + targetFile,
                        TITLE, StdIcon_Error, StdButton_Ok)).execute();
                return undefined;
            }
        }
        return targetFile;
    }
    
    function setSelectedTarget(){
        if (data.savedSelectedTarget){
            for ( let i = 0; i < self.files_TreeBox.numberOfChildren; ++i ){
                let node = self.files_TreeBox.child( i );
                if (node.text(COL_FULL_FILENAME) === data.savedSelectedTarget){
                    node.selected = true;
                }
            }
        }
    }
    
    function getReference(){
        let filename = data.cache.getRefFilename();
        if (filename){
            return filename;
        }
        (new MessageBox("Specify the reference file", TITLE, StdIcon_Error, StdButton_Ok)).execute();
        return undefined;
    }
    
    /**
     * Get noise from first FITSHeader. Return L or Red noise, scaled for exposure and airmass.
     * @param {FitsHeader} fitsHeader
     * @param {FITSKeyword[]} fitsKeywords
     * @param {Number | undefined} exposure
     * @param {Number | undefined} airmass
     * @returns {undefined|{{Number} noise, {String} type}
     */
    function getScaledNoise0(fitsHeader, fitsKeywords, exposure, airmass){
        let noiseData = fitsHeader.getNoise(fitsKeywords);
        if (noiseData){
            let index = noiseData.index === -1 ? 0 : noiseData.index;
            let scaledNoise = noiseData.noise[index];
            if (exposure){
                scaledNoise /= exposure;
            }
            if (airmass){
    //                console.note("Airmass: ", airmass.toPrecision(3), ", Noise: ", scaledNoise.toPrecision(3));
                scaledNoise /= Math.exp(-airmass);
    //                console.noteln(" / ", (Math.exp(-airmass)).toPrecision(3), " = ", scaledNoise.toPrecision(3));
            }
            return {noise:scaledNoise, type:noiseData.type[index]};
        }
        return undefined;
    }
    
    /**
     * @param {String[]} targetFiles
     * @returns {Number} maximum noise NOISExx value
     */
    function getMaxDisplayNoise(targetFiles){
        let maxValue = -1;
        let fitsHeader = new FitsHeader();
        for (let filename of targetFiles){
            try {
                let keywords = getFITSKeywords( filename );
                let exposure = fitsHeader.getExposure(keywords);
                let airmass = fitsHeader.getAirmass(keywords);
                let noiseData = getScaledNoise0(fitsHeader, keywords, exposure, airmass);
                if (noiseData){
                    maxValue = Math.max(maxValue, noiseData.noise);
                }
            } catch (error){
                continue;
            }
        }
        return maxValue > 0 ? maxValue : 1.0;
    }
    
    /**
     * @param {String[]} targetFiles
     * @returns {Number} maximum (PSFSGL / NOISExx) value
     */
    function getMaxPSFSGL(targetFiles){
        let maxValue = -1;
        let fitsHeader = new FitsHeader();
        for (let filename of targetFiles){
            try {
                let keywords = getFITSKeywords( filename );
                let psfsgl = fitsHeader.getPsfSignalWeight(keywords);
                if (psfsgl){
                    maxValue = Math.max(maxValue, psfsgl);
                }
            } catch (error){
                continue;
            }
        }
        return maxValue > 0 ? maxValue : 1.0;
    }
    
    /**
     * 
     * @param {String} filename
     * @param {Number} maxPSFSGL
     * @param {Number} maxNoise
     * @throws {Error} If FITS header cannot be read
     */
    function addTargetFileNode(filename, maxPSFSGL, maxNoise){
        let fitsKeywords = getFITSKeywords( filename );
        // File read exception was not thrown, so OK to create the node
        let node = new TreeBoxNode( self.files_TreeBox );
        let fitsHeader = new FitsHeader();
        let alt = fitsHeader.getAltitude(fitsKeywords);
        let airmass = fitsHeader.getAirmass(fitsKeywords);
        let filter = fitsHeader.getFilter(fitsKeywords);
        let exposure = fitsHeader.getExposure(fitsKeywords);
        let psfsgl = fitsHeader.getPsfSignalWeight(fitsKeywords);
        let noiseData = getScaledNoise0(fitsHeader, fitsKeywords, exposure, airmass);
        if (psfsgl){
            node.setText( COL_PSFSGL, (psfsgl/maxPSFSGL).toPrecision(3));
        }
        if (noiseData){
            node.setText( COL_NOISE, (noiseData.noise/maxNoise).toPrecision(3) + " " + noiseData.type);
        }
        if (alt){
            let altString = alt > 9.9995 ? alt.toPrecision(4) : '0' + alt.toPrecision(4);
            node.setText( COL_ALT, altString );
        }
        if (airmass){
            node.setText( COL_AIRMASS, airmass.toPrecision(5) );
        }
        if (filter){
            node.setText( COL_FILTER, filter );
        }
        if (exposure){
            node.setText( COL_EXPOSURE, exposure.toPrecision(5) );
        }

        node.setText( COL_FULL_FILENAME, filename);
        let fileOrPath = data.useFullPath ? filename : File.extractName(filename);
        node.setText( COL_FILENAME, fileOrPath );
        node.setToolTip( COL_FILENAME, filename); 
        let psfToolTip = "<p>Normalized PSF Signal Weight (0.0 to 1.0)</p>" +
                "<p>Set reference to an image with a high PSF Signal Weight.</p>" +
                "<p>Works best for narrow band or images taken from a dark site.</p>";
        node.setToolTip( COL_PSFSGL, psfToolTip);
        let noiseTip = "<p>Normalized noise estimate (0.0 to 1.0)</p>" +
                "<p>Set reference to an image with low noise.</p>" +
                "<p>Works best for moonlight, twilight and light polluted sites.</p>" +
                "<p>If both MRS and K-Sigma noise estimates are present, " +
                "all noise based weight algorithms will be less accurate because " +
                "MRS and K-Sigma are not compatible. " +
                "Try increasing your exposure time.</p>";
        node.setToolTip( COL_NOISE, noiseTip);
        node.setToolTip( COL_AIRMASS, "Set reference to an image with a low airmass.");
        node.setToolTip( COL_ALT, "Set reference to an image with a high altitude.");
        node.setToolTip( COL_FILTER, "All filters should be of the same type.");
        node.setToolTip( COL_EXPOSURE, "Exposure time in seconds.");
    }
    
    /**
     * Adjust the filename column width to fit the filename.
     */
    function adjustFilenameColumnWidth(){
        self.files_TreeBox.adjustColumnWidthToContents( COL_PSFSGL );
        self.files_TreeBox.adjustColumnWidthToContents( COL_NOISE );
        self.files_TreeBox.adjustColumnWidthToContents( COL_ALT );
        self.files_TreeBox.adjustColumnWidthToContents( COL_AIRMASS );
        self.files_TreeBox.adjustColumnWidthToContents( COL_EXPOSURE );
        self.files_TreeBox.adjustColumnWidthToContents( COL_FILTER );
        self.files_TreeBox.setColumnWidth( COL_FULL_FILENAME, 0);
        self.files_TreeBox.adjustColumnWidthToContents( COL_FILENAME );
        let colWidth = self.files_TreeBox.columnWidth(COL_FILENAME);
        if (colWidth > originalFilenameColumnWidth){
            // Add a 20 pixel margin to ensure there really is enough room. 
            self.files_TreeBox.setColumnWidth( COL_FILENAME, colWidth + 20);
        }
    }
    
    function setDialogValues(){
        // Ref filename
        let filterName;
        let refFilename = data.cache.getRefFilename();
        if (refFilename && File.exists(refFilename)){
            self.refTextBox.text = refFilename;
            let hdrEntries = self.getHdrEntries(refFilename);
            if (hdrEntries.pixelSize){
                data.pixelSize = hdrEntries.pixelSize;
            }
            if (hdrEntries.focalLength){
                data.focalLength = hdrEntries.focalLength;
            }
            filterName = hdrEntries.filter;
        } else {
            self.refTextBox.text = "";
            filterName = "NONE";
        }
        self.setPixelScaleFields(data.pixelSize, data.focalLength, filterName);
        
        // Tgt files
        self.fullPathCheckBox.checked = data.useFullPath;
        updateTargetImagesList();
        self.files_TreeBox.sort(COL_FILENAME, true);
        
        // -----------------
        // Output directory
        // -----------------
        self.outputDirTextBox.text = data.outputDir;

        // PostFix TextBox
        self.postFixTextBox.text = data.outputPostFix;
        self.noiseWeightTextBox.text = data.noiseWeightKeyword;
        
        // PreFix weight CheckBox
        self.weightPrefixCheckBox.checked = data.weightPrefix;
        self.rescaleCheckBox.checked = data.rescale;
        
        // Overwrite check box
        self.overwriteCheckBox.checked = data.overwrite;
        self.displayGradientToggle.checked = data.displayGradient;
        
        // On error ComboBox
        self.onErrorComboBox.currentItem = data.onErrorIndex;
        
        // -----------------
        // Star detection
        // -----------------
        self.refLogStarDetection_Control.setValue(data.logStarDetection);
        
        // -----------------
        // Star search
        // -----------------
        self.starFluxTolerance_Control.setValue(data.starFluxTolerance);
        self.starSearchRadius_Control.setValue(data.starSearchRadius);
        
        // -----------------
        // Photometry
        // -----------------
        self.autoPhotometryCheckBox.checked = data.useAutoPhotometry;
        self.apertureGrowthRate_Control.setValue(data.apertureGrowthRate);
        self.apertureAdd_Control.setValue(data.apertureAdd);
        self.apertureGap_Control.setValue(data.apertureGap);
        self.apertureBgDelta_Control.setValue(data.apertureBgDelta);
        self.limitPhotoStarsPercent_Control.setValue(data.limitPhotoStarsPercent);
        self.outlierRemoval_Control.setValue(data.outlierRemoval);
        self.linearRangeRef_Control.setValue(data.linearRangeRef);
        
        // -----------------
        // Sample Generation
        // -----------------
        self.autoSampleGenerationCheckBox.checked = data.useAutoSampleGeneration;
        self.sampleSize_Control.setValue(data.sampleSize);
        self.limitSampleStarsPercent_Control.setValue(data.limitSampleStarsPercent);
        self.sampleStarGrowthRate_Control.setValue(data.sampleStarGrowthRate);
        
        // --------------------
        // Gradient correction
        // --------------------
        self.gradientSmoothness_Control.setValue(data.gradientSmoothness);
        
        self.smallScreenToggle.checked = data.smallScreen;
        self.autoExitToggle.checked = data.autoExit; 
        // --------------------
        // ImageIntegration
        // --------------------
        self.imageIntegrationBar.checkBox.checked = data.imageIntegration;
        self.imageIntegrationSection.enabled = data.imageIntegration;
        self.minimumWeight_Control.setValue(data.minimumWeight);
        self.sortByWeightCheckBox.checked = data.sortByWeight;
        let nthItem = self.templateIconIdComboBox.findItem( data.imageIntegrationTemplateId, 0, true );
        if (nthItem < 0) nthItem = 0;
        self.templateIconIdComboBox.currentItem = nthItem;
    }
}

NsgDialog.prototype = new Dialog;
